# -*- coding: utf-8 -*-

"""
This is one of 2 files required to run the 5s RWIP simulation (i.e. calculate the motion of a RWIP) on either SpiNNaker or NEST.  The other is SpiNNakerRWIPSystem.py to create and run the system
on SpiNNaker and the other is NestRWIPSystem for NEST.

This script creates a neural RWIP forward model/neural system, i.e. a model that calculates one time step of the RWIP simulation
(i.e. calculates and outputs the next RWIP state (which is the next theta (anglular displacement of the inverted pendulum from the upright (equilibrium)
position (theta=0)) measured in radians, and thetadot (angular velocity) values) using the given theta and thetadot values (i.e. the current RWIP state)
and the torque/force value calculated by a PID controller to stabilise the RWIP.  The model contains the following input and output populations:
    i) input population thetaPop, which represents the theta input value (a value in this system is represented by the number of neurons in a population
       that fire).
    ii) input population thetaDotPop, which represents the thetadot input value. 
    iii) input neuron RWIPModActForceCell, which activates (fires) when the model is next to calculate a RWIP state, i.e. will receive theta, thetadot, and
         force values.  This is described below.  
    iv) input populations forcePop/forceBufferPop, which represents the force input value.  forcePop is a set of spike injectors in SpiNNaker and a set
        of if_cond_exp neurons in NEST, used to inject the force value into the model.  However, as it is unknown when the model running on SpiNNaker will receive
        the injected spikes, the forcePop is connected to forceBufferPop, which waits for spikes (a force value) from forcePop, holds them once received
        (i.e. forcepop activates the same neurons in forceBufferPop, which continuously fire) for a period of time, and then releases the spikes (the value)
        into the model (this is discussed below). Furthermore, to simulate a RWIP, more than one model is used so the same forcePop is used for each model
        to reduce the number of neurons/spike injectors used, i.e. is connected to each model's forceBufferPop, and, so only the forceBufferPop in model X
        (i.e. the model next to calculate a RWIP state) activates (receives the force value) the RWIPModActForceCell associated to model X  fires (activates)
        along with the forcePop.  Thus,  RWIPModActForceCell indicates that model X is next to calculate a RWIP state and enables model X's forceBufferPop to
        receive the force value from (be activated by) self.forcePop.  Note that, due to the above, the force value injected is delayed entering the model
        and therefore the theta and thetadot values are also delayed so that all values are received by the model at the same time (this is described below).
        Also, Note that a forceBufferPop is not required for NEST, as spikes are not injected into the system (as in SpiNNaker).  However  as it is used and
        built in for SpiNNaker, it is used for NEST.  The model will contain an extra input population initForcePop if its the first model in the sequence
        calculating the motion the RWIP, which is used to represent the initial force value, thus, is only used once, i.e. activated at the start of the
        simulation.  This is not a set of spike injectors so bypasses forceBufferPop.            
    v) output population nextThetaPop, which represents the next theta output value.  
    vi) output population nextThetaDotPop, which represents the next thetadot output value.  

The system implements a numerical solution of the RWIP equation of motion to calculate the next theta and thetadot values (next RWIP state) using the above
inputs:

nexttheta=theta+(thetadot*dt)
nextThetaDot=thetadot+(B*(Force-Friction)+Sin(theta)*A)*dt

where the calculation is implemented by a set of neural mathematical functions/components (see below).

See "Block D.J, Astrom K,J and Spong M.W (2007) The Reaction Wheel Pendulum.  Morgan and Claypool publishers" for more information about the RWIP equation
of motion.

Notes:

A value in this system is represented by the number of neurons that simulateously fire a spike, thus, the more neurons that fire, the greater the number represented. 
For example (example 1), if we had a population of 10 neurons and each neuron represented a value of 0.1, a value of 0.6 would be represented by 6 neurons (neurons:0-5)
firing.  Many values can be represented at the same time, where different populations represent different values.

A population can represent negative and positive values.  We do this by splitting a population in 2, where the first set of neurons represent negative
values and the rest represent positive values.  For example (example 2), if we had a population of
10 neurons, we could use neurons 0-4 to represent a negative number and 5-9 to represent positive neurons.  If a 0 value needs to be represented
(which is required for the force value) we add one neuron in between the positive and negative sets.  For example if we had 11 neurons, we could have neurons
0-4 representing negative values, neuron 5 represents a 0 value, and neurons 6-11 represent positive values.  The value associated to a neuron is referred to as
the neuron value.

For the system to work, a value needs to be represented by the first and consecutive neurons of a population.  For example, a value of -0.2 and 0.2 need to be
represented by neurons 0,1 and 5,6 in the above population (example 2).  If a value is not represented by a population in the correct way, a reordered version
of the value needs to be produced.  In other words, a reorder network is connected to such a population, which produces a reordered version of the value.

Neural Mathematical functions are created by connecting populations (layers) of neurons in specific ways and using specific weights so that the function's
output population represents the result of a particular calculation on the values given by the function's input populations, i.e. when a specific number of
neurons in the input populations fire, a specific number of neurons in the output population will fire representing the result.  Various neural mathematical
functions are connected to achieve the above equations, where inputs/outputs to each function are appropriately delayed so that the input values to a neural
function arrive at the same time.  See either SpiNNakerRWIPSystem.py and NestRWIPSystem.py for details about the time input values are received and output
values are sent by the model.

A highID pop is sometimes connected to a population, so that the value represented by the population (i.e. represented by
the first and consecutive neurons in, depending on the sign of the value, the set of neurons representing positive or negative values) is represented by one
neuron in the highID population, i.e. the neuron with the same index as the one with the highest index (last neuron) in the population that contributed to the
positive or negative value.  Using population example 1 above, a value of 0.5 would be represented by only one neuron (neuron 4) in the highId population.  We
connect a highID pop to a population when we require a value to be represented by only one neuron. We do this either for computational purposes and/or to
prevent SpiNNaker overloading.

With our current system configuration, a neuron with a large number of excitatory connections can overload SpiNNaker.  Thus, for SpiNNaker to run the model
correctly, we found a neuron can only have up to 12 connections.  However, neurons in many of the populations (i.e. neural layers in the neural mathematical
functions) require more than 12 connections to work.  To resolve this, when neurons in a population y require > 12 connections from neurons in population x,
an intermediate population between x and y is created called expand 1, where duplicates of each neuron r in population y are added to it, where the number of
duplicates for each neuron r = (the required Number of Connections to neuron r)/12 + 1 for any remainder.  The
connections from population x associated to a neuron r in y are then spread across all of the duplicates of r in expand 1, and the duplicate neurons of r
in expand 1 are connected to neuron r in y.  For example, if 36 neurons in x need to be connected to a neuron r in y, an expand 1 population between them
is added and 3 neurons associated to r is added to expand 1, and connected as follows:

x                       expand 1            y
(12 neurons)--------->  one neuron--------->neuron r
(12 neurons)--------->  one neuron--------->
(12 neurons)--------->  one neuron--------->

Thus, when neurons in population x need to activate a neuron r in y, neurons in x will activate a neuron in expand 1 associated to r, which will
activate the corresponding neuron in y.  If the number of connections between neurons in expand 1 and their corresponding neuron r in y is >12, another
intermediate layer (expand2) between expand1 and y is added, and the connections are spread in the same way.  Whether expand 1 and 2 populations are required,
and their sizes, are determined at run time.  This system can only add 2 expand populations.

Note, this system currently only works with a neural time step of 1.0ms.  In addition, our neural parameter values are set so a post synaptic neuron fires X
(connection delay) ms after a pre synaptic neuron if it receives sufficient input (i.e. self.minWt below).  SpiNNaker achieves this but, due to NEST's
neural simulation method, NEST does not, i.e. a post synaptic neuron fires X+1 ms after if it receives sufficient input.  This could not be reduced.

"""

from numpy import array, hstack, linspace, pi
from numpy import zeros, cos, sin, arange, around
from scipy.integrate import odeint
from matplotlib import pyplot as plt
from matplotlib import animation
from matplotlib.patches import Rectangle
from sympy import symbols
from sympy.physics.mechanics import *
from numpy import array, hstack, zeros, linspace, pi, ones
import os
import math
from decimal import *
import numpy as np
from pyNN.utility.plotting import Figure, Panel  ###this has a warning
import spynnaker8.external_devices as externaldevices

#http://spinnakermanchester.github.io/
# http://apt.cs.manchester.ac.uk/projects/SpiNNaker/

class RWIPClass:

    def __init__(self, sim, minWt, dt,nest,additionalWt,ForceBufferDuration,numModels,inputMinMax,ModelID,frictVal,BDt,ADt,RWIPOutputDelay,InputPopsOfNextRWIPMod,timePerLayer,NeuronValues,inputPops):

        ###################RWIP Model And Neural Simulator Parameter Values
        self.ModelID=ModelID  #this is the ID given to the model, which is the position (ID) of the model in the sequence of connected RWIP models
        self.NextConnectedModel=[]     #the next RWIP model connected to this model.        
        self.minWt=minWt          ##the minimum weight (input) to cause an if_cond_exp neuron to fire (activate)
        self.thetaMinMaxVals=inputMinMax[0] #min and max values that can be represented by thetaPop
        self.thetaDotMinMaxVals=inputMinMax[1] #min and max values that can be represented by thetaDotPop
        self.forceMinMaxVals=inputMinMax[2]  #min and max values that can be represented by forcePop/forceBufferPop
        self.thetaNeuronValue=NeuronValues[0]     #the value each neuron in population thetaPop represents
        self.thetaDotNeuronValue=NeuronValues[1]  #the value each neuron in population thetaDotPop represents
        self.forceNeuronValue=NeuronValues[2]     #the value each neuron in population forcePop/forceBufferPop represents
        self.RWIPmodNumCells=0    #the number of neurons used in the system (model).  This is calculated as the model is built.
        self.sim=sim              #neural simulator object
        self.nest=nest            #neural simulator being used.  1 for NEST, 0 for SpiNNAker
        self.timePerLayer=timePerLayer #the time (neural) after a pre synaptic neuron a post synaptic neuron will fire if it receives sufficient input
                              #(i.e. self.minWeight) and the connection delay and neural time step is 1.0ms.  It is 1.0ms in SpiNNaker and 2.0ms in NEST.  A
                              #post synaptic neuron will always fire 1.0ms after the delay in NEST.  self.timePerLayer is used to calculate appropriate
                              #connection delays within the RWIP model. 
        self.additionalWt=additionalWt  #additional weight (input) required to cause some neurons to fire.  This is important for SpiNNaker.
        self.inhibWt=0.1        ##the maximum inhibition weight used in the model
        self.RWIPOutputDelay=RWIPOutputDelay      #the output delay (neural time), i.e. the delay sending the output ( the next theta and thetadot values) from
                                                  #this model to its next connected model.
        self.ForceBufferDuration=ForceBufferDuration  ##This defines the period of time (neural) the forceBufferPop is enabled to receive the force value from forcePop.
                                   #In other words, its the time a forceBufferPop will wait for spikes from self.forcePop and the maximum time forceBufferPop can
                                   #hold the spikes (be activated) for, i.e. the time they are held for is dependent on when the spikes are received (see the start of the file for
                                   #more information.
                                   
        self.MaxConnectionsPerCell=12  #maximum number of connections per neuron.  This is important for SpiNNaker.  
        self.RWIPCalcTimePlusDelay=0  #the time (neurally) it takes the RWIP model to calculate the next RWIP state, i.e. the next theta and thetadot values.  This
                                     #is calculated as the model is built.  It also includes the model's output delay (i.e. self.RWIPOutputDelay+2.0 ms), see above
                                     #for more information.
                                        
        #################The RWIP model inputs (input populations). forceBufferPop, initForcePop, thetaDotPop and thetaPop are connected to the input populations
        #################of their corresponding neural functions. forcePop and RWIPModActForceCell are connected to forceBufferPop, where RWIPModActForceCell
        #################is also connected to the force buffer network that manages forceBufferPop.  (See function CreateRWIPModel
        #################for more information).
        self.forcePop=[]                  
        self.thetaPop=[]       
        self.thetaDotPop=[]     
        self.initForcePop=[]            ##an self.initForcePop is only required if the model is the first in a sequence of models.
        self.forceBufferPop=[]          
        self.RWIPModActForceCell=self.createPopulation(1,1)   #create self.RWIPModActForceCell.  
        
        if(self.ModelID==0):   ##if we are creating the first RWIP model
            ##create the input populations self.forcePop, self.thetaPop,self.thetaDotPop and self.initForcePop
            self.CreateInputPopulations()  
        else:
            self.thetaPop=inputPops[0]      #self.thetaPop is the output population (nextThetaPop) of the previous connected RWIP model.
            self.thetaDotPop=inputPops[1]   #self.thetaDotPop is the output population (nextThetaDotPop) of the previous connected RWIP model.
            self.forcePop=inputPops[2]      #self.forcePop is the same for all the models.          

        ##create force buffer network, including self.forceBufferPop, which, with self.RWIPModActForceCell, manages self.forceBufferPop.  Also connect
        ##self.forcePop to self.forceBufferPop.
        self.CreateAndManageForceInput()

        #################The RWIP model outputs (output populations), which are the outputs populations of the last neural functions implemented (see function
        #################CreateRWIPModel for more info).
        self.outputPops=InputPopsOfNextRWIPMod  #contains the output populations of the model, i.e. nextThetaPop and nextThetaDotPop.
                                                ## InputPopsOfNextRWIPMod contains the input populations (i.e. the populations used to represent the theta and
                                                ##thetadot values) of the next connected RWIP model, and therefore will be used as the output
                                                ##populations of this model, i.e. used to represent the next theta and thetadot values.  if empty, output
                                                ##populations will be created.
        
        #################RWIP Parameter Values                        
        self.dt=dt              ##RWIP simulation time step 
        self.ADt=ADt            #A*self.dt
        self.BDt=BDt            #B*self.dt.  A and B are based on the physical quantities of the RWIP (see SpiNNakerRWIPModel.py or NestRWIPModel.py for more
                                #information)
        self.frictVal=frictVal  #friction torque value

        ##create RWIP model
        self.CreateRWIPModel()

        
    def CreateInputPopulations(self):
        ###create the input populations (i.e. thetaPop, thetaDotPop and forcePop) of the RWIP model.  These represent the theta, thetadot and force
        ###input values.
        
        ####create the theta pop
        model=1              #use if_cond_exp neurons
        addzero=0            #a neuron to represent a value of 0 is not required.
        self.thetaPop=self.createValuePop(self.thetaNeuronValue, self.thetaMinMaxVals, model,addzero,[]) ##[]== no spike injection port

        ####create the thetadot pop
        self.thetaDotPop=self.createValuePop(self.thetaDotNeuronValue, self.thetaDotMinMaxVals, model,addzero,[]) ##[]== no spike injection port
                                                                                    
        ####create forcePop (which are if_cond_exp neurons for NEST and spike injectors for SpiNNaker) and initForcePop.
        addzero=1  #a neuron in the populations are required to represent a value of 0.
        self.initForcePop=self.createValuePop(self.forceNeuronValue, self.forceMinMaxVals, model,addzero,[]) ##[]== no spike injection port

        if(self.nest==0):
            forceport=12345  #port on which the spiNNaker machine should listen for force packets
            model=3 ##injector neurons
            self.forcePop=self.createValuePop(self.forceNeuronValue, self.forceMinMaxVals,model,addzero,forceport) 
        else:
            self.forcePop=self.createValuePop(self.forceNeuronValue, self.forceMinMaxVals,model,addzero,[]) ##[]== no spike injection port

        
    def createValuePop(self, neuronValue, MinMaxVals,model,addzero,port):
        
        #This function creates a population of neurons (using the neuron model given by model) to represent a set of values given by both the
        #min and max values in MinMaxVals and the value each neuron represents (neuronValue).  A neuron representing a value of 0 can also be
        #added when addzero==1.

        #model can take 3 values, 1,2 and 3. 1 and 2 are for if_cond_exp neurons.  Both have exactly the same parameter values but 1 has a refractory
        #period of 3.0ms and 2 has no refractory period.  3 is for injector neurons, which are injected via the port number given by port.

        minValue=MinMaxVals[0]
        maxValue=MinMaxVals[1]
       
        numNeg=int(np.ceil(abs(minValue)/neuronValue)) #number of neurons required to represent negative values          
        numPos=int(np.ceil(maxValue/neuronValue)) #number of neurons required to represent positive values

        popSize=numPos+numNeg  #size of the population
        
        if(addzero==1): #if a neuron is required to represent a value of 0.
            popSize+=1
        
        #create the population and attach its population information.
        NewPops=self.createPopAndInfoSet(popSize,numNeg,neuronValue,minValue,maxValue,model,port)

        return NewPops

    def createPopulation(self, popSize, whichModel):
        ##This creates a population of if_cond_exp neurons of size popSize, where whichModel==1 has a refractory period of 3.0ms and
        ##whichModel==2 has no refractory period.

        CELL_PARAMS = {'v_thresh':-52.0, 'v_reset' : -65.0, 'tau_m' : 15.0, 
                'tau_refrac': 3.0 , 'tau_syn_E':0.35,
                'e_rev_I': -130.0, 
                'tau_syn_I': 0.35,'e_rev_E': 0.0, 
                'v_rest' : -65.0,'i_offset':0.0, 'cm':1.0}
        CELL_PARAMS2 = {'v_thresh':-52.0, 'v_reset' : -65.0, 'tau_m' : 15.0, 
                'tau_refrac': 0.0 , 'tau_syn_E':0.35,
                'e_rev_I': -130.0, 
                'tau_syn_I': 0.35,'e_rev_E': 0.0, 
                'v_rest' : -65.0,'i_offset':0.0, 'cm':1.0} ##no tau_refrac               

        if(whichModel==1):
            Pop=self.sim.Population(popSize,self.sim.IF_cond_exp,CELL_PARAMS)
        elif(whichModel==2):
            Pop=self.sim.Population(popSize,self.sim.IF_cond_exp,CELL_PARAMS2)

        self.RWIPmodNumCells+=popSize  #increment the number of neurons used in the RWIP system.
        return Pop

    def createInjPopulation(self,numNeuronsInEachPop,port):
        #This creates a population of spike injector neurons of size numNeuronsInEachPop, where the spikes are injected via the given port on SpiNNaker.

        cell_params_spike_injector = {
            'port': port,  # The port on which the spiNNaker machine should listen for packets.
        }        

        ##create the population of injector neurons
        Pop=self.sim.Population(numNeuronsInEachPop,externaldevices.SpikeInjector(**cell_params_spike_injector))
        return Pop

    def ActivateCellsUsingValue(self, Population, value, inclZero,sender):
        #This function activates a specific number of neurons in the given population so that it represents the given value (i.e. activates the first and
        #consecutive neurons in, depending on the sign of the value, the set of neurons representing negative or positive values) . More information
        #about how values are represented is given at the start of the file.

        #An injector connection is provided in sender when the value needs to be injected into SpiNNaker. 

        #if the population includes a neuron to represent a value of 0, inclZero needs to = 1.

        actVoltage=-51.0    #this is a voltage > the membrane potential threshold.  This is used to activate a neuron
        rest=-65.0          #resting membrane potential
 
        #extract population information from Population
        [NeuronValue,TotalNumNeurons,NumNegNeurons,NumPosNeurons]=self.extractPopInfo(Population)

        if(inclZero==1):
            
            NumPosNeurons-=1 #-1 as the 0 neuron is included in the number of positive neurons.

        #initialise the neuron list.  This contains the neurons to activate.   
        if(sender!=[]):
            InitialiseNeuronList=[]
        else:
            InitialiseNeuronList=np.ones(TotalNumNeurons)*rest

        ##adddec is used to add a tiny amount to the value to ensure it rounds correctly
        adddec=0.0
        if(value<0):
            adddec=-0.00001
        elif(value>0):
            adddec=0.00001
        else:
            adddec=0.0

        ##calculate the number of cells to activate

        if(((value==0.0)and(inclZero==0))or(NeuronValue==0)):
            ##do not activate any cells.
            numberOfCellsToActivate=0
        else:

            numberOfCellsToActivate=int(np.fix(abs((value/NeuronValue)+adddec))) #calculate the number of neurons to activate
            #if the value is 0
            if((numberOfCellsToActivate==0)and(inclZero==1)):
                ##activate the cell that represents 0
                if(sender!=[]):
                    InitialiseNeuronList+=[NumNegNeurons]
                else:
                    InitialiseNeuronList[NumNegNeurons]=actVoltage
                
            else:

                ##activate the first and consecutive neurons in, depending on the sign of the value, the set of neurons representing negative or positive values.
                for eachCellToActivate in range(0, numberOfCellsToActivate):  
                    #if a negative value, add the first and the correct number of consecutive neurons representing negative values to the list.
                    if(value<0):
                        if(eachCellToActivate<NumNegNeurons):
                            if(sender!=[]):
                                InitialiseNeuronList+=[eachCellToActivate]
                            else:
                                InitialiseNeuronList[eachCellToActivate]=actVoltage
                        else:
                        #    print "WARNING:Number Of Cells To Activate is larger than the number of negative neurons given.
                            break
                    #if a positive value, add the first and the correct number of consecutive neurons representing positive values to the list.
                    elif(value>0):
                        if(eachCellToActivate<NumPosNeurons):
                            neuronID=eachCellToActivate+NumNegNeurons
                            if(inclZero==1):
                                neuronID+=1   #if a neuron representing a value of 0 is in the population, we want to bypass this.
                            if(sender!=[]):
                                InitialiseNeuronList+=[neuronID]
                            else:
                                InitialiseNeuronList[neuronID]=actVoltage
                        else:
                            #       print "WARNING:Number Of Cells To Activate is larger than the number of positive neurons given.  
                            #    ####self.#pause()
                            break
                        
        ##activate the neurons
        if(sender!=[]):
            sender.send_spikes(Population[0].label,InitialiseNeuronList,send_full_keys=True)
        else:
            Population[0].initialize(v=InitialiseNeuronList)  

    def oneToOneExcitatoryConnector(self,inputPop, OutputPop,weight):
        ##This function connects inputPop to OutputPop via one to one excitatory connections using the given weight
        ExcConnector=[]
        inputSize=inputPop[1][0]
        for eachCell in range(0,inputSize):
            
            ExcConnector+=[(eachCell, eachCell, weight, 1.0)]

        if isinstance(inputPop, list):   #if inputPop contains information along with the input population, extract the population
            inputPop=inputPop[0]
        if isinstance(OutputPop, list):   #if OutputPop contains information along with the output population, extract the population
            OutputPop=OutputPop[0]

        #connect the populations
        self.connectPopsUsingExcitatory(inputPop, OutputPop,ExcConnector)

    def oneToAllInhibitoryConnector(self, inputPop, neuronInInput, outputPop, weight, delay):
        #This function connects (using inhibitory connections) the neuron neuronInInput in inputPop to all of the neurons in the outputPop using the given
        #weight and delay

        InhConnector=[] #inhibitory connections 

        outputPopSize=outputPop[1][0]
        #create the connections
        for eachOutCell in range(0,outputPopSize): #for each cell in the outputPop
            InhConnector+=[(neuronInInput, eachOutCell, weight, delay)]  

        #connect the populations
        self.connectPopsUsingInhibitory(inputPop,outputPop[0],InhConnector)

    def oneToAllExcitatoryConnector(self, inputPop, neuronInInput, outputPop, weight, delay):
        #This function connects (using excitatory connections) the neuron neuronInInput in inputPop to all of the neurons in the outputPop using the given
        #weight and delay

        ExcConnector=[] #excitatory connections 

        outputPopSize=outputPop[1][0]
        
        #create the connections
        for eachOutCell in range(0,outputPopSize): #for each cell in the outputPop
            ExcConnector+=[(neuronInInput, eachOutCell, weight, delay)]
            
        #connect the populations    
        self.connectPopsUsingExcitatory(inputPop,outputPop[0],ExcConnector)

        
    def genConnsNegPosHighIDToOutPops(self,HighIDNegPopAndInfo,HighIDPosPopAndInfo,HighIdNeuronsResNegOut,HighIdNeuronsResPosOut,OutputPopAndInfo,Layers,connectPops,connsToProduce):
        ##This function generates and returns connections (the connections to generate is specified in connsToProduce)  between both the given input
        ##populations: HighIDNegPop (i.e. HighIDNegPopAndInfo[0], a HighID population representing negative values) and a HighIDPosPop
        ##(i.e. HighIDPosPopAndInfo[0]) a HighID population representing positive values) and an output population (OutputPopAndInfo[0]), so that the output
        ##population represents the result of a calculation (given by connsToProduce) on the value given by one of the input populations (i.e. only one of
        ##populations will provide a value).  More specifically, the result is represented by the first and consecutive neurons in, depending on the sign of
        ##the result, the set representing positive or negative values in the output population.

        #In summary, it generates the connections for the following network:

        ##HighIDNegPop---->OutputPop 
        ##HighIDPosPop---->OutputPop

        #HighID populations are described at the beginning of the file.

        ##HighIDNegPopAndInfo and HighIDPosPopAndInfo can be the same HighId population, where the population contains neurons that represent
        #positive and negative values.

        ##the connections that can be generated between the inputs (HighIDNegPop and HighIDPosPop) and the output pop are as follows:
        ##connsToProduce=0, the output pop represents the result of the calculation: Sin(theta)*self.ADt, thus HighIDNegPop and HighIDPosPop represent theta values
        ##connsToProduce=1, the output pop represents the result of the calculation: (F-Fr)*B*Dt, thus HighIDNegPop and HighIDPosPop represent force values
        ##connsToProduce=2, the output pop represents the result of the calculation: ThDot*Dt, thus HighIDNegPop and HighIDPosPop represent thetadot values
        ##connsToProduce=3, the value given by a neuron in a HighID pop is represented by the first and consecutive neurons in, depending on the sign of
        ##                  value, the set representing positive or negative values in the output pop.                   
        
        ##HighIdNeuronsResNegOut contains a list of 3 elements in relation to the neurons in HighIDNegPopAndInfo and HighIDPosPopAndInfo
        ##that will result in a negative value (result), i.e. the output population representing a negative value.  HighIdNeuronsResNegOut[0]=number of neurons,
        ##HighIdNeuronsResNegOut[1]=the first neuron in HighIDNegPop (i.e. HighIDPosPopAndInfo[0]) that results in the output population representing a negative
        ##value, and HighIdNeuronsResNegOut[2]=the value of the first neuron in HighIDNegPop that results in the output population representing a negative value

        ##HighIdNeuronsResPosOut is the same as above but in relation to the neurons that produce a positive result (value).

        ##layers contains the number of neural layers prior to HighIDNegPop and HighIDPosPop that this function needs to take into account when calculating the
        ##delays so that the outputPop receives spikes from both HighIDNegPop and HighIDPosPop at the same time.

        ##The function will also produce expand 1 and expand 2 population connections between highIDNegPop and the output pop and highIDPosPop and the output
        ##pop.  For example, highIDNegPop--->exp1-->exp2-->Out.  However, they are only created if required, i.e. when neurons in the output population
        ##require > self.MaxConnectionsPerCell.  See the start of the file for more information.

        ##The function will connect the populations (and create expand 1 and 2) when connectPops=1

        ##The function will return the number of neural layers required/created for this network and the input and output populations and connections

        numberOfLayers=0               ##number of neural layers created by this function
        AllConnAndNetInfo=[]           ## two sets of connections (one set between HighIDNegPop and OutPop (may also contain connections for exp1 and 2),
                                        ##and the same for HighIDPosPop
        ExtraLayersEachAddSub=[]        ##contains two numbers, one states how many extra layers (i.e. either an expand1 and an expand2) were required/added 
                                        ##between highIDPopNeg and the output, and the other between highIDPopPos and the output pop

        if(Layers==[]):     #no prior layers
            Layers=[0,0]
        
        #######################
        ##if an output population is not given, create one using the specification provided in OutputPopAndInfo
        if(OutputPopAndInfo[0]==[]):  # if there is no given output pop, create one
            addzero=0 #a neuron is not required to represent a 0 value
            model=1 
            OutPopAndInfo=self.createValuePop(OutputPopAndInfo[2], OutputPopAndInfo[3],model,addzero,[])

        else:      ##if an output population is given
            OutPopAndInfo=OutputPopAndInfo

        [OutNeuronValue,outPopSize,outPopNegSize,outPopPosSize]=self.extractPopInfo(OutPopAndInfo)  #extract output population information

        
        ######################################################
        ##connect the highIDNegPop (or if the same population is given in HighIDNegPopAndInfo and HighIDPosPopAndInfo,
        ##connect the neurons that produce a negative result) to the neurons representing negative values in the outputPop 

        firstCellHighIDNeg=HighIdNeuronsResNegOut[1]   #the id of the first neuron that produces a negative result.
        HighIDPopNegResSize=HighIdNeuronsResNegOut[0]     #number of high ID neurons that produce a negative result
        firstHighIDValNegRes=HighIdNeuronsResNegOut[2] #the value of the first highID neuron that produces a negative result.

        [highIDNegNeuronValue,HighIDPopSize,HighIDNegNum,HighIDPosNum]=self.extractPopInfo(HighIDNegPopAndInfo) #extract highIDNeg population information

        highIDNegNeuronValue*=-1.0                   ##the neurons represent negative values.
        HighIDNegPop=HighIDNegPopAndInfo[0]          ##highID population    
        firstCellOutNeg=0     ##index of the first neuron representing negative values in the output pop.

        ##generate connections between the HighIDNegPop starting at firstCellHighIDNeg to HighIDPopNegResSize and output pop from firstCellOutNeg to outPopNegSize
        ConnAndNetInfoNeg=self.genConnsBetHighIDAndOutPops(firstCellHighIDNeg,firstCellOutNeg,HighIDPopNegResSize,outPopNegSize,highIDNegNeuronValue,OutNeuronValue,connsToProduce,HighIDNegNum,firstHighIDValNegRes) 


        NetInfoNeg=ConnAndNetInfoNeg[1]    ####extract out the network information 
        ExtraNumLayers=NetInfoNeg[4]       ##number of layers added e.g.. expand1 and expand2.
        ExtraLayersEachAddSub+=[ExtraNumLayers+Layers[0]]
        
        AllConnAndNetInfo+=[ConnAndNetInfoNeg]

        ######################################
        ##connect the highIDPosPop (or if the same population is given in HighIDNegPopAndInfo and HighIDPosPopAndInfo,
        ##connect the neurons that produce a positive result) to the neurons representing positive values in the outputPop 
    
        firstCellOutPos=outPopNegSize

        firstCellHighIDPos=HighIdNeuronsResPosOut[1] #the id of the first neuron that producs a positive result.
        HighIDPopPosResSize=HighIdNeuronsResPosOut[0] #number of high ID neurons that produces a positive result
        firstHighIDValPosRes=HighIdNeuronsResPosOut[2] #the value of the first highID neuron that produces a positive result.

        [highIDPosNeuronValue,HighIDPopSize,HighIDNegNum,HighIDPosNum]=self.extractPopInfo(HighIDPosPopAndInfo) #extract highIDPos population information

        HighIDPosPop=HighIDPosPopAndInfo[0]

        ##generate connections between the HighIDPosPop starting at firstCellHighIDPos to HighIDPopPosResSize and output pop from firstCellOutPos to outPopPosSize
        ConnAndNetInfoPos=self.genConnsBetHighIDAndOutPops(firstCellHighIDPos,firstCellOutPos,HighIDPopPosResSize,outPopPosSize,highIDPosNeuronValue,OutNeuronValue,connsToProduce,[],firstHighIDValPosRes)

        NetInfoPos=ConnAndNetInfoPos[1]
        ExtraNumLayers=NetInfoPos[4]
        
        ExtraLayersEachAddSub+=[ExtraNumLayers+Layers[1]]     
        AllConnAndNetInfo+=[ConnAndNetInfoPos]                ##add the new connection set

        ################################################################
        ##modify the connection delays between the 2 HighIDPops and the OutputPop so spikes from the HighIDPops activate the outputPop at the same time.  This
        ##is important if, for example, one network contains an expand1 pop and the other does not
        NumSetsOfConnections=2  ##number of sets of connections in AllConnAndNetInfo
    
        ##modify the delays
        ModAllConnAndNetInfo=self.ModifyConnectionDelays(NumSetsOfConnections,AllConnAndNetInfo,ExtraLayersEachAddSub) 
        numberOfLayers+=max(ExtraLayersEachAddSub)
        
        negConnections=ModAllConnAndNetInfo[0]
        posConnections=ModAllConnAndNetInfo[1]

        ##########################################
        ##connect the populations if required and return the connections and populations
        inOutPopsAndConnections=[OutPopAndInfo,HighIDNegPop,HighIDPosPop,negConnections,posConnections]
        if(connectPops==1):
            ##connect the network highIDPop--->expand1 (if required) --> expand2 (if required)---> Output
            self.connectFinalLayerOfNeuralFunction(inOutPopsAndConnections)

        return [inOutPopsAndConnections,numberOfLayers]

    def generateReorderConnections(self, PopSize, StartInputCell, StartReorderCell):
        #This function generates and returns the connections between the first neuron (StartInputCell) and last neuron (StartInputCell+PopSize)
        #of an input population and a reorder population, so that the first (StartReorderCell) and
        #consecutive neurons (where the last is StartReorderCell+PopSize) in the reorder population represent the reordered version of the value given by the
        #neurons (StartInputCell to StartInputCell+PopSize) in the input population.
        
                                                                                                              
        ExcConnector=[]  ##connections between the input and reorder population.

        for eachInputCell in range(StartInputCell,PopSize+StartInputCell):
            reorderCellNum=1.0
            for eachReorderCell in range(StartReorderCell,PopSize+StartReorderCell): #for each reorder neuron
                WeightProp=1.0/reorderCellNum   #proportion of the weight to associate to the connection
                wtB=self.minWt*WeightProp
                ExcConnector+=[(eachInputCell, eachReorderCell, wtB, 1.0)]
                reorderCellNum+=1

        return ExcConnector

    def connectHighIDPopToOutPop(self,HighIDPop, OutPop,connectionsAndNetInfo):
        #This connects an input (HighIDPop) and Output population (OutPop) using the given connections (connectionsAndNetInfo).
        #if connections associated to expand 1 and 2 pops are included in connectionsAndNetInfo, which are intermediate populations between
        #the input and output (see the start of the file for more information) the function creates the expand 1 and 2 pops
        #and connects them.
        
        netInfo=connectionsAndNetInfo[1] #extract network information.
        expand1PopRequired=netInfo[2]
        if(expand1PopRequired==0): ##if no expand 1 and 2 populations are required, connect the input directly to the output.  
            connections=connectionsAndNetInfo[0]
            ExcConnectorHighIDToOut=connections[4] ##connections between input and output.            
            self.connectPopsUsingExcitatory(HighIDPop, OutPop,ExcConnectorHighIDToOut)
        else: 
            #create the expand 1 and 2 pops and connect them to the input (HighID) and output populations.
            self.connectHighIDToExpandsToOutPop(HighIDPop,OutPop,connectionsAndNetInfo)

    
    def createNetToAdd3HighIDSubs(self,HighIDinputPop,firstCellHighIDInputSub1,HighIDInputSubSize1,HighIDInputSubSize2,HighIDInputSubSize3,AddPop):
        ##This function creates a network between a highID population (i.e. between the neurons in 3 consecutive subsets of HighIDinputPop
        ##(neurons in sub1 are firstCellHighIDInputSub1 to HighIDInputSubSize1, neurons in sub 2 are firstCellHighIDInputSub2=HighIDInputSubSize1+firstCellHighIDInputSub1
        ##to firstCellHighIDSub2+highIDSubSize2 and neurons in sub 3 are firstCellHighIDInputSub3=firstCellHighIDInputSub1+HighIDInputSubSize1+HighIDInputSubSize2
        ##to firstCellHighIDInputSub3+HighIDInputSubSize3) and an output (add) population (i.e. one subset of neurons firstCellAddSubset=firstCellHighIDInputSub1 to
        ##HighIDInputSubSize1+HighIDInputSubSize2+HighIDInputSubSize3), so the first and consecutive neurons in the subset of the output (AddPop) population
        ##repesents the total of the 3 values given by the 3 HighIDinputPop subsets.  In other words, when a specific neuron in each of the three subsets
        ##fire, a specific number of neurons in the output population will fire.

        ##The function returns the number of neural layers created by this function.

        ###HighID populations are described at the beginning of the file.

        ##The network is as follows:

        #HighIDinputPop (subset 1)
        #HighIDinputPop (subset 2) -----> expand1 (if req) ----> expand2 (if req)----->interAddPop--->HighIDinterPop--->HighIDPop (subset 1)--->expand1 ( if req)---> expand2 (if req)-->AddPop (subset 1)
        #HighIDinputPop (subset 3)------------------------------------------------------------------------------------->HighIDPop (subset 2)

        #interAddPop (of size interAddSize=HighIDInputSubSize1+HighIDInputSubSize2) represents the total of the first 2 HighIDinput subsets, which is connected
        ##to a new HighIDpop (HighIDinterPop of size interAddSize).  HighIDinterPop is connected to subset 1 (neurons 0 to HighIDInputSubSize1+HighIDInputSubSize2)
        ##of another HighIDpop (HighIDPop of size
        #HighIDPopSize=HighIDInputSubSize1+HighIDInputSubSize2+HighIDInputSubSize3) and subset 3 of the HighIDinput is connected to subset 2 of HighIDPop (neurons
        #in subset 3 are connected to neurons from HighIDInputSubSize1+HighIDInputSubSize2 to HighIDInputSubSize1+HighIDInputSubSize2+HighIDInputSubSize3 in HighIDPop).
        #The total of the values of the 2 subsets of HighIDPop is represented by AddPop, thus AddPop represents the total of the values in the 3 subsets in HighIDinput.

        ##As the network above shows, the function will also produce expand 1 and expand 2 population connections if required between the HighIDinput and interAddPop
        ##and/or the HighIDPop and Add populations.  Expand 1 and 2 populations are required if 1 or more neurons in the output (add) pop subset require >
        ##MaxConnectionsPerCell connections.  expand 1 and 2 are described at the start of the file.        

        numLayers=3   #number of neural layers created by this function.  It creates a minimum of 3 (i.e.interAddPop--->HighIDinterPop--->HighIDPop).
                      #numLayers is increased if expand 1 and 2 pops are created/required.
        model=1       #use if_cond_exp neurons

        ####################create a network to add the first two HighIDinput subsets
        firstCellHighIDInputSub1=firstCellHighIDInputSub1                   #start cell in HIGHIDinput subset 1
        firstCellHighIDInputSub3=firstCellHighIDInputSub1+HighIDInputSubSize1+HighIDInputSubSize2  # #start cell in HIGHIDinput subset 3
        firstCellInterAddPop=0                                 ##the first neuron in interAddPop

        interAddSize=HighIDInputSubSize1+HighIDInputSubSize2   #interAddPop size
        
        interAddPop=self.createPopulation(interAddSize,model)  ##create interAddPop

        ##generate connections between HighIDinput (subsets 1 and 2) --> expand1 (if req) --->expand2 (if req) --> interAddPop so that interAddPop
        ##represents the total of the values in HighIDinput (subsets 1 and 2)
        connSub1A2ToInterAndNetInfo=self.generateConns2HighIDSubsTo1AddSub(HighIDInputSubSize1,HighIDInputSubSize2,firstCellHighIDInputSub1,firstCellInterAddPop)
        netInfo=connSub1A2ToInterAndNetInfo[1]
        extraLayers=netInfo[4]
        numLayers+=extraLayers         ##the number of extra layers required, i.e. the number of expand pops between between HighIDinput and interAddPop

        #use the connections in connSub1A2ToInterAndNetInfo to connect HighIDinputPop and interAddPop (and create and connect expand 1 and 2 pops
        #if included in connSub1A2ToInterAndNetInfo)
        self.connectHighIDPopToOutPop(HighIDinputPop, interAddPop,connSub1A2ToInterAndNetInfo)

        ##############################################connect interAddPop to HighIDinterPop
        HighIDinterPop=self.createPopulation(interAddSize,model)
        delay=1.0
        IntAddToHighIDinterConnectors=self.HighIDPopConnections(interAddSize, firstCellInterAddPop, delay) ##generate connections between interAddPop and HighIDinterPop
        self.connectPopsUsingExciteAndInhib(interAddPop,HighIDinterPop,IntAddToHighIDinterConnectors) #connect interAddPop to HighIDinterPop

        ###################################connect HighIDinterPop to HighIDPop (subset 1)
        HighIDPopSize=HighIDInputSubSize1+HighIDInputSubSize2+HighIDInputSubSize3  #size of HighIDPop (this is the size of the ADDPop subset)
        HighIDPop=self.createPopulation(HighIDPopSize,model)  ##create HighIDPop
        self.oneToOneExcitatoryConnector([HighIDinterPop,[interAddSize]], HighIDPop,self.minWt) #connect HighIDinterPop to HighIDPop from neuron 0 to interAddSize
                                                                                             #using weight self.minWt to activate them.
        ####################################connect HighIDinputPop (subset 3) to HighIDPop (subset 2)
        ExcConnectorIn3ToHiID=[]
        counter=interAddSize  ##the first neuron in subset 3 in HighIDinputPop
        delayBetIn3ToHiID=numLayers*self.timePerLayer  #delay between HighIDinputPop (subset3) and HighIDPop
        if(self.nest==1):
            delayBetIn3ToHiID-=1

        ##generate connections so that the neurons activated in HighIDinput (subset 3) activates the same neurons in HighIDPop (subset 2)
        for eachCell in range(firstCellHighIDInputSub3,firstCellHighIDInputSub3+HighIDInputSubSize3):
            ExcConnectorIn3ToHiID+=[(eachCell, counter, self.minWt, delayBetIn3ToHiID)]
            counter+=1

        ##connect
        self.connectPopsUsingExcitatory(HighIDinputPop, HighIDPop,ExcConnectorIn3ToHiID)  #connect HighIDinputPop to HighIDPop

        ###################################generate the connections between HighIDPop and AddPop
        firstCellAddSubset=firstCellHighIDInputSub1                    #Start cell in the ADDpop Subset
        firstCellHighIDSub1=0        ##the first neuron in the first subset of HighIDPop
        HighIDPopFirstSubSize=interAddSize  #size of the first subset in the HighIDPop 

        ##generate connections between HighIDPop (subsets 1 and 2) --> expand1 (if req) --->expand2 (if req) --> AddPop(subset 1) so that AddPop
        ##represents the total of the values in HighIDPop (subsets 1 and 2)
        connHiIDToAddAndNetInfo=self.generateConns2HighIDSubsTo1AddSub(HighIDPopFirstSubSize,HighIDInputSubSize3,firstCellHighIDSub1,firstCellAddSubset)
        netInfo=connHiIDToAddAndNetInfo[1]
        extraLayers=netInfo[4]

        numLayers+=extraLayers  ##the number of extra layers required, i.e. the number of expand pops created between HighIDPop and AddPop.

        #use the connections in connHiIDToAddAndNetInfo to connect HighIDPop and AddPop (and create and connect expand 1 and 2 pops if included in
        #connHiIDToAddAndNetInfo)
        self.connectHighIDPopToOutPop(HighIDPop, AddPop,connHiIDToAddAndNetInfo) 

        return numLayers
 

    def connectHighIDToExpandsToOutPop(self,HighIDPop,OutPop,connectionsAndNetInfo):
        ##This function uses the connections in connectionsAndNetInfo to connect the following populations:
        ##HighIDPop --> expand 1 pop--->expand 2 pop (if required)--->  OutPop, and creates an expand 1 population.
        
        ##An expand 2 population is inserted between expand 1 and the outPop if expand2PopRequired=1 in connectionsAndNetInfo (see below).

        ###HighID and expand populations are described at the beginning of the file.

        ExcConnector=connectionsAndNetInfo[0]  ##connections
        networkInfo=connectionsAndNetInfo[1] #information about the network associated to the connections

        expand1PopSize=networkInfo[0]         #the number of neurons required for expand1 
        
        expand2PopSize=networkInfo[1]         #the number of neurons required for expand2
        expand1PopRequired=networkInfo[2]       #is the expand1 pop required
        expand2PopRequired=networkInfo[3]       #is the expand2 pop required
        numConnectionsSets=len(ExcConnector[0])       #number of connection sets between HighIDPop and the expand1 population
        
        model=1 #use if_cond_exp neurons.
        expand1Pop=self.createPopulation(expand1PopSize,model) #create an expand 1 population

        ##connect HighIDPop to expand1Pop
        self.connectPopsUsingExcitatory(HighIDPop, expand1Pop,ExcConnector[0])
        if(expand2PopRequired==0):  ##an expand2 population is not required, connect the expand 1 population directly to the OutPop
            self.connectPopsUsingExcitatory(expand1Pop, OutPop,ExcConnector[1])

        if(expand2PopRequired==1):    #if an expand2 population is required
            expand2Pop=self.createPopulation(expand2PopSize,model)                  ##create one
            self.connectPopsUsingExcitatory(expand1Pop, expand2Pop,ExcConnector[2]) ##connect the expand1Pop to the expand2Pop
            self.connectPopsUsingExcitatory(expand2Pop, OutPop,ExcConnector[3])     ##connect the expand2Pop the OutPop

    def connectPopsUsingExciteAndInhib(self,InPop,OutPop,Connections):
        ##This connects the input population (InPop) to an output pop (OutPop) with the excitatory and inhibitory connections in Connections
        ExcConnections=Connections[0]
        InhConnections=Connections[1]        

        self.connectPopsUsingExcitatory(InPop,OutPop,ExcConnections)
        self.connectPopsUsingInhibitory(InPop,OutPop,InhConnections)
        
    def connectPopsUsingExcitatory(self,InPop,OutPop,ExcConnections):
        ##This connects the input population (InPop) to an output population (OutPop) with the excitatory connections in ExcConnections
        
        connList = self.sim.FromListConnector(ExcConnections)
        input_proj=self.sim.Projection(InPop, OutPop, connList,  receptor_type='excitatory')

    def connectPopsUsingInhibitory(self,InPop,OutPop,InhConnections):
        ##This connects the input population (InPop) to an output population (OutPop) with the inhibitory connections in InhConnections

        connList = self.sim.FromListConnector(InhConnections)
        input_proj=self.sim.Projection(InPop, OutPop, connList,  receptor_type='inhibitory')        

    def ModifyConnectionDelays(self,numNetworks,ConnectionsAndNetInfo,NumNeuralLayersInEachNetwork):
        ## This function modifies the delays in the given connections sets (ConnectionsAndNetInfo), where:
        ##   1. each connection set contains the connections for the following network:
        ##   
        ##      HighIDPop---->expand 1(if included)--->expand 2(if included)--->outPop

        ##   2. all of the connection sets given (i.e. networks (number of networks=numNetworks)) run in parallel in the RWIP system,
        ## so that all of the output pops (in each network) receive spikes at the same time.  For example, one network may have an expand 2 pop,
        ## and the other may only have an expand 1, thus, the output pops of both networks need to receive spikes (i.e. from expand 2 and 1)  at the same time.
        ##  
        ## The function modifies the delays in the connections between HighIDPop and expand 1, if included, or between HighIDPop and outPop.  The
        ## delay is calculated using the difference between the number of layers in each network, which are contained in
        ## NumNeuralLayersInEachNetwork.

        ## HighID and expand populations are described at the beginning of the file.

        ## The function returns the modified connections.
        
        if(NumNeuralLayersInEachNetwork!=[]): 
            maxNumLayers=(max(NumNeuralLayersInEachNetwork))  ##the highest number of layers in the given networks 

        for eachNetwork in range(0,numNetworks):   #for the connection set of each network

            connections=ConnectionsAndNetInfo[eachNetwork][0] ##extract the connections for the network
            NetInfo=ConnectionsAndNetInfo[eachNetwork][1]     #extract the network information
            expand1PopIncluded=NetInfo[2]                   
            
            if(NumNeuralLayersInEachNetwork!=[]): 
                NetDelay=(maxNumLayers-NumNeuralLayersInEachNetwork[eachNetwork])*self.timePerLayer  ##calculate the required delay for the current network
            else:
                NetDelay=0.0

            ExcConnectorHighIDToEx1=connections[0] #connections between expand 1 pop and OutPop
            ExcConnectorHighIDToOut=connections[4] #connections between HighIDPop and OutPop

            ###extract out the connection set to modify          
            if(expand1PopIncluded==1):   #modify the delay between HighIDPop and Expand 1 pop.
                connectionSet=ExcConnectorHighIDToEx1
            else:  #modify the delay between HighIDPop and OutPop.
                connectionSet=ExcConnectorHighIDToOut

            numConnections=len(connectionSet)
            for eachConn in range(0,numConnections): ####modify the delay in each of the connections 
                connectionSet[eachConn]=(connectionSet[eachConn][0],connectionSet[eachConn][1],connectionSet[eachConn][2],connectionSet[eachConn][3]+NetDelay)

            if(expand1PopIncluded==1):
                ConnectionsAndNetInfo[eachNetwork][0][0]=connectionSet
            else:
                ConnectionsAndNetInfo[eachNetwork][0][4]=connectionSet

        return ConnectionsAndNetInfo

    def createNetToAddAllHighIDSubs(self, HighIDInputPopSize, HighIDInputPop , HighIDReorderNumInSubs):
        ##This function creates a neural network so that the output population (which is a highIDPop and returned by the function) represents the
        ##total of the values given by each of the subsets (where the value is represented by one neuron in a subset) in the input population
        ##(HighIDInputPop).  In other words, when a specific neuron in each of the subsets fire, a specific neuron will fire in the output population.
        ##The size of HighIDInputPop is HighIDInputPopSize, and the size of each subset is in HighIDReorderNumInSubs.

        ##The adding process is implemented over a series of HighIDPop and addPop neural layers (neural layer sets):

        ## |########### neural layer set N-1 ########################|#############neural layer set N #########################################|
        ## HighIDPop(N-1) (subset X)
        ## HighIDPop(N-1) (subset X+1)------> addPop(N-1)(subset Y)----->HighIDPop(N)(subset Y)--------->
        ## HighIDPop(N-1) (subset X+2)                                                                    AddPop(N) (subset Z)
        ## HighIDPop(N-1) (subset X+3)------> addPop(N-1)(subset Y+1)--->HighIDPop(N)(subset Y+1)------->
    
        ##where in one neural layer set, the first and consecutive neurons in a subset of an addPop layer represents the total value of two consecutive subsets
        ##in the HighIDPop.  One neuron in each subset in HighIDPop represents the value of its corresponding subset in the previous neural layer's addPop.
        ##See the start of the file for more information about highIDPops.

        ##expand 1 and 2 populations (i.e. up to 2 neural layers) will be created and connected between every 2 subsets in HighIDPop and their
        ##corresponding subset in addPop if required.  Expand 1 and 2 populations are required if 1 or more neurons in the  addPop subset require >
        ##MaxConnectionsPerCell connections.  These will be added to the same neural layer set.  expand 1 and 2 are described at the start of the file

        ##if the number of subsets in a HighIDPop are odd, the function creates a network so that the last subset in the corresponding AddPop represents the
        ##total value of the last 3 subsets in HighIDPop.

        ##The function returns the number of neural layers created and the final HighIDPop (output population).

        #####################################
        currentHighIDPop=HighIDInputPop         ##This is the first HighIDPop in the first neural set
        currentHighIDNumInSubs=HighIDReorderNumInSubs  ##This contains the number of neurons in each subset of the first HighIDPop
        
        
        model=1                       #the neurons in this network are if_cond_exp neurons.
        AddPopSet=[]                  #contains the AddPop in each neural layer set
        HighIDPopSet=[]                 #contains the HighIDPop in each neural layer set
        StopAddingNeuralLaySets=0             #is 1 when no more HighIDPop and addPop neural layers are required
        NeuralLaySetNum=0               #number of neural layer sets
        TotalNumLayers=0.0           #number of neural layers in the neural network, i.e. the number the function creates 
        
        while(StopAddingNeuralLaySets!=1): #if another neural layer set is required

            TotalNumLayers+=2  #a neural set contains a minimum of 2 layers: a highID and add pop
            
            isNumHighIDSubetsOdd=int(np.ceil(float(len(currentHighIDNumInSubs)))%2.0)  #is the number of subsets in currentHighIDPop odd.
            if(isNumHighIDSubetsOdd==1):
                numAddSubs=int(float(len(currentHighIDNumInSubs)-1)/2.0)            ##number of 2 subset additions, +1 for adding the values of the last 3 subsets  
            else:
                numAddSubs=int(float(len(currentHighIDNumInSubs))/2.0)              ##number of 2 subset additions.

            NextHighIDNumInSubs=[]                                                  ##this contains the size of each subset in the HighIDPop in the next neural layer set.
            currentAddPop=self.createPopulation(HighIDInputPopSize,model)    ##create a new addPop for the current neural layer set
            AddPopSet+=[currentAddPop]                                       #add the new addPop

            firstCellInHighIDSub1=0                             ##the first neuron in the first of the two subsets in the current HighIDPop to be added.                                                               
            AllconnBetHi2SubAddAndNetInfo=[]          ##holds the connections between all of the populations (subsets) in the current neural layer set,
                                                      ##including expand 1 and 2 and set information.
                                     
            ExcConBetAddAndNextHighID=[] ##stores sets of excitatory connections, where each set contains the connections between one addpop subset in the
                                         ##current neural layer set and the corresponding 2 or 3 subsets in the HighIDPop of the next.
            InhConBetAddAndNextHighID=[] ##same as above but for inhibitory connections.
  
            ##################################GENERATE THE CONNECTIONS BETWEEN THE HIGH ID, EXPAND POPS(IF REQUIRED) AND ADD POPS SUBSETS IN THE CURRENT NEURAL
            ##################################LAYER SET.  IN ADDITION, GENERATE THE CONNECTIONS BETWEEN THE ADDPOP OF THE CURRENT NEURAL LAYER SET AND THE
            ##################################HIGHIDPOP OF THE NEXT NEURAL LAYER SET
            subset=0                 #subset ID in highIDPop
            ExtraLayersEachAddSub=[] #this keeps a record of the extra layers (i.e. expand 1 and expand2) between every 2 subsets in the current HighIDPop (or
                                     #the last 3 subsets if the number of subsets in HighIDPop is odd)  and
                                     #the corresponding subset in the AddPop. 
            for eachAddSub in range(0,numAddSubs):   ##for every addPop subset
                if((isNumHighIDSubetsOdd==1)and(eachAddSub==numAddSubs-1)):  
                    ##create and connect a network between the last 3 subsets of the current HighIDPop and the last subset of the current addPop (which may
                    ##also include expand 1 and 2 pops) so that the subset represents the total value of the last 3 subsets in HighIDPop.  
                    AddPopSubSize=currentHighIDNumInSubs[subset]+currentHighIDNumInSubs[subset+1]+currentHighIDNumInSubs[subset+2] ##size of the current
                                                                                                                                   ##AddPop subset
                    ##generate the connections.  Function returns the number of extra layers created between the highIDPop and AddPop subsets.
                    ExtraNumLayers=self.createNetToAdd3HighIDSubs(currentHighIDPop,firstCellInHighIDSub1,currentHighIDNumInSubs[subset],currentHighIDNumInSubs[subset+1],currentHighIDNumInSubs[subset+2],currentAddPop)                    
                    NextHighIDNumInSubs+=[AddPopSubSize]##size of the last highIDPop subset in the next neural layer set

                else:
                    ##generate the connections between the current 2 subsets of the current HighIDPop and their corresponding subset in the current addPop
                    ##(which may also include expand 1 and 2 connections) so that the subset represents the total value of the current 2 subsets in HighIDPop.
                    AddPopSubSize=currentHighIDNumInSubs[subset]+currentHighIDNumInSubs[subset+1] #size of the current AddPop subset
                    ##generate the connections.  Function returns the connections and associated network information.
                    connBetHighID2SubAddAndNetInfo=self.generateConns2HighIDSubsTo1AddSub(currentHighIDNumInSubs[subset],currentHighIDNumInSubs[subset+1],firstCellInHighIDSub1,[]) #[]==VALUE IS NOT REQUIRED                   
                    netInfo=connBetHighID2SubAddAndNetInfo[1]
                    ExtraNumLayers=netInfo[4] #number of extra layers created between the highIDPop and AddPop subsets.
                    AllconnBetHi2SubAddAndNetInfo+=[connBetHighID2SubAddAndNetInfo]
                    NextHighIDNumInSubs+=[AddPopSubSize] ##size of the highIDPop subset (associated to the current AddPop subset) in the next neural layer set
                    
                ExtraLayersEachAddSub+=[ExtraNumLayers]

                ##GENERATE THE CONNECTIONS BETWEEN THE CURRENT SUBSET IN THE ADDPOP OF THE CURRENT NEURAL LAYER SET AND ITS CORRESPONDING SUBSET IN
                ##THE HIGHIDPOP OF THE NEXT NEURAL LAYER SET
                firstCellInADDPopSub1=firstCellInHighIDSub1 
                connBetAddNxtHighID=self.HighIDPopConnections(AddPopSubSize, firstCellInADDPopSub1, 1.0)  ##1.0 is the connection delay
                ExcConBetAddAndNextHighID+=connBetAddNxtHighID[0] #excitatory connections
                InhConBetAddAndNextHighID+=connBetAddNxtHighID[1] #inhibitory connections

                ##THE NEXT TWO SUBSETS TO ADD IN THE HIGHIDPOP IN THE CURRENT NEURAL LAYER SET
                firstCellInHighIDSub1+=AddPopSubSize #the first cell in the first subset of the next 2 subsets in HighIDPop.
                subset+=2 ##id of the first subset in the next 2 subsets in highIDPop.
                
            #################################MODIFY THE CONNECTION DELAYS IN EACH CONNECTION SET (I.E. THE CONNECTIONS BETWEEN EVERY TWO SUBSETS
            #################################IN THE CURRENT HIGHID POP AND THE CORRESPONDING SUBSET IN THE ADD POP) SO THAT THE SPIKES FROM THE HIGHID
            #################################SUBSETS ACTIVATE THE ADDPOP SUBSETS AT THE SAME TIME.  THIS IS REQUIRED IF SOME SETS INCLUDE EXPAND 1 AND 2 POPS
            ################################# AND/OR IF A NETWORK WAS CREATED TO ADD THE LAST 3 SUBSETS OF THE CURRENT HIGHIDPOP (IMPLEMENTED IF THE NUMBER OF
            #################################SUBSTS IN THE CURRENT HIGHIDPOP IS ODD) AS THIS REQUIRES MORE LAYERS THAN JUST ADDING THE VALUE OF 2
            #################################HIGHID SUBSETS.

            numAddSubsToModify=numAddSubs
            if(isNumHighIDSubetsOdd==1): 
                numAddSubsToModify-=1   #we do not need to modify the connection delays of the network that adds the last 3 subsets of the highIDPop.
                
            ModAllconnBetHi2SubAddAndNetInfo=self.ModifyConnectionDelays(numAddSubsToModify,AllconnBetHi2SubAddAndNetInfo,ExtraLayersEachAddSub)
            TotalNumLayers+=max(ExtraLayersEachAddSub)     ##this is the maximum number of layers created between the current highIDPop and addPop.

            ###############################CONNECT THE HIGHIDPOP SUBSETS (EXPAND 1 AND 2 POPULATIONS IF REQUIRED) AND ADDPOP SUBSETS USING THE PREVIOUSLY
            ###############################GENERATED  CONNECTIONS.  IF THE LAST ADDPOP SUBSET REPRESENTS THE TOTAL OF THE LAST 3 HIGHIDPOP SUBSETS, WE DO NOT
            ###############################NEED TO CONNECT THIS NETWORK HERE AS IT WAS ALREADY DONE IN FUNCTION createNetToAdd3HighIDSubs
            for eachAddSub in range(0,numAddSubs): ## for each addPop subset
                if(((isNumHighIDSubetsOdd==1)and(eachAddSub==numAddSubs-1))==0):
                    self.connectHighIDPopToOutPop(currentHighIDPop, currentAddPop,ModAllconnBetHi2SubAddAndNetInfo[eachAddSub])
                    
            ###############################USE THE PREVIOUSLY GENERATED CONNECTIONS TO CONNECT THE ADDPOP SUBSETS IN THE CURRENT NEURAL LAYER, TO
            ###############################EITHER THEIR CORRESPONDING SUBSETS IN THE HIGHIDPOP IN THE NEXT NEURAL LAYER OR THE HIGHIDPOP THE FUNCTION RETURNS.

            NewHighIDPop=self.createPopulation(HighIDInputPopSize,model)   ##create a new highID pop for either the next neural layer set or to return
            ##connect the subsets
            self.connectPopsUsingExciteAndInhib(currentAddPop,NewHighIDPop,[ExcConBetAddAndNextHighID,InhConBetAddAndNextHighID])#            
            
            ##DETERMINE WHETHER TO ADD/CONNCT ANOTHER NEURAL LAYER SET
            if(numAddSubs==1): ##if the current addPop represents the total of all the HighIDPop subsets, i.e. the current addPop has only one subset, 
                StopAddingNeuralLaySets=1    ##stop adding neural layer sets to the network
            else:
                ##############ADD ANOTHER NEURAL LAYER SET
                NeuralLaySetNum+=1  #increment the number of neural layer sets.
                NextHighIDPop=NewHighIDPop
                HighIDPopSet+=[NextHighIDPop]               ##add the new HighIDPop
                currentHighIDPop=NextHighIDPop  ##the highIDPop for the next neural layer set
                currentHighIDNumInSubs=NextHighIDNumInSubs  #the number of neurons in each subset in the highIDPop of the next neural layer.

        OutPop=[NewHighIDPop,[0,0],0,[0,0]]  ##information about the NewHighIDPop is not required for the other functions that use it,  however,
                                             ##we need to add 3 values to the list in order for the other functions to work.

        return [OutPop,TotalNumLayers]

    def CreateReorderSubsNet(self,NumNeurons,FirstNeuronInFirstSubset,InputPop):
        ##This functions creates a neural network, where a reorder population is split into subsets of size self.MaxConnectionsPerCell and
        ##each subset represents a reordered version of the value represented by the same subset in
        ##the InputPop.  In other words, the first and consecutive neurons in each subset of the reorder population
        ##represent the value given by the same subset in the InputPop.  The function also connects a HighIDPop,which is the output of the network, to the
        ##ReorderPop, so that one neuron in each subset in HighIDPop represents the value in the same subset in ReorderPop.  (see the start of the file
        ##for more information about the HighID pop).

        ##In Summary, the function creates the following network:

        ##InputPop(subset X) ------->ReOrderPop(Subset X) ------->HighIDPop(Subset X)
        ##InputPop(subset X+1) ----->ReOrderPop(Subset X+1) ----->HighIDPop(Subset X+1)

        ##The InputPop is split into several subsets starting from neuron FirstNeuronInFirstSubset to neuron FirstNeuronInFirstSubset+NumNeurons.

        ##The function returns the HighIDPop and the number of neurons in each subset.
        
        model=1                                                       ##use if_cond_exp neurons for the below populations
        ReOrderPop=self.createPopulation(NumNeurons,model)              ##create reorder pop.
        HighIDPop=self.createPopulation(NumNeurons,model)               ##create HighIDPop.
        NumSubSets=int(float(NumNeurons)/float(self.MaxConnectionsPerCell))#number of neuron subsets of size self.MaxConnectionsPerCell
        NumRem=int(float(NumNeurons)%float(self.MaxConnectionsPerCell))  ##the remainder of neurons after the subset division.  
        FirstInSubsetInputPop=FirstNeuronInFirstSubset                  ##id of the first neuron in the first subset in InputPop,
        FirstInSubsetReAndHighPop=0                                     ##id of the first neuron in the first subset in ReOrderPop and HighIDPop.
        NumberInEachSubset=[]                                           ##is a list containing the number of neurons in each subset, where the last contains the
                                                                        ##remainder of neurons
        
        ExcConnectorRe=[]           #connections between InputPop and ReOrderPop
        ExcConnectorHigh=[]         #excitatory connections between ReOrderPop and HighIDPop
        InhConnectorHigh=[]         #inhibitory connections between ReOrderPop and HighIDPop

        ConnectionDelays=1.0        #spike delay between neurons in the ReOrderPop and HighIDPop

        ##generate the connections between each subset of neurons in InputPop to ReOrderPop, and ReorderPop to HighIDPop

        for eachSubset in range(0,NumSubSets):
            
            #generate the reorder connections between the neurons in subset (eachSubset) in InputPop (starting from neuron FirstInSubsetInputPop
            #to neuron FirstInSubsetInputPop+self.MaxConnectionsPerCell) and ReOrderPop (starting from neuron FirstInSubsetReAndHighPop to
            #FirstInSubsetReAndHighPop+self.MaxConnectionsPerCell)
            ExcConnectorRe+=self.generateReorderConnections(self.MaxConnectionsPerCell, FirstInSubsetInputPop, FirstInSubsetReAndHighPop) 
                                                                                                              
            #generate the reorder connections between the neurons in subset (eachSubset) in ReOrderPop and HighIDPop (both starting from
            #FirstInSubsetReAndHighPop to FirstInSubsetReAndHighPop+self.MaxConnectionsPerCell)
            Connectors=self.HighIDPopConnections(self.MaxConnectionsPerCell, FirstInSubsetReAndHighPop,ConnectionDelays) 
                                                                                                            
            ExcConnectorHigh+=Connectors[0] ##extract the excitatory connections between ReOrderPop and HighIDPop
            InhConnectorHigh+=Connectors[1] ##extract the inhibitory connections ReOrderPop and HighIDPop           
            FirstInSubsetInputPop+=self.MaxConnectionsPerCell #first neuron in the next subset in InputPop
            FirstInSubsetReAndHighPop+=self.MaxConnectionsPerCell #first neuron in the next subset in ReOrderPop and HighIDPop
            NumberInEachSubset+=[self.MaxConnectionsPerCell] #record the size of the subset

        ##connect the remainder of neurons in InputPop to ReOrderPop, and ReorderPop to HighIDPop.
        if(NumRem!=0): #if there are remaining neurons
            ExcConnectorRe+=self.generateReorderConnections(NumRem, FirstInSubsetInputPop, FirstInSubsetReAndHighPop) 
            Connectors=self.HighIDPopConnections(NumRem, FirstInSubsetReAndHighPop,ConnectionDelays)
            ExcConnectorHigh+=Connectors[0]
            InhConnectorHigh+=Connectors[1]            
            NumberInEachSubset+=[NumRem]  #add the remaining number of neurons to the list


        ##connect InputPop to ReOrderPop
        self.connectPopsUsingExcitatory(InputPop[0],ReOrderPop,ExcConnectorRe)

        ##connect ReOrderPop to HighIDPop
        self.connectPopsUsingExciteAndInhib(ReOrderPop,HighIDPop,[ExcConnectorHigh,InhConnectorHigh])

        return [HighIDPop, NumberInEachSubset]

    def ModifyNextThetaThetaDotConnDelays(self,RWIPNumLayersNextTh,RWIPNumLayersNextThDot,nextThetaDotPopAndConn,nextThetaPopAndConn):
        ##This function modifies the connection delays in one of the two given connection sets (nextThetaDotPopAndConn,nextThetaPopAndConn), so
        ##that both the RWIP model outputs (i.e. the next theta and theta dot values) are sent at the same time, i.e. nextThetaDotPop and nextThetaPop
        ##fire at the same time.
        
        ##nextThetaDotPopAndConn contains the connections (network) between the layer (population) representing the next thetaDot value and the previous layer
        ##(containing two separate populations, one representing positive, and the other representing negative, values). nextThetaPopAndConn contains the
        ##connections (network) between the layer (population) representing the next theta value and the previous layer
        ##(containing two separate populations, one representing positive, and the other representing negative, values).  As the next theta and thetadot
        ##values are calculated in parallel, but with a different number of layers, the delay of one connection set is changed so that nextThetaDotPop and
        ##nextThetaPop fire at the same time.  The delay is
        ##calculated using the difference between the number of neural layers used to calculate the next theta value (RWIPNumLayersNextTh) and the next
        ##thetaDot value (RWIPNumLayersNextThDot).

        ##The function returns the modified connections.

        #first two and last two of numLayersInEachNet contains the number of neural layers used to calculate the
        #next theta value and thetadot value respectively.
        numLayersInEachNet=[RWIPNumLayersNextTh,RWIPNumLayersNextTh,RWIPNumLayersNextThDot,RWIPNumLayersNextThDot]

        #contains the connections to modify, the first two represent the connections between nextThetaPop and the previous layer (populations) representing
        #negative and positive values respectively, and the last two is the same as above but for nextThetaDotPop
        connectionsToModify=[nextThetaPopAndConn[3],nextThetaPopAndConn[4],nextThetaDotPopAndConn[3],nextThetaDotPopAndConn[4]] 

        NumSetsOfConnections=len(connectionsToModify)

        ##change the delays
        modifiedConnections=self.ModifyConnectionDelays(NumSetsOfConnections,connectionsToModify,numLayersInEachNet)

        nextThetaPopAndConn[3]=modifiedConnections[0]
        nextThetaPopAndConn[4]=modifiedConnections[1]
        nextThetaDotPopAndConn[3]=modifiedConnections[2]
        nextThetaDotPopAndConn[4]=modifiedConnections[3]
                          
        return [nextThetaDotPopAndConn,nextThetaPopAndConn]
            

    def CreateReorderNetwork(self,popToReorder,OutputPop,connectFinalLayer):
        ##This function creates a reorder network, which is connected to the given input population (popToReorder), where its output population (
        ##which is either given in OutputPop, or one is created if OutputPop=[]) represents the reordered
        ##version of (i.e. the first and consecutive neurons in, depending on the sign of the value given by popToReorder, the set representing
        ##positive or negative values in the output population represent) the value given by popToReorder.  In other words, when a specific number of neurons
        ##in popToReorder fire, representing a positive or negative value, the same number in OutputPop fire but the neurons that fire are the first and
        ##consecutive neurons in the set representing positive or negative values.  This is required for the system to work correctly.

        ##The network is implemented using several neural networks.  Firstly, one neural network (which we call a sub reorder network) is connected to the
        ##neurons representing positive values in popToReorder, and another to the neurons representing negative values.  The output layer of each (which
        ##is a highIDPop and explained at the beginning of the file) is split into several subsets of size self.MaxConnectionsPerCell and each subset
        ##represents a reordered version of the value in their corresponding subset in popToReorder.  Another
        ##neural network (which we call the adding network) is connected to each of the sub reorder network outputs (i.e. the HighIDPop)
        ##, where the output of each adding network (which is also a highIDPop) represents the total value of all the subsets in the output population of
        ##their connected sub reorder network.  The adding networks contains (i.e. the adding process is implemented over) several neural layers, i.e. each
        ##subset in one layer represents the total value of 2 or 3 consecutive and different subsets in the previous layer.  Finally, the output populations
        ##of the adding networks are connected to a final neural layer (i.e. the given or created outputPop).  The output of the adding network
        ##associated to positive values is connected to the neurons in the final layer representing positive values, and the other network is connected to the
        ##neurons representing negative values.

        ##expand populations are used in the adding networks.  expand populations are described at the beginning of the file.

        ##if connectFinalLayer==1, the function will connect the OutputPop (or if OutputPop==[] the created one) to the outputs of the adding networks.

        ##The function returns the following: the number of neural layers created for the whole reorder network, the output population and its connections.

        ##Reordering is implemented in this way due to the neuron input limitation described at the beginning for the file and to prevent SpiNNaker overloading.

 
        numReorderedLayers=2 #number of layers in the whole reorder network.  The initial number is 2,  the HighIDPop and reorder pop created in
                             #function CreateReorderSubsNet for the sub reorder networks.

        [NeuronValue,totalPopNum,negNum,posNum]=self.extractPopInfo(popToReorder)   ##extract information about popToReorder

        ###############create a sub reorder network for, and connect it to, the neurons representing negative values in popToReorder.
        ###############negNum is the number of neurons in popToReorder representing negative values.  In addition, acquire the output
        ###############layer of the network (negReHighIDPop) and the number of neurons in each subset in negReHighIDPop (NumberInEachNegIDSubset)
        FirstNegNeuron=0 #the index of the first neuron representing a negative value in popToReorder
        [negReHighIDPop, NumberInEachNegIDSubset]=self.CreateReorderSubsNet(negNum,FirstNegNeuron,popToReorder)    

        ###############create a sub reorder network for, and connect it to, the neurons representing positive values in popToReorder. 
        ###############posNum is the number of neurons in popToReorder representing positive values.  In addition, acquire the output
        ###############layer of the network (posReHighIDPop) and the number of neurons in each subset in posReHighIDPop (NumberInEachPosIDSubset).
        FirstPosNeuron=negNum #the index of the first neuron representing a positive value in popToReorder
        [posReHighIDPop, NumberInEachPosIDSubset]=self.CreateReorderSubsNet(posNum,FirstPosNeuron,popToReorder) 


        ###create and connect an adding network to negReHighIDPop.  Acquire the output layer (NegAddHighIDPop) and the number of neural layers created/used
        ###in the adding network (NegNumLayers).
        [NegAddHighIDPop,NegNumLayers]=self.createNetToAddAllHighIDSubs(negNum, negReHighIDPop,NumberInEachNegIDSubset)

        ###create and connect an adding network to posPosHighIDPop.  Acquire the output layer (PosAddHighIDPop) and the number of neural layers created/used
        ###in the adding network (PosNumLayers).
        [PosAddHighIDPop,PosNumLayers]=self.createNetToAddAllHighIDSubs(posNum, posReHighIDPop,NumberInEachPosIDSubset) 

        ######################generate the connections between both negReHighIDPop and posPosHighIDPop and the final layer (either OutputPop or a new one if
        ######################OutputPop==[]).  connect them if connectFinalLayer==1.  The following parameters are required for the function that
        ######################generates the connections: self.genConnsNegPosHighIDToOutPops

        connsToProduce=3 ##is the option to generate the connections between a HighIDPop and an output population,
                         ##so the first and consecutive neurons in the set of neurons representing positive or negative values
                         ##in the output population represents the positive or negative value given by one of the neurons in NegAddHighIDPop or PosAddHighIDPop.

        ##information related to the neuron (values) in NegAddHighIDPop that will result in the OutputPop representing a negative value (result):
        firstCellAddHighIdNeg=0  #the first neuron in NegAddHighIDPop that will result in OutputPop representing a negative value
        firstAddHighIDValNegRes=NeuronValue*-1.0 #the value of the first neuron in NegAddHighIDPop that will result in OutputPop representing a negative
                                                 #value
        HighIdNeuronsResNegOut=[negNum,firstCellAddHighIdNeg,firstAddHighIDValNegRes]                        

        ##information related to the neuron (values) in PosAddHighIDPop that will result in the OutputPop representing a positive value (result):
        firstCellAddHighIdPos=0 #the first neuron in PosAddHighIDPop that will result in OutputPop representing a positive value
        firstAddHighIDValPosRes=NeuronValue #the value of the first neuron in PosAddHighIDPop that will result in OutputPop representing a positive value
        HighIdNeuronsResPosOut=[posNum,firstCellAddHighIdPos,firstAddHighIDValPosRes]

        Layers=[NegNumLayers,PosNumLayers]  ##number of layers created/used in each adding network.  This is required by genConnsNegPosHighIDToOutPops to
                                            ##calculate apporpriate delays between the outputs of the adding networks and the outputPop so any spikes sent
                                            ##from either adding network activate the outputPop at the same time.

        ##generate connections.      
        [OutputReorderedPopAndConnections, numLayers]=self.genConnsNegPosHighIDToOutPops(NegAddHighIDPop,PosAddHighIDPop,HighIdNeuronsResNegOut,HighIdNeuronsResPosOut,OutputPop,Layers,connectFinalLayer,connsToProduce)

        numReorderedLayers+=numLayers+1 ##number of layers used in the reorder network.  +1 for the output layer

        return [OutputReorderedPopAndConnections,numReorderedLayers]

    def createAdditionPopulation(self, Input1Num, Input2Num, InputNeuronValue):
        #This function creates an addition population (which will represent the result of the addition calculation) for the addition neural function
        #The function returns the population and information related to it.
        
        addPopSize=Input1Num[1][0]+Input2Num[1][0] #size of the addition population
        addNumNeg=Input1Num[1][1]+Input2Num[1][1] #number of neurons representing a negative value in the addition population
        addNumPos=addPopSize-addNumNeg            #number of neurons representing a positive value in the addition population.
        
        minSumValue=-1.0*addNumNeg*InputNeuronValue  #minimum value the addition pop can represent
        maxSumValue=addNumPos*InputNeuronValue       #maximum value the addition pop can represent

        model=1  #use neuron model if_cond_exp for

        addzero=0 #a neuron to represent a value of 0 is not required
        addPopAndInfo=self.createValuePop(InputNeuronValue, [minSumValue,maxSumValue],model,addzero,[]) #create the population


        return addPopAndInfo

    def generateAdditionConnections(self,input1Size,input1NegNum,Input1Delay,startAddNegNeuronToExc,firstInp2PosNeuron,firstInp2NegNeuron,startAddPosNeuronToExc,input2PosNum,input2NegNum):
        #This generates the connections between an input pop (referred to as input 1) and an addition (output) pop, using information about a second input
        #population.  The addition pop is used to represent the total of two values, each represented by an input population.  A positive or negative total value
        #is represented by the corresponding set of neurons (i.e. the neurons representing positive or negative values) in the add population.  In addition, 
        #the value given by the input pop (depending on the sign of the value) is represented by the first and consecutive neurons in the set of neurons
        #representing negative or positive values.

        #The connections are generated so that the same number of neurons representing a particular value in the input 1 population,
        #activates the same number of neurons in the (add) population (i.e. neurons representing negative values (0 to input1NegNum-1) in the input 1 population
        #have one to one excitatory connections with the neurons (startAddNegNeuronToExc to startAddNegNeuronToExc+input1NegNum) in the addition population.
        #Similarly, the neurons representing positive values (input1NegNum to input1Size) in the input 1 population have one to one excitatory connections with
        #the neurons(startAddPosNeuronToExc to startAddPosNeuronToExc+(input1Size-input1NegNum) in the addition population). Note, the input 2 population will
        #activate a different set of neurons to the input1 population.  Thus, the number of neurons activated in the output population will equal the number
        #of neurons activated in the input 1 and 2 populations. Connections are also generated so any activated neurons
        #in input 1 will inhibit the same number of neurons, which represent the negative of the value represented by input 1, in the output (addition)
        #population, i.e. inhibit the neurons that input 2 can activate.  Thus inhibitory connections are used as follows: the neurons representing negative
        #values (0 to input1NegNum-1) in the input 1 population have one to one inhibitory connections
        #with the neurons (firstInp2PosNeuron to firstInp2PosNeuron+input1NegNum) in the addition population representing positive values, where
        #firstInp2PosNeuron is the first neuron in the addition population that represents a positive value that the first neuron in input 2, which also
        #represents a positive value, can activate.  Similarly, the neurons representing positive values (input1NegNum to input1Size) in the input 1 population
        #have one to one inhibitory connections with the neurons (firstInp2NegNeuron to firstInp2NegNeuron+(input1Size-input1NegNum) in the addition population
        #representing negative values, where firstInp2NegNeuron is the first neuron in the addition population that represents a negative value that the first neuron
        #in input 2, which also represents a negative value, can activate.  Thus, if the two inputs represent differently signed values, the number of neurons
        #activated in the output population equals the difference between the number of neurons activated in inputs 1 and 2.

        #Input1Delay is the delay required between input 1 and the output(addition) population so that the spikes from input1 and input2 activate the
        #output (addition) population at the same time.

        #input2PosNum and input2NegNum are the number of neurons in the input 2 population that represent positive and negative numbers respectively.

        #This function returns the excitatory and inhibitory connections.
        
        EXConnections=[]    #contains the excitatory connections
        INHconnections=[]   #contains the inhibitory connections

        inp2PosCellCounter=0 #counts the number of neurons representing a positive value in the input2 population
        inp1PosCellCounter=0 #counts the number of neurons representing a positive value in the input1 population
        inp2NegCellCounter=0 #counts the number of neurons representing a negative value in the input2 population
        
        ##connect the input 1 (given) population to the output (addition) population using one to one excitatory and inhibitory connections.
        for inp1Cell in range(0,input1Size): #for each cell in the input 1 population
            
            if(input1NegNum>inp1Cell):  #if inp1Cell neuron represents a negative value.
                #connect the neuron inp1Cell to (using self.minWt to activate) the output neuron inp1Cell+startAddNegNeuronToExc that represents a negative value.
                EXConnections += [(inp1Cell,inp1Cell+startAddNegNeuronToExc,self.minWt,1.0+Input1Delay)] 
                if(input2PosNum>0): ##if input 2 has neurons representing positive values.
                    if(inp2PosCellCounter<input2PosNum): #for neuron inp2PosCellCounter in input 2 representing a positive value
                        #connect neuron inp1Cell (using an inhibitory connection) to output neuron inp2PosCellCounter+firstInp2PosNeuron that represents
                        #a positive value, which inp2PosCellCounter can activate.
                        INHconnections+= [(inp1Cell,inp2PosCellCounter+firstInp2PosNeuron,self.inhibWt,1.0+Input1Delay)]
                        inp2PosCellCounter+=1
            else:
                #if inp1Cell neuron represents a positive value, connect the neuron to (using self.minWt to activate) the
                #output neuron inp1PosCellCounter+startAddPosNeuronToExc representing a positive value.
                EXConnections += [(inp1Cell,inp1PosCellCounter+startAddPosNeuronToExc,self.minWt,1.0+Input1Delay)]
                if(input2NegNum>0): ##if input 2 has neurons representing negative values.
                    if(inp2NegCellCounter<input2NegNum): #for neuron inp2NegCellCounter in input 2 representing a negative value
                        #connect neuron inp1Cell (using an inhibitory connection) to the output neuron inp2NegCellCounter+firstInp2NegNeuron that
                        #represents a negative value, which inp2NegCellCounter can activate
                        INHconnections+= [(inp1Cell,inp2NegCellCounter+firstInp2NegNeuron,self.inhibWt,1.0+Input1Delay)]
                inp2NegCellCounter+=1
                inp1PosCellCounter+=1
        return [EXConnections,INHconnections]

    def CreateAdditionNetwork(self,Input1PopsAndInfo, Input2PopsAndInfo,Input1Delay, Input2Delay):
        #This function creates an addition neural network, where the output (addition) population represents the total value given by Input1Pop (i.e. Input1PopsAndInfo[0])
        #and Input2Pop (i.e. Input2PopsAndInfo[0]).  For example, if 3 neurons representing positive values in input populations 1 and 2 fire, 6 neurons
        #in the addition population representing positive values will fire. Similarly, if 3 neurons representing positive values in input 1 fire, and
        #3 neurons representing negative values in input 2 fire, no neurons will fire in the addition population.

        #Input1Delay and Input2Delay are the delays required
        #between Input1Pop and the addition population, and between Input2Pop and the addition population respectively, so the spikes from both input populations
        #activate and/or inhibit the addition population at the same time. The output (addition) population of the network is returned.

        #extract the infomation about input1 and 2 populations
        [Input1NeuronValue,input1Size,input1NegNum,input1PosNum]=self.extractPopInfo(Input1PopsAndInfo)
        [Input2NeuronValue,input2Size,input2NegNum,input2PosNum]=self.extractPopInfo(Input2PopsAndInfo)

        input1Pop=Input1PopsAndInfo[0]  #Input population 1
        input2Pop=Input2PopsAndInfo[0]  #Input population 2

        #check whether the Neuron Values of both input populations are the same.
        if(Input1NeuronValue!=Input2NeuronValue):
            print "Error: Input Neuron Values in the createAdditionNeuralFunction function do not match"
            print "input1"
            print Input1NeuronValue
            print "input2"
            print Input2NeuronValue
            adfadf

        ##create the addition (output) population
        addPopAndInfo=self.createAdditionPopulation(Input1PopsAndInfo, Input2PopsAndInfo, Input1NeuronValue)

        #extract the information about the newly created addition (output) population
        [addNeuronValue,addPopSize,addNegNum,addPosNum]=self.extractPopInfo(addPopAndInfo)

        
        addPop=addPopAndInfo[0]  ##add population

        #############################generate the connections between the input 1 population and the add (output) population
        firstAddNegNeuronToExc=0              #the first neuron in the output population representing a negative value that the first neuron in input1
                                              #representing a negative value can activate, and the first neuron represening a positive neuron in input 2
                                              #can inhibit
        firstAddPosNeuronToInh=addNegNum      #the first neuron in the output population representing a positive value that the first neuron in input 1
                                              #representing a negative value can inhibit, and the first neuron represening a positive value in input 2 can activate
        firstAddNegNeuronToInh=input1NegNum   #the first neuron in the output populaton representing a negative value that the first neuron in input 1 representing a
                                              #positive value can inhibit, and the first neuron representing a negative value in input 2 can activate.
        firstAddPosNeuronToExc=addNegNum+input2PosNum  #the first neuron in the output population representing a positive value that the first neuron in input
                                                       #1 representing a positive value can activate, and the first neuron representing a negative value in
                                                       #input 2 can inhibit.
        #generate connections
        [EXConnections,INHconnections]=self.generateAdditionConnections(input1Size,input1NegNum,Input1Delay,firstAddNegNeuronToExc,firstAddPosNeuronToInh,firstAddNegNeuronToInh,firstAddPosNeuronToExc,input2PosNum,input2NegNum)

        ##############################connect input 1 to the add (output) population
        self.connectPopsUsingExciteAndInhib(input1Pop,addPop,[EXConnections,INHconnections])

        ##############################generate the connections between the input 2 population and the add population
        firstAddNegNeuronToExc=input1NegNum #the first neuron in the output population representing a negative value that the first neuron in input2
                                              #representing a negative value can activate, and the first neuron representing a positive value in
                                              #input 1 can inhibit
        firstAddPosNeuronToInh=addNegNum+input2PosNum #the first neuron in the output population representing a positive value that the first neuron in input 2
                                              #representing a negative value can inhibit, and the first neuron representing a positive value in input 1 can
                                              #activate
        firstAddNegNeuronToInh=0              #the first neuron in the output population representing a negative value that the first neuron in input 2 representing
                                              #a positive value can inhibit, and the first neuron representing a negative value in input 1 can activate.
        firstAddPosNeuronToExc=addNegNum     #the first neuron in the output population representing a positive value that the first neuron in input 2
                                             #representing a positive value can activate, and the first neuron representing a negative value in input 1
                                             #can inhibit.
        #generate connections
        [EXConnections,INHconnections]=self.generateAdditionConnections(input2Size,input2NegNum,Input2Delay,firstAddNegNeuronToExc,firstAddPosNeuronToInh,firstAddNegNeuronToInh,firstAddPosNeuronToExc,input1PosNum,input1NegNum)    

        ###############################connect input 2 to the add (output) population
        self.connectPopsUsingExciteAndInhib(input2Pop,addPop,[EXConnections,INHconnections])

        return addPopAndInfo


    def createAdditionNeuralFunction(self, InputPops1, InputPops2, Input1Delay, Input2Delay, OutputAddPop, connectFinalLayer):
        ##creates an addition neural function (network), where its output population (which is either given in OutputAddPop
        ##or if OutputAddPop==[] the function creates a new one) represents the total value given by the two input
        ##populations (InputPops1 and InputPops2), i.e. a specific number of neurons will fire in the OutputAddPop, when a specific number of neurons
        ##fire in the input populations.  The positive or negative total is represented by the first and consecutive neurons in, depending on the
        ##sign of the value, the set that represent positive or negative values in the output population.
        
        #The addition neural function consists of two neural networks: an addition network, whose 
        ##output represents the total value of the two inputs, and a reorder network, whose output population (which is the OutputAddPop) represents a
        ##reordered version of the total value given by the addition network (as described above).
        ##A reorder network is required as the version of the value outputted by the addition network can be unordered.

        ##The output population of addition neural function is returned, along with its connections and the number of neural layers that comprise the
        ##neural function.

        #Input1Delay and Input2Delay are the delays required between Input1Pop and the addition population (see CreateAdditionNetwork) and between Input2Pop
        #and the addition population respectively, so the spikes from both input populations activate the addition population at the same time.

        ##connectFinalLayer ==1 to connect the final layer of the AdditionNeuralFunction, i.e. the output (final layer) of the reorder network.

        numAddLayers=1  #This is the number of neural layers that comprise the neural function.  The initial number is 1 for the addition population.

        ##create the addition network, and acquire the output of the network, i.e. the addition population (addPop)
        addPop=self.CreateAdditionNetwork(InputPops1, InputPops2, Input1Delay, Input2Delay)

        ##if no outputAddPop, pass the required specification of the population to CreateReorderNetwork so one is created.
        if(OutputAddPop==[]): 
            OutputAddPop=[[],addPop[1],addPop[2],addPop[3]]

        ##create the reorder network and connect it to addPop.  Acquire the following: the outputPop of the reorder network, its connections, and the number
        ##of neural layers created/used in the reorder network.
        [OutputPopAndConns,numReorderedLayers]=self.CreateReorderNetwork(addPop,OutputAddPop,connectFinalLayer)

        numAddLayers+=numReorderedLayers

        return [OutputPopAndConns,numAddLayers]

    def CreateAndConnectAHighIDPop(self, inputPop,delay):
        ##This function creates and connects a HighIDPop to the given inputPop, so that the value represented by the
        ##inputPop (which is represented by the first and consecutive neurons in, depending on the sign of the value,
        #the set that represents positive or negative values in inputPop) is represented by one neuron in the highID population, i.e. the neuron in the
        #highIdPop with the same ID as the one with the highest ID (last neuron) in the input population that contributed to the value.  The given delay is
        #added to the connections.

        #The function returns the HighIDPop and information about the population.

        model=1  #use if_cond_exp
        addzero=0 #a neuron to represent a 0 value is not required.
        HighIDPopAndInfo=self.createValuePop(inputPop[2], inputPop[3],model,addzero,[])
        
        InputPopSize=inputPop[1][0] #number of neurons in InputPop
        FirstNeuron=0 #the first neurons in inputPop to start the connections from

        #generate inhibitory and excitatory connections between all of the neurons in InputPop and HighIDPop
        ExcAndInhConns=self.HighIDPopConnections(InputPopSize, FirstNeuron, 1.0+delay) 

        ##connect the inputPop to the highIDPop using the ExcAndInhConns
        HighIDPop=HighIDPopAndInfo[0]
        self.connectPopsUsingExciteAndInhib(inputPop[0],HighIDPop,ExcAndInhConns)

        return HighIDPopAndInfo

    def HighIDPopConnections(self, SetSize, StartCell, delay):
        ##This function generates and returns the connections (with the given delay) between neurons startCell and StartCell+SetSize in an input population
        ##and the same neurons in a highID population, so that the value represented by the first and consecutive neurons (i.e. between startCell and
        ##StartCell+SetSize) in the input population is represented by one neuron in the highID population, i.e. the neuron in the highIDPop with the same
        ##ID as the one with the highest ID (last neuron) in the input population that contributed to the value.

        ##The function returns the connections.
        
        ExcConnector=[] #contains the excitatory connections
        InhConnector=[] #contains the inhibitory connections


        for eachCell in range(StartCell,SetSize+StartCell): #for each cell in an input population between startCell to setSize+StartCell 
            ExcConnector+=[(eachCell, eachCell, self.minWt+self.additionalWt, delay)]  ##connect eachcell in the inputPop to (using self.minWt to activate) the
                                                                                    ##same cell in the HighIDPop. 
                                                                                    ##add self.additionalWt to the weight due to the limitations of SPINNAKER,
                                                                                    ##to force the cell in the HighIDPop to fire.
            
            ##eachCell in InputPop inhibits all of the cells in the HighIDPop below it, i.e between eachCell-1 and StartCell
            if(eachCell>StartCell):
                if((eachCell-1)>StartCell):
                    for belowCell in range(eachCell-1,StartCell-1,-1): 

                        InhConnector+=[(eachCell, belowCell, self.inhibWt, delay)]
                else:
                    InhConnector+=[(eachCell, eachCell-1, self.inhibWt, delay)]
        
        return [ExcConnector,InhConnector] 

    def ConnectForceToHighIDPop(self,forcePop,HighIDPop, delay):
        #This function connects the given forcePop to the given HighIDPop, so that the value represented by forcePop (i.e. by the first and consecutive
        #neurons in, depending on the sign of the value, the set representing positive or negative values in forcePop, or one neuron when the value is 0) is
        #represented by one neuron in the highID population, i.e. the neuron in the highIdPop with the same ID as the one with the highest ID (last neuron) in
        #the forcePop that contributed to the value.  

        ExcConnector=[] #contains the excitatory connections
        InhConnector=[] #contains the inhibitory connections

        ##generate the connections
        ForcePopSize=forcePop[1][0] #extract the force population
        for eachCell in range(0,ForcePopSize): #for eachCell in the force Pop
            ExcConnector+=[(eachCell, eachCell, (self.minWt)+0.1, delay)]  ##connect eachcell in the forcePop to (using self.minWt+0.1 to activate) the
                                                                        ##same cell in the HighIDPop.
            ##eachCell in forcePop inhibits all of the cells in the HighIDPop below it.
            if((eachCell>0)and(eachCell!=forcePop[1][1])): ## prevent the neurons representing positive values in forcePop inhibiting the neurons in highIDPop
                                                           ## representing negative values and 0.
                InhConnector+=[(eachCell, eachCell-1, self.inhibWt+0.05, delay)]

        #connect the populations
        self.connectPopsUsingExciteAndInhib(forcePop[0],HighIDPop[0],[ExcConnector,InhConnector])   

    def ControlForceBufferOutput(self):
        ##This function creates a neural network to control when the spikes from self.forceBufferPop (i.e. the force value) are released into the system
        ##(i.e. activate the corresponding neuron in self.HighIDForcePop, which is the input population/layer to the FFrBDtNeuralFunction ).  In other words,
        ##the network prevents any spikes from self.forceBufferPop, which receives them from (i.e. is activated by) self.forcePop anytime between t (when
        ##self.forceBufferPop is enabled) and t+self.ForceBufferDuration, being released into the system (i.e. activating a neuron in self.HighIDForcePop)
        ##until t+self.ForceBufferDuration+1.0, and then deactivates/inhibits self.forceBufferPop.  self.HighIDForcePop is a highId population and represents
        ##the value from self.forceBufferPop/self.forcePop using one neuron.

        ##The network works as follows:
        ## self.inhibForceBufCell is activated when self.RWIPModActForceCell activates.  see self.RWIPModActForceCell at the beginning of the file for more
        ## information.  self.inhibForceBufCell fires every
        ## 1.0ms in SpiNNaker (i.e. 1 neural time step) and 2.0ms in NEST (i.e. 2 neural time steps) from
        ## t to t+self.ForceBufferDuration and inhibits all of the neurons in self.HighIDForcePop at these times, i.e. prevents any neurons in
        ## self.forceBufferPop activating any neuron in self.HighIDForcePop during this period.  self.inhibForceBufCell is deactivated by neuron 
        ## self.RWIPModActForceCell at t+self.ForceBufferDuration in SpiNNaker and t+self.ForceBufferDuration-1.0 in NEST, allowing
        ## neurons in self.forceBufferPop to activate a neuron in self.HighIDForcePop. self.RWIPModActForceCell deactivates the
        ## (i.e. inhibits all of the neuron in) self.forceBufferPop at t+self.ForceBufferDuration+1.0

        ##creates one self.inhibForceBufCell neuron
        model=2  ##create neurons using the if_cond_exp model with no refractory period.
        self.inhibForceBufCell=self.createPopulation(1,model) 

        ##self connect self.inhibForceBufCell (using self.minWt) so, once activated, it continously fires every 1.0 in SpiNNaker or 2.0ms in NEST (i.e. 1 or 2 neural time steps)
        ExcConnectorIToI=[(0, 0, self.minWt+0.01, 1.0)] 
        self.connectPopsUsingExcitatory(self.inhibForceBufCell, self.inhibForceBufCell,ExcConnectorIToI)

        #connect self.RWIPModActForceCell to (using self.minWt to activate) the self.inhibForceBufCell 
        ExcConnectorSToI=[(0, 0, self.minWt+0.01, 1.0)]                                                                     
        self.connectPopsUsingExcitatory(self.RWIPModActForceCell, self.inhibForceBufCell,ExcConnectorSToI) #connect the neurons        

        #connect self.inhibForceBufCell (using inhibitory connections) to all of the neurons in self.HighIDForcePop.
        delay=1.0
        neuronID=0  #neuron Id of self.inhibForceBufCell
        self.oneToAllInhibitoryConnector(self.inhibForceBufCell,neuronID,self.HighIDForcePop, self.minWt+0.01, delay)
        
        #connect (using inhibitory connections) self.RWIPModActForceCell to self.inhibForceBufCell so self.RWIPModActForceCell deactivates
        #self.inhibForceBufCell at t+self.ForceBufferDuration for SpiNNaker and t+self.ForceBufferDuration-1.0 in NEST
        if(self.nest==0):
            InhConnectorSToI=[(0, 0, self.minWt+0.01, self.ForceBufferDuration)] 
        else:
            InhConnectorSToI=[(0, 0, self.minWt+0.01, self.ForceBufferDuration-1.0)] 

        self.connectPopsUsingInhibitory(self.RWIPModActForceCell, self.inhibForceBufCell,InhConnectorSToI) 

        ##connect (using inhibitory connections) self.RWIPModActForceCell to all of the neurons in self.forceBufferPop so self.RWIPModActForceCell
        ##deactivates self.forceBufferPop at t+self.ForceBufferDuration+1.0

        delay=self.ForceBufferDuration+1.0
        self.oneToAllInhibitoryConnector(self.RWIPModActForceCell,0,self.forceBufferPop, self.minWt+0.01, delay)
        
    def CreateForceBufferNet(self):
        #This function creates a force buffer network, which includes self.forceBufferPop.   The network and self.RWIPModActForceCell enables
        #self.forceBufferPop at time t to receive the force value from self.forcePop, i.e. wait for spikes (the force value) from self.forcePop for duration
        #self.ForceBufferDuration ms, i.e. between t and t+self.ForceBufferDuration, and hold them once received (thus, the maximum time
        #they are held (i.e. how long self.forceBufferPop fires/is activated) for is self.ForceBufferDuration, depending on when they are received.).  The neurons in
        #self.forcePop are spike injectors in SpiNNaker and if_cond_exp neurons in NEST. 
        
        #The network works as follows:
        # self.forceBufferPop is self connected, so when its neurons fire/activate, they continously fire every 1ms in SpiNNaker (i.e. every
        # time step) and 2ms in NEST (i.e. every 2 time steps, due to NEST's neural simulation method).
        # When neurons in self.forcePop fire, they partially activate the same neurons in self.forceBufferPop.  The neurons in
        # self.forceBufferPop (i.e. the same neurons in forcePop that fired) are fully activated/fire when the neuron self.RWIPModActForceCell
        # activates.  However, as it is unknown when self.forceBufferPop will receive spikes from self.forcePop
        # (i.e. the spike injectors in SpiNNaker) (as discussed earlier), and, so self.forceBufferPop can receive spikes (i.e. be activated) at any time between t+1 to
        # t+self.ForceBufferDuration, self.forceBufferPop is continously partially activated every 1ms in SpiNNaker (1 neural time step) and 2ms in
        # NEST (2 neural time steps) for duration t+1 to t+self.ForceBufferDuration by a population of
        # self.ForceBufferDuration-1 sequentially connected neurons called bufferTrigcells, which are also activated by RWIPModActForceCells.  Thus,
        # neurons in self.forcePop can fully activate the same neurons in self.forceBufferPop at any point within this period.
        # Each neuron in bufferTrigcells is connected to all of the neurons in self.forceBufferPop, where the first neuron is activated by neuron
        # self.RWIPModActForceCell, and each neuron fires 1ms in SpiNNaker (1 neural time step) and 2ms in NEST (2 neural time steps) after the previous
        # neuron in the sequence.

        #########################################create the self.forcebuffer                                        
        forceNeuronValue=self.forcePop[2]  ##the value of each neuron in self.forcePop
        forceMinMaxVal=self.forcePop[3]  #the minimum and maximum values self.forcePop can represent.
        model=2               #create if_cond_exp neurons with no refractory period
        addzero=1            #add a neuron to represent a value of 0 in the population.
        self.forceBufferPop=self.createValuePop(forceNeuronValue, forceMinMaxVal, model,addzero,[]) ## create the neural force buffer.
        
        ###############self connect self.forceBufferPop (using self.minWt) so, once activated, it continously fires every 1.0ms in SpiNNaker (1 neural time step) and 2.0ms
        ###############in NEST (2 neural time steps).
        
        self.oneToOneExcitatoryConnector(self.forceBufferPop,self.forceBufferPop,self.minWt)  

        ###########connect self.forcePop to self.forceBufferPop (using self.minWt*0.9) so self.forcePop partially activates the neurons in self.forceBufferPop
        self.oneToOneExcitatoryConnector(self.forcePop, self.forceBufferPop,self.minWt*0.9) 

        ############connect self.RWIPModActForceCell to self.forceBufferPop (using self.minWt*0.1) so self.RWIPModActForceCell partially
        ############activates the neurons in self.forceBufferPop 
        delay=1.0
        self.oneToAllExcitatoryConnector(self.RWIPModActForceCell,0,self.forceBufferPop, self.minWt*0.1, delay)

        ##############################create the population bufferTrigcells
        model=1 #create if_cond_exp neurons with refractory period
        bufferTrigcells=self.createPopulation(int(self.ForceBufferDuration)-1,model)

        ############################# connect the neurons in bufferTrigcells sequentially using self.minWt, so that one fires 1ms in
        #############################SpiNNaker (1 neural time step) or 2ms in NEST (i.e. 2 neural time steps) after the previous in the sequence.
        delay=1.0
        ExcConnectorTrToTr=[] #excitatory connections between bufferTrigcells
        for eachCell in range(0,int(self.ForceBufferDuration)-2):  #for each cell in bufferTrigcells apart from the last 
            if(eachCell<int(self.ForceBufferDuration)-1):
                ExcConnectorTrToTr+=[(eachCell, eachCell+1, self.minWt, delay)] ##connect eachCell to the next using self.minWt
                                    
        self.connectPopsUsingExcitatory(bufferTrigcells,bufferTrigcells,ExcConnectorTrToTr)

        ##################connect self.RWIPModActForceCell to (using self.minWt to activate) the first neuron in bufferTrigcells     
        ExcConnectorStoTr=[(0, 0, self.minWt, delay)]  
        self.connectPopsUsingExcitatory(self.RWIPModActForceCell,bufferTrigcells,ExcConnectorStoTr)        

        ##################connect each neuron in bufferTrigcell to all of the neurons in self.forceBufferPop (using self.minWt*0.1) so each neuron
        ##################in bufferTrigcell partially activates all of the neurons in self.forceBufferPop 
        delay=1.0
        for eachCell in range(0,int(self.ForceBufferDuration)-1):        ##for each neuron in bufferTrigcells       
            self.oneToAllExcitatoryConnector(bufferTrigcells,eachCell,self.forceBufferPop, self.minWt*0.1, delay)

    def ConnectForceInputToFFrBDtNeuralFunction(self):
        ##This function connects self.forceBufferPop to self.HighIDForcePop and connects self.initForcePop to self.HighIDForcePop if self.ModelID==0.
        ##self.HighIDForcePop is the first (input) layer of the neural function that calculates: (Force-friction)*B*dt), which is created by function
        ##CreateFFrBDtNeuralFunction.  See the start of the file for more information about self.forceBufferPop.

        ##self.HighIDForcePop represents the force value in self.forceBufferPop and self.initForcePop using one neuron.
        
        ##connect the self.forceBufferPop to self.HighIDForcePop.
        delay=1.0       
        self.ConnectForceToHighIDPop(self.forceBufferPop, self.HighIDForcePop,delay)
                                                          
        if(self.ModelID==0):  ##self.initForcePop is created to connect to self.HighIDForcePop of the first RWIP model (i.e. self.ModelID==0).
                                ##The delay is set so the spikes (input) from initForcePop reach self.HighIDForcePop at the same time as the 
                                ## other RWIP model inputs (i.e. the spikes (inputs) from self.thetaPop and self.thetaDotPop) reach the first layer
                                # of their corresponding neural functions implemented at the same time.     
            self.ConnectForceToHighIDPop(self.initForcePop, self.HighIDForcePop,2.0+self.RWIPOutputDelay)

    def createPopAndInfoSet(self,PopSize,NumNeg,NeuronValue,minVal,maxVal,model,port):
        #This creates a population of neurons (spike injectors if model==3 (port is the port which the spiNNaker machine should listen for packets) and
        #if_cond_exp neurons if model==1 or 2) of Size (PopSize).  The population is returned in a list along with the following information: 
        #PopSize, NumNeg (which is the number of neurons in the population representing negative values), NeuronValue (which is the value each neuron
        #represents), minVal and maxval (which are the minimum and maximum values the population can represent).

        if(model==3): ##create a population of spike injectors
            Pop=self.createInjPopulation(PopSize,port)
        else: ##create a population of if_cond_exp neurons
            Pop=self.createPopulation(PopSize,model)
        
        PopAndInfo=[Pop,[PopSize,NumNeg],NeuronValue,[minVal,maxVal]]

        return PopAndInfo

    def createFFrBDtPopulation(self, FFrBdtNeuronValue):
        #This function creates a population of neurons (referred to as FFrBdtPop) to represent the result of, i.e. is the last layer (output population) of the
        #neural function that calculates, the following part of the numerical solution of the RWIP equation of motion: (force-friction)*B*dt.  The full
        #solution is shown at the top of the file.  The size of FFrBdtPop is based on both the given value each neuron in FFrBdtPop needs to represent
        #(i.e. FFrBdtNeuronValue) and the following information regarding the population (i.e.self.HighIDForcePop) that represents the force value in the
        #neural function: the value each neuron represents, and the number of neurons representing positive and negative values.  The function returns
        #FFrBdtPop along with the following information about it, i.e. the population size, the number of neurons in FFrBdtPop representing negative values,
        #FFrBdtNeuronValue, and the minimium and maximum value FFrBdtPop can represent.
        
        ##extract information about the population self.HighIDForcePop 
        [forceNeuronValue,ForcePopSize,ForceNegNum,ForcePosNum]=self.extractPopInfo(self.HighIDForcePop)

        minFFrBdtVal=self.BDt*((-1.0*ForceNegNum*forceNeuronValue)-self.frictVal)   ##minimum value FFrBdtPop can represent

        maxFFrBdtVal=self.BDt*((ForcePosNum*forceNeuronValue)-self.frictVal)        ##maximum value FFrBdtPop can represent   
         
        ##create population FFrBdtPop And Info Set
        model=1    #if_cond_exp neurons
        addzero=0  #a neuron to represent 0 is not required.
        FFrBdtPopAndInfo=self.createValuePop(FFrBdtNeuronValue, [minFFrBdtVal,maxFFrBdtVal],model,addzero,[])

        return FFrBdtPopAndInfo

    def CreateFFrBDtNeuralFunction(self):
        ##This creates a neural function to calculate/output the result of the following calculation: (force-friction)*B*dt.

        ##The neural function consists of a minimum of two layers, the first (input) layer is self.HighIDForcePop, where one neuron represents the force
        ##value given by self.forceBufferPop or self.initForcePop, and a second (output) layer FFrBDtPop represents the result of the above calculation, i.e.
        ##when a specific number of neurons in self.forceBufferPop or self.initForcePop fire, a specific neuron in self.HighIDForcePop will fire causing
        ##a specific number of neurons in FFrBDtPop to fire.  More specifically, the first and consecutive neurons in, depending on the sign of the result,
        #the set representing positive or negative values in FFrBDtPop will represent the result.
        ##HighID Pops are described at the beginning of the file.

        ##up to 2 expand populations may be added in between self.HighIDForcePop and FFrBDtPop.  Expand 1 and 2 populations are required if 1 or more
        ##neurons in the FFrBDtPop require > self.MaxConnectionsPerCell connections.  See the start of the file for more information.

        ##The function returns FFrBDtPop and information related to it, i.e. the population size, the number of neurons representing negative values,
        ##the value each neuron represents, and the minimium and maximum value FFrBdtPop can represent. The function also returns the number of layers
        ##created for the neural function.
        
        numberOfLayers=2  #this stores the number of neural layers created for the neural function.  The minimum number is 2, i.e. self.HighIDForcePop
                          #and FFrBDtPop

        self.ConnectForceInputToFFrBDtNeuralFunction()   ##connects the force input (either self.initForcePop and/or self.forceBufferPop) to the first layer
                                                         ##of the neural function, i.e. self.HighIDforcePop.

        ###########create FFrBDtPop
        FFrBdtNeuronValue=self.thetaDotPop[2]  ###the value each neuron in FFrBDtPop needs to represent.  The value needs to be same as
                                               ###the value each neuron in self.thetaDotPop represents.

        FFrBdtPopAndInfo=self.createFFrBDtPopulation(FFrBdtNeuronValue)

        ############connect self.HighIDForcePop to FFrBDtPop so that FFrBDtPop represents the result of the above calculation when a force value is given by
        ############self.HighIDForcePop.  The following parameters are required for the function self.genConnsNegPosHighIDToOutPops, which
        ###########generates the connections and connects the populations: 

        connsToProduce=1 ##is the option to generate the connections between self.HighIDForcePop and FFrBDtPop so that the above calculation is implemented.
        connect=1  ##connect the layers (populations)

        #####extract information about self.HighIDForcePop
        [forceNeuronValue,ForcePopSize,ForceNegNum,ForcePosNum]=self.extractPopInfo(self.HighIDForcePop) 
        ForcePosNum-=1   #as ForcePosNum contains both the number of neurons representing positive values and the neuron representing a 0 value,
                         #-1 is taken off so ForcePosNum contains only the number of neurons that represent positive values. 

        #####extract information about FFrBDtPop
        [FFrBdtNeuronValue,FFrBdtPopSize,FFrBdtPopNumNeg,FFrBdtPopNumPos]=self.extractPopInfo(FFrBdtPopAndInfo)

        ##information related to the neurons in self.HighIDForcePop (i.e. the values it can represent) that (if they fire) will result in the FFrBDtPop representing a
        ##negative value (result):
        NumExtraNeg=int(self.frictVal/forceNeuronValue)##number of neurons in self.HighIDForcePop representing a positive value that will
                                                       ##result in FFrBDtPop representing a negative value                                                      
        ForceNegNumTotal=ForceNegNum+NumExtraNeg+1 #number of neurons in self.HighIDForcePop (including positive neurons and a +1 for a 0 value) that
                                                   # will result in FFrBDtPop representing a negative value
        firstCellForceNeg=0                        #the first neuron in self.HighIDForcePop that will result in FFrBDtPop representing a negative value
        firstForceValNegRes=forceNeuronValue*-1.0  #the value represented by the first neuron in self.HighIDForcePop that will result in FFrBDtPop representing a
                                                   #negative value.
        ForceNeuronsResNegFFrBdt=[ForceNegNumTotal,firstCellForceNeg,firstForceValNegRes]                        

        ##information related to the neurons in self.HighIDForcePop (i.e. the values it can represent) that (if they fire) will result in the FFrBDtPop representing a
        ##positive value (result):
        ForcePosNumTotal=ForcePosNum-NumExtraNeg        ##number of neurons in self.HighIDForcePop that will result in FFrBDtPop representing a positive value.
        firstCellForcePos=ForceNegNum+NumExtraNeg+1     #the first neuron in self.HighIDForcePop that will result in FFrBDtPop representing a positive value         
        firstForceValPosRes=forceNeuronValue*(NumExtraNeg+1) #the value represented by the first neuron in self.HighIDForcePop that will result in FFrBDtPop representing 
                                                           #a positive value
        ForceNeuronsResPosFFrBdt=[ForcePosNumTotal,firstCellForcePos,firstForceValPosRes]

        ##generate the connections between self.HighIDForcePop to FFrBDtPop and connect them. The function returns the number of extra layers created, i.e.
        ##whether expand 1 and 2 populations (layers) were created between self.HighIDForcePop to FFrBDtPop.  The function also returns the FFrBDtPop,
        ##HighIDForcePop and connections, which are redundant as the network has already been connected.
        [PopAndConns,extralayers]=self.genConnsNegPosHighIDToOutPops(self.HighIDForcePop,self.HighIDForcePop,ForceNeuronsResNegFFrBdt,ForceNeuronsResPosFFrBdt,FFrBdtPopAndInfo,[],connect,connsToProduce)

        numberOfLayers+=extralayers

        return [FFrBdtPopAndInfo,numberOfLayers]

    def createThetaDotDtPopulation(self,HighIDThetaDotPopAndInfo,ThDotDtNeuronValue):
        #This function creates a population of neurons (referred to as ThetaDotDtPop) to represent the result of, i.e. is the last layer (output population) of the
        #neural function that calculates, the following part of the numerical solution of the RWIP equation of motion: thetaDot*dt.  The full solution is shown at
        #the top of the file.  The size of ThetaDotDtPop is based on both the given value each neuron in ThetaDotDtPop needs to represent (i.e. ThDotDtNeuronValue)
        #and the following information regarding the population (i.e. HighIDThetaDotPop in HighIDThetaDotPopAndInfo) that represents the thetaDot value in the
        #neural function: the value each neuron represents, and the number of neurons representing positive and negative values.  The function returns
        #ThetaDotDtPop along with the following information about it, i.e. the population size, the number of neurons in ThetaDotDtPop representing negative
        #values, ThDotDtNeuronValue, and the minimium and maximum value ThetaDotDtPop can represent.
        
        ##extract information about the population HighIDThetaDotPop
        [ThetaDotNeuronValue,ThetadotPopSize,ThetaDotNegNum,ThetaDotPosNum]=self.extractPopInfo(HighIDThetaDotPopAndInfo)

        minThetaDotDtVal=self.dt*(-1.0*ThetaDotNegNum*ThetaDotNeuronValue)   ##minimum value ThetaDotDtPop can represent

        maxThetaDotDtVal=self.dt*(ThetaDotPosNum*ThetaDotNeuronValue)   ##maximum value ThetaDotDtPop can represent

        ##create ThetaDotDtPop population And Info Set
        model=1   ##if_cond_exp neurons
        addzero=0 #a neuron to represent 0 is not required.
        ThetaDotDtPopAndInfo=self.createValuePop(ThDotDtNeuronValue, [minThetaDotDtVal,maxThetaDotDtVal],model,addzero,[])

        return ThetaDotDtPopAndInfo

    def extractPopInfo(self, PopAndInfo):
        #This function extracts and returns the information about the population in PopAndInfo
        
        NeuronValue=PopAndInfo[2]  #the value each neuron represents
        PopSize=PopAndInfo[1][0]  ##number of neurons in the population
        NegNum=PopAndInfo[1][1] ##number of neurons representing negative values
        PosNum=PopSize-NegNum  ##number of neurons representing positive values

        return [NeuronValue,PopSize,NegNum,PosNum]
    
    def CreateThetaDotDtNeuralFunction(self):
        ##This creates a neural function to calculate/output the result of the following calculation: thetaDot*dt.

        ##The neural function consists of a minimum of two layers, the first (input) layer is HighIDThetaDotPop, where one neuron represents the thetaDot
        ##value given by self.thetaDotPop, and a second (output) layer ThetaDotDtPop represents the result of the above calculation, i.e. when a specific number
        ##of neurons in self.thetaDotPop fire, a specific neuron in HighIDThetaDotPop will fire causing a specific number of neurons in ThetaDotDtPop to fire.
        ##More specifically, the first and consecutive neurons in, depending on the sign of the result, the set representing positive or negative values
        ##in ThetaDotDtPop will represent the result.  HighID Pops are described at the beginning of the file.

        ##up to 2 expand populations may be added in between HighIDThetaDotPop and ThetaDotDtPop.  Expand 1 and 2 populations are required if 1 or more
        ##neurons in the ThetaDotDtPop require > MaxConnectionsPerCell connections.  expand 1 and 2 are described at the start of the file

        ##The function returns ThetaDotDtPop and information related to it, i.e. the population size, the number of neurons representing negative values,
        ##the value each neuron represents, and the minimium and maximum value ThetaDotDtPop can represent. The function also returns the number of layers
        ##created for the neural function

        numberOfLayers=2  #this stores the number of neural layers created for the neural function.  The minimum number is 2, i.e. HighIDThetaDotPop
                          #and ThetaDotDtPop

        ################create the first layer of the neural function (i.e. the HighIDThetaDotPop) and connect the thetadot input (i.e. self.thetaDotPop) to
        ################HighIDThetaDotPop.  The delay between the two is set so the spikes from self.thetaDotPop (which is also the output of the previous
        ################connected RWIP model) reach HighIDThetaDotPop at the same time as the other RWIP model inputs (i.e. the spikes from self.forcePop and
        ################self.thetaPop) reach the first layer of their corresponding neural functions implemented at the same time.
        delay=1.0
        HighIDThetaDotPopAndInfo=self.CreateAndConnectAHighIDPop(self.thetaDotPop,delay+self.RWIPOutputDelay) #connects self.thetaDotPop to HighIDThetaDotPop
                                                         
        ################create ThetaDotDtPop
        ThDotDtNeuronValue=self.thetaPop[2]    ###the value each neuron in ThetaDotDtPop needs to represent.  The value needs to be same as
                                               ###the value each neuron in self.thetaPop represents.
        ##create ThetaDotDtPop using the population specification in HighIDThetaDotPopAndInfo and ThDotDtNeuronValue
        ThetaDotDtPopAndInfo=self.createThetaDotDtPopulation(HighIDThetaDotPopAndInfo,ThDotDtNeuronValue)

        ############connect HighIDThetaDotPop to ThetaDotDtPop so that ThetaDotDtPop represents the result of the above calculation when a thetaDot value is
        ############given by HighIDThetaDotPop.  The following parameters are required for the function self.genConnsNegPosHighIDToOutPops that generates the
        ############connections and connects the populations:
        connsToProduce=2 ####is the option to generate the connections between HighIDThetaDotPop and ThetaDotDtPop so that the above calculation is implemented.
        connect=1  ####connect the layers (populations)

        #####extract information about HighIDThetaDotPop
        [thetaDotNeuronValue,ThetadotPopSize,ThetaDotNegNum,ThetaDotPosNum]=self.extractPopInfo(HighIDThetaDotPopAndInfo)

        #####extract information about ThetaDotDtPop
        [thetaDotDtNeuronValue,ThetadotDtPopSize,ThetaDotDtPopNumNeg,ThetaDotDtPopNumPos]=self.extractPopInfo(ThetaDotDtPopAndInfo)

        ##information related to the neurons in HighIDThetaDotPop (i.e. the values it can represent) that (if they fire) will result in the ThetaDotDtPop
        ##representing a negative value (result):
        firstCellThetaDotNeg=0               #the first neuron in HighIDThetaDotPop that will result in ThetaDotDtPop representing a negative value
        firstHighIDThetaDotValNegRes=thetaDotNeuronValue*-1.0 #the value represented by the first neuron in HighIDThetaDotPop that will result in ThetaDotDtPop
                                                              #representing a negative value
        HighIdThetaDotNeuronsResNegOut=[ThetaDotNegNum,firstCellThetaDotNeg,firstHighIDThetaDotValNegRes]

        ##information related to the neurons in HighIDThetaDotPop (i.e. the values it can represent) that (if they fire) will result in the ThetaDotDtPop
        ##representing a positive value (result):

        firstCellThetaDotPos=ThetaDotNegNum   #the first neuron in HighIDThetaDotPop that will result in ThetaDotDtPop representing a positive value
        firstHighIDThetaDotValPosRes=thetaDotNeuronValue #the value represented by the first neuron in HighIDThetaDotPop that will result in ThetaDotDtPop
                                                              #representing a positive value
        HighIdThetaDotNeuronsResPosOut=[ThetaDotPosNum,firstCellThetaDotPos,firstHighIDThetaDotValPosRes]

        ##generate the connections between HighIDThetaDotPop to ThetaDotDtPop and connect them. The function returns the number of extra layers created, i.e.
        ##whether expand 1 and 2 populations (layers) were created between HighIDThetaDotPop and ThetaDotDtPop.  The function also returns the ThetaDotDtPop,
        ##HighIDThetaDotPop and connections, which are redundant as the network has already been connected.

        [PopsAndConns,extralayers]=self.genConnsNegPosHighIDToOutPops(HighIDThetaDotPopAndInfo,HighIDThetaDotPopAndInfo,HighIdThetaDotNeuronsResNegOut,HighIdThetaDotNeuronsResPosOut,ThetaDotDtPopAndInfo,[],connect,connsToProduce)

        numberOfLayers+=extralayers

        return [ThetaDotDtPopAndInfo,numberOfLayers]

    def calculateOutCellsToActivate(self,whichCalc,inputValue,OutCellNeuronValue):
        ##This function calculates and returns the number of neurons in the last (output) layer of a neural mathematical function (the neural function is
        ##given in whichCalc) that need to fire/be activated (i.e. the value the last layer needs to represent) when the input population represents value
        ##inputValue.

        ##OutCellNeuronValue is the value each neuron in the last (output) layer represents, and is also used to calculate the number of output neurons to
        #activate.
    
        if(whichCalc==0):
            #neural function that calculates Sin(theta)*A*dt
            OutValue=self.ADt*sin(inputValue)   #the value the last (output) layer of the neural function needs to represent when the input population
                                                #represents theta=inputValue
        elif(whichCalc==1):
            #neural function that calculates (force-friction)*B*dt
            OutValue=self.BDt*(inputValue-self.frictVal)#the value the last (output) layer of the neural function needs to represent when the input population
                                                        #represents force=inputValue
        elif(whichCalc==2):
            #neural function that calculates thetaDot*dt
            OutValue=self.dt*inputValue #the value the last (output) layer of the neural function needs to represent when the input population
                                        #represents thetaDot=inputValue
            
        if(OutValue<0):
            addec=-0.000001
        else:
            addec=0.000001

        ##calculate the number of neurons in the last (output) layer of the neural function that need to be activated to represent the OutValue.
        numCellsToActivate=int(round(abs((OutValue/OutCellNeuronValue)+addec)))

        return numCellsToActivate

    #here
    def generateInExp1Exp2OutConnections(self,NetInfo,InputCell,OutputCell, wt,inCellCounter,determineMaxConnReached):
        #This function is used by generateConns2HighIDSubsTo1AddSub and genConnsBetHighIDAndOutPops to determine whether expand1 and 2 populations
        #are required between an input and output population i.e. input-->expand1-->expand2-->output, generates connections between them and increases the size
        #of the expand 1 and 2 populations.  More specifically, it does the following:
        # A) generates the following connections between the given neuron (InputCell) from an input population (a HighIDPop) and
        #    an output cell (OutputCell) form an output population, where the network containing the populations (i.e. the information and connections generated
        #    so far) is contained in NetInfo and the connection weight to use is wt:

        #    1. InputCell and OutputCell, which is added to connection set ExcConnectorHighIDToOut in NetInfo.
        #    2. InputCell and an expand 1 cell from an expand 1 population (given in NetInfo), which is added to connection set ExcConnectorHighIDToEx1
        #       in NetInfo.
        #    3. the given expand 1 cell and OutputCell, which is added to connection set ExcConnectorEx1ToOut in NetInfo.
        #    4. the given expand 1 cell and an expand 2 from an expand 2 population (given in NetInfo), which is added to connection set
        #       ExcConnectorEx1ToEx2 in NetInfo. 
        #    5. the given expand 2 cell and OutputCell, which is added to connection set ExcConnectorEx2ToOut in NetInfo.
        #    Note, each output cell is initially associated to one expand 1 and 2 cell, which the function generates connections for. However, whether they
        #    are used or not, i.e. connected, is dependent on below.
 
        #B) determines from the information provided by NetInfo and the newly generated (above) connections whether an expand 1 and
        #   2 population is required (i.e. need to be created) between the input and output populations.  An expand 1 population is required
        #   (i.e. the function sets expand1PopRequired to 1, see below) when an output cell in the output population
        #   requires >=self.MaxConnectionsPerCell connections from neurons in the input population (where the number of connections for the given output cell
        #   is given by numInToOutConn in NetInfo), and an expand 2 population is required (i.e. the function sets expand2PopRequired to 1, see below) when the
        #   number of expand 1 cells associated to an output cell is == self.MaxConnectionsPerCell (where the number of expand 1 cells associated to the given
        #   output cell is Expand1CellCounterForOutCell).  The connection sets used to connect HighIDPop and the output population, and whether expand 1
        ##  and 2 populations are created, is decided by function self.connectHighIDPopToOutPop using the information in NetInfo.  That is, connection sets
        ##  2,4,5 (above) would be used when expand1PopRequired==1 and expand2PopRequired==1, connection sets 2,3 would be used when expand1PopRequired==1
        ##  and expand1PopRequired==0, and connection set 1, when expand1PopRequired==0 and expand1PopRequired==0.  

        #C) determines whether another expand 1 and 2 cell needs to be added to the expand 1 and 2 populations for the given output cell.
        
        #The function returns the above information and connections by updating the connections sets and parameters in NetInfo and returning NetInfo.

        #For more information about highIDPops and expand populations, and an example, see the beginning of the file.

        #The function parameter inCellCounter is the index of InputCell in a subset of neurons in the input population that needs to be connected to the
        #given output cell and an expand 1 cell associated to the output cell. InputCell (below) is the index of the neuron
        #in the entire input population.  The two values can be the same.

        #The function parameter determineMaxConnReached=1 when the number of output cell connection needs to be checked in relation to the above.
        
        
        connections=NetInfo[0]                       #NetInfo contains the connections for a network so far.
        connectionAndPopsInfo=NetInfo[1]
        
        ExcConnectorHighIDToEx1=connections[0]       ##connections between the input (HighIdPop) and expand 1 populations
        ExcConnectorEx1ToOut=connections[1]          ##connections between the expand 1 and output populations
        ExcConnectorEx1ToEx2=connections[2]          ##connections between the expand 1 and expand 2 populations
        ExcConnectorEx2ToOut=connections[3]          ##connections between the expand 2 and output populations
        ExcConnectorHighIDToOut=connections[4]       ##connections between the input and output populations

        numInToOutConn=connectionAndPopsInfo[6]        ##number of connections that have been made between neurons in the input population and the given
                                                       ##output cell.  This also corresponds to the number of connections between the same neurons in the
                                                       ##input population and the given expand 1 cell.  This is reset everytime an expand 1 cell is added to
                                                       ##the expand 1 population, i.e. everytime numHighIDToOutConn >=self.MaxConnectionsPerCell.
                                                           
        alreadyconnExp1ToExp2AndOut=connectionAndPopsInfo[7]  ##a list stating whether the expand 1 cells (associated to the given output cell) are already 
                                                              ##connected to the given output cell and expand 2 cell.

        Expand1CellCounterForOutCell=connectionAndPopsInfo[4]  ##number of expand 1 cells associated to the output cell
        expand2Cell=connectionAndPopsInfo[3]                   ##number of expand 2 cells in the expand 2 population
        expand1PopRequired=connectionAndPopsInfo[0]            ##==1 if an expand 1 population is required
        expand2PopRequired=connectionAndPopsInfo[1]            ##==1 if an expand 2 population is required
        expand1Cell=connectionAndPopsInfo[2]                   ##total number of expand 1 cells in the expand 1 population.

        alreadyconnInToOutAndExp1=connectionAndPopsInfo[5] ## a list stating whether the neurons in an input population (that need to be connected to
                                                           ## the given output cell and expand 1 cell) are already connected, i.e. ==1.
        
        ##a connection is generated for every connection set (described above). 
        
        if(alreadyconnInToOutAndExp1[inCellCounter]==0): ##if the given input cell has not been connected to the given output cell and expand 1 cell           
            ExcConnectorHighIDToOut+=[(InputCell, OutputCell, wt, 1.0)]                                                                                                                          
            ExcConnectorHighIDToEx1+=[(InputCell, expand1Cell , wt, 1.0)] 
            if(alreadyconnExp1ToExp2AndOut[Expand1CellCounterForOutCell]==0): #if the given expand 1 cell is not connected to the given output cell and expand 2 cell                                                                                                                           
                ExcConnectorEx1ToOut+=[(expand1Cell, OutputCell, self.minWt+self.additionalWt, 1.0)] 
                ExcConnectorEx1ToEx2+=[(expand1Cell, expand2Cell, self.minWt+self.additionalWt, 1.0)]                                  
                ExcConnectorEx2ToOut+=[(expand2Cell, OutputCell, self.minWt+self.additionalWt, 1.0)]
                alreadyconnExp1ToExp2AndOut[Expand1CellCounterForOutCell]=1   #connections between the given expand 1, expand 2 and output cells have been made              
            numInToOutConn+=1  #the given output and expand 1 cell has another connection between them.
            alreadyconnInToOutAndExp1[int(inCellCounter)]=1  #connections between the given input, expand 1 and output cells have been made            

        ##determine whether an expand 1 and 2 pop are required and whether to add another expand 1 and expand 2 cell to the expand 1 and 2 populations
        ##for the current output cell.

        if(determineMaxConnReached==1): ## check whether the given output cell requires another expand 1 and or 2 cell.
            if(numInToOutConn>=self.MaxConnectionsPerCell): ##if the output cell has another 12 or more connections
                Expand1CellCounterForOutCell+=1  ##number of expand 1 cells associated to the output cell.
                numInToOutConn=0  ##reset the number of connections for the output cell
                alreadyconnExp1ToExp2AndOut+=[0] #add a connection flag for the new expand 1 cell.
                expand1Cell+=1   ##add another expand 1 cell to the expand 1 population
                expand1PopRequired=1  ##an expand1 pop is required
                if(((Expand1CellCounterForOutCell%self.MaxConnectionsPerCell)==0)): ##if self.MaxConnectionsPerCell expand 1 cells are associated to the given
                                                                                    ##output cell
                    expand2Cell+=1  ##add another expand 2 cell to the expand 2 population
                    expand2PopRequired=1 #an expand2 pop is required

        ##update all info sets
        connections[0]=ExcConnectorHighIDToEx1
        connections[1]=ExcConnectorEx1ToOut
        connections[2]=ExcConnectorEx1ToEx2
        connections[3]=ExcConnectorEx2ToOut
        connections[4]=ExcConnectorHighIDToOut

        connectionAndPopsInfo[7]=alreadyconnExp1ToExp2AndOut 
        connectionAndPopsInfo[6]=numInToOutConn

        connectionAndPopsInfo[4]=Expand1CellCounterForOutCell
        connectionAndPopsInfo[3]=expand2Cell
        connectionAndPopsInfo[0]=expand1PopRequired
        connectionAndPopsInfo[1]=expand2PopRequired
        connectionAndPopsInfo[2]=expand1Cell
        connectionAndPopsInfo[5]=alreadyconnInToOutAndExp1

        NetInfo[0]=connections
        NetInfo[1]=connectionAndPopsInfo

        return NetInfo

    def generateConns2HighIDSubsTo1AddSub(self,highIDSubSize1,highIDSubSize2,firstCellHighIDSub1,firstCellADDsub):
        ##This function generates and returns the connections between a highID population (i.e. between the neurons in 2 consecutive subsets of a highIDPop
        ##(neurons in sub1 are firstCellHighIDSub1 to firstCellHighIDSub1+highIDSubSize1, and neurons in sub 2 are
        ##firstCellHighIDSub2=highIDSubSize1+firstCellHighIDSub1 to firstCellHighIDSub2+highIDSubSize2) and an output (add) population (i.e.a subset of neurons
        ##firstCellADDsub to firstCellADDsub+highIDSubSize1+highIDSubSize2), so the first and consecutive neurons in the subset of the output population
        ##repesents the total of the 2 values given by the two HighID subsets, i.e. when a specific neuron in each of the two highID subsets fire, a
        ##specific number of neurons in the output population will fire.

        ###HighID populations are described at the beginning of the file.

        ##The function also produces connections for expand 1 and 2 populations in between the HighIDPop and output population subsets, i.e.

        ##HighIDPop(subset1)--->expand 1----->expand 2----->OutputPop(subset)
        ##HighIDPop(subset2)

        ##These connections would be used, and expand 1 and 2 populations would be created, if required (i.e. if 1 or more neurons in the output pop subset
        ##require > MaxConnectionsPerCell connections) by self.connectHighIDPopToOutPop using the information returned by this function.  see the beginning of
        ##the file for more information about the expand populations.

        ##Thus, the function produces the following connection sets:

        #1. between HighIDPop and AddPop, stored in array ExcConnectorHighIDToAdd.
        #2. between HighIDPop and exp1, stored in array ExcConnectorHighIDToEx1. 
        #3. between exp1 and AddPop, stored in array ExcConnectorEx1ToAdd.
        #4. between exp1 and exp2, stored in array ExcConnectorEx1ToEx2
        #5. between exp2 and AddPop, stored in array ExcConnectorEx2ToAdd

        

        ##calculate sub sizes
        AddSubSize=highIDSubSize1+highIDSubSize2#size of ADDpop subset  
        firstCellHighIDSub2=highIDSubSize1+firstCellHighIDSub1   ##first cell in HIGHID subset 2
        if(firstCellADDsub==[]): #if the first cell in the add pop sub is not given
            firstCellADDsub=firstCellHighIDSub1  #first cell in ADDpop subset


        ##initialise connection sets, described above
        ExcConnectorHighIDToEx1=[]  
        ExcConnectorEx1ToAdd=[]  
        ExcConnectorEx1ToEx2=[] 
        ExcConnectorEx2ToAdd=[]
        ExcConnectorHighIDToAdd=[]

        ##create a set of the above connections
        connections=[ExcConnectorHighIDToEx1,ExcConnectorEx1ToAdd,ExcConnectorEx1ToEx2,ExcConnectorEx2ToAdd,ExcConnectorHighIDToAdd]

        ##initialise other parameters
        expand1PopRequired=0  #==1 if an expand 1 population is required
        OutCellCounter=0   #the index of the first output neuron in the subset.
        expand1Cell=0      ##Number of neurons in the expand1 population.
        expand2Cell=0     ##Number of neurons in the expand2 population
        expand2PopRequired=0    ##==1 if an expand 2 population is required
        Expand1CellCounterForOutCell=0   ##number of expand 1 cells associated to an output cell.

        ##create an information list of the above
        connectionAndPopsInfo=[expand1PopRequired,expand2PopRequired,expand1Cell,expand2Cell,Expand1CellCounterForOutCell,[],[],[]] #==[] are to be filled in below
        
        ################generate the connections between the neurons in the first and second HighID subsets and the neurons in the output pop subset.
        for eachOutputSubCell in range(firstCellADDsub,firstCellADDsub+AddSubSize+1): ##for each output (add) cell in the subset.  Note, we add +1
                                                                                      ##as a total of two inputs can be 0, which needs to be skipped.
                                                                                   
            inSub1CellCounter=0   ###index of the current neuron in the first input subset  
            alreadyconnExp1ToExp2AndOut=[0] ##a list stating whether the expand 1 cells (associated to the given output cell) are already 
                                            ##connected to the current output cell and expand 2 cell.  An element is added for each expand 1 cell.  
            Expand1CellCounterForOutCell=0
            numHighIDToAddConn=0   #number of connections between the input neurons and the current output and expand 1 cell.
                                #It resets every self.MaxConnectionsPerCell is reched 
            alreadyconnInSub1ToOutAndExp1=zeros(highIDSubSize1) ##a list stating whether the input neurons in subset 1 are already 
                                                                ##connected to the given output cell and expand 1 cell.
            alreadyconnInSub2ToOutAndExp1=zeros(highIDSubSize2) ##a list stating whether the input neurons in subset 2 are already 
                                                                ##connected to the given output cell and expand 1 cell.

            ###update the connection/network information.
            connectionAndPopsInfo[7]=alreadyconnExp1ToExp2AndOut
            connectionAndPopsInfo[6]=numHighIDToAddConn
            connectionAndPopsInfo[4]=Expand1CellCounterForOutCell

            for eachInputSub1Cell in range(firstCellHighIDSub1,firstCellHighIDSub1+highIDSubSize1+1): #for each cell in the highIDPop subset 1, 
                inSub2CellCounter=0  ###index of the current input neuron in the second subset  
                if(inSub1CellCounter>OutCellCounter):  #the value represented by an input neuron cannot be > the total
                    break                 
                for eachInputSub2Cell in range(firstCellHighIDSub2,firstCellHighIDSub2+highIDSubSize2+1): #for each cell in the highIDPop subset 2

                    inputsSumResult=inSub1CellCounter+inSub2CellCounter  #total value represented by eachInputSub1Cell and eachInputSub2Cell
                    if(inputsSumResult>OutCellCounter):  #if the total is > the value eachOutputSubCell can represent
                        break
                    
                    if((inputsSumResult==OutCellCounter)and(OutCellCounter!=0)): #if the current output cell represents the total value represented by
                                                                                 #the current neurons in the first and second input subsets
                        ###generate a connection between the current neuron in input subset 1 (eachInputSub1Cell) and the current output cell
                        if(inSub1CellCounter>0):                        #if eachInputSub1Cell contributes to the total.
                            In1ValueDif=float(inSub1CellCounter)/float(OutCellCounter)  ##eachInputSub1Cell's contribution to the total  
                            wt=(self.minWt)*In1ValueDif                    ##the input contribution of eachInputSub1Cell to eachOutputSubCell. self.minWt is the
                                                                        ##minimum input required to activate eachOutputSubCell
                            if(inSub2CellCounter==0):              ##if eachInputSub2Cell contributes nothing 
                                determineMaxConnReached=1            ###check the number of connections to the output cell
                            else:
                                determineMaxConnReached=0            ##do not check the number of connections to the output cell

                            #update the connection/network information
                            connectionAndPopsInfo[5]=alreadyconnInSub1ToOutAndExp1
                            NetInfo=[connections,connectionAndPopsInfo]
                            ##generate the connections, and update NetInfo as to whether expand 1 and 2 populations are required etc.
                            NetInfo=self.generateInExp1Exp2OutConnections(NetInfo,eachInputSub1Cell-1,eachOutputSubCell-1, wt,inSub1CellCounter-1,determineMaxConnReached)
                            connectionAndPopsInfo=NetInfo[1]
                            #update the connection/network information
                            alreadyconnInSub1ToOutAndExp1=connectionAndPopsInfo[5]
                            
                        ###generate a connection between the current neuron in input subset 2 (eachInputSub2Cell) and the current output cell (eachOutputSubCell)
                        if(inSub2CellCounter>0):                        #if eachInputSub2Cell contributes to the total
                            In2ValueDif=float(inSub2CellCounter)/float(OutCellCounter) ##eachInputSub2Cell's contribution to the total  
                            wt=(self.minWt)*In2ValueDif                    ##the input contribution of eachInputSub2Cell to eachOutputSubCell.  self.minWt is the
                                                                        ##minimum input required to activate eachOutputSubCell                                            
                            determineMaxConnReached=1  ###check the number of connections to the output cell

                            #update the connection/network information
                            connectionAndPopsInfo[5]=alreadyconnInSub2ToOutAndExp1
                            NetInfo=[connections,connectionAndPopsInfo]
                            ##generate the connections, and update NetInfo as to whether expand 1 and 2 populations are required etc. 
                            NetInfo=self.generateInExp1Exp2OutConnections(NetInfo,eachInputSub2Cell-1,eachOutputSubCell-1, wt,inSub2CellCounter-1,determineMaxConnReached)        
                            connectionAndPopsInfo=NetInfo[1]
                            #update the connection/network information
                            alreadyconnInSub2ToOutAndExp1=connectionAndPopsInfo[5]
                    inSub2CellCounter+=1
                inSub1CellCounter+=1
            connectionAndPopsInfo[2]+=1# expand1Cell+=1
            connectionAndPopsInfo[3]+=1# expand2Cell+=1
            OutCellCounter+=1
        
        expand1PopRequired=connectionAndPopsInfo[0]
        expand2PopRequired=connectionAndPopsInfo[1]
        expand1Cell=connectionAndPopsInfo[2]
        expand2Cell=connectionAndPopsInfo[3]
                
        numlayers=expand1PopRequired+expand2PopRequired  ##maximum of 2

        NetInfo=[expand1Cell,expand2Cell,expand1PopRequired,expand2PopRequired,numlayers]

        return[connections,NetInfo]

    def genConnsBetHighIDAndOutPops(self,firstCellHighIDSub,firstCellOutPop,highIDSubSize,OutPopSubSize,HighIDCellNeuronValue,OutCellNeuronValue, whichCalc,firstPosHighIDCell, firstHighIDValue):
        ##This function generates and returns the connections between a highID population (i.e. a subset of neurons from firstCellHighIDSub to
        ##firstCellHighIDSub+highIDSubSize1) and an output population (i.e. a subset of neurons from firstCellOutPop to firstCellOutPop+OutPopSubSize), so the
        ##first and consecutive neurons in the subset of the output population represents, i.e. when a specific neuron in the HighIDPop subset fires, a
        ##specific number of neurons in the output population subset will fire representing:
        ##1. when whichCalc<=2, the result of a calculation (the calculation is given by whichCalc) on the value represented/given by the neuron that fired in
        ##   the HighIDPop.  The value represented by a neuron in the HighIDPop is as follows: firstHighIDValue+(index of the neuron in the
        ##   subset*HighIDCellNeuronValue).
        ##2. when whichCalc==3, the value represented by the neuron that fired in the HighIDPop, where the value represented by a neuron in HighIDPop is as
        ##   follows: index of the neuron in the subset*HighIDCellNeuronValue

        ##The value represented by each neuron in the output population is OutCellNeuronValue.

        ##All of the neurons in the output population subset represent positive or negative values.  Thus, it is important that the subset of highID neurons
        ##given, i.e. the result of a calculation on the values the highID population subset can represent, can be represented by the output population subset.

        #The subset of highID neurons may include neurons that can represent negative or positive values.  Thus, for the function to work, it requires
        ##the index of the first neuron in the highID subset (i.e. the index of the neuron in the whole highID population) that represents a
        ##positive value, i.e. firstPosHighIDCell.

        ##When whichCalc==0, the output population will represent the result of self.ADt*sin(theta), where theta is represented by HighIDPop.
        ##When whichCalc==1, the output population will represent the result of self.BDt*(force-self.frictVal), where force is represented by HighIDPop.
        ##When whichCalc==2, the output population will represent the result of thetaDot*self.dt, where thetaDot is represented by HighIDPop. 

        ##The function also produces connections for expand 1 and 2 populations in between the HighIDPop and output population subsets, i.e:
        ##HighIDPop(subset)--->expand 1----->expand 2----->OutputPop(subset)
        
        ##These connections would be used, and expand 1 and 2 populations would be created, if required (i.e. if 1 or more neurons in the output pop subset
        ##requires >self.MaxConnectionsPerCell connections) by function self.connectHighIDPopToOutPop using the information returned by this function.  see the
        ##beginning of the file for more information about the expand populations.

        ##Thus, the function produces the following connection sets:

        #1. between HighIDPop and OutPop, stored in array ExcConnectorHighIDToOut.
        #2. between HighIDPop and exp1, stored in array ExcConnectorHighIDToEx1. 
        #3. between exp1 and OutPop, stored in array ExcConnectorEx1ToOut.
        #4. between exp1 and exp2, stored in array ExcConnectorEx1ToEx2
        #5. between exp2 and OutPop, stored in array ExcConnectorEx2ToOut


        #The connections and information about the network is returned.


        ##initialise connection sets, described above
        ExcConnectorHighIDToEx1=[]  
        ExcConnectorEx1ToOut=[]  
        ExcConnectorEx1ToEx2=[] 
        ExcConnectorEx2ToOut=[]
        ExcConnectorHighIDToOut=[]

        ##create a set of the above connections
        connections=[ExcConnectorHighIDToEx1,ExcConnectorEx1ToOut,ExcConnectorEx1ToEx2,ExcConnectorEx2ToOut,ExcConnectorHighIDToOut]

        ##initialise other parameters
        expand1PopRequired=0  #==1 if an expand 1 population is required
        OutCellCounter=0   #the index of the first output neuron in the subset.
        expand1Cell=0      ##Number of neurons in the expand1 population.
        expand2Cell=0     ##Number of neurons in the expand2 population
        expand2PopRequired=0    ##==1 if an expand 2 population is required
        Expand1CellCounterForOutCell=0   ##number of expand 1 cells associated to an output cell, 
        
        ####create an information list of the above
        connectionAndPopsInfo=[expand1PopRequired,expand2PopRequired,expand1Cell,expand2Cell,Expand1CellCounterForOutCell,[],[],[]] ##[] are filled in below

        storeHighIDCellNeuronValue=HighIDCellNeuronValue
        
        ##generate the connections
        for eachOutputCell in range(firstCellOutPop,firstCellOutPop+OutPopSubSize): #for each output cell in the subset
            inCellCounter=0     ##index of the current neuron in the input subset
            alreadyconnExp1ToExp2AndOut=[0] ##a list stating whether the expand 1 cells (associated to the current output cell) are already 
                                            ##connected to the given output cell and expand 2 cell.  An element is added for each expand 1 cell.
            numHighIDToAddConn=0            #number of connections between the input neurons and the current output and expand 1 cell.
                                            #It resets every self.MaxConnectionsPerCell is reached 
            Expand1CellCounterForOutCell=0   ##is the number of expand 1 cells associated to the current output cell
            alreadyconnInToOutAndExp1=zeros(highIDSubSize) ##a list stating whether the input neurons are already 
                                                           ##connected to the given output cell and expand 1 cell.  
            HighIDCellNeuronValue=storeHighIDCellNeuronValue
            HighIDvalue=firstHighIDValue
            
            ###update the connection/network information.
            connectionAndPopsInfo[7]=alreadyconnExp1ToExp2AndOut
            connectionAndPopsInfo[6]=numHighIDToAddConn
            connectionAndPopsInfo[5]=alreadyconnInToOutAndExp1
            connectionAndPopsInfo[4]=Expand1CellCounterForOutCell

            ##generate the connections between the input neurons and eachOutputCell
            for eachInputCell in range(firstCellHighIDSub,firstCellHighIDSub+highIDSubSize): #for each neuron in the input (HighIDPop) subset
                if((HighIDCellNeuronValue<0)and(eachInputCell==firstPosHighIDCell)): #if the current neuron in the input subset (eachInputCell) represents a
                                                                                      #positive value but the HighIDCellNeuronValue is currently negative,
                                                                                      #change it
                    HighIDvalue=0                                             
                    HighIDCellNeuronValue=abs(HighIDCellNeuronValue) #use the opposite sign of HighIDCellNeuronValue
                    
                ##calculate the number of output neurons that need to fire (be active) for the given whichCalc and when eachInputCell fires
                if(whichCalc==3):
                    numCellsToActivate=inCellCounter+1 
                else:
                    numCellsToActivate=self.calculateOutCellsToActivate(whichCalc,HighIDvalue,OutCellNeuronValue)
                #generate the connections between the corresponding each input, expand 1, expand 2 and output cells, so that when
                #eachInputCell fires, the correct neurons in the output population fire (are activated).
                if(numCellsToActivate>=(OutCellCounter+1)): #if the number of output cells to activate includes the current output cell
                    wt=(self.minWt+self.additionalWt) 
                    determineMaxConnReached=1   #check the number of connections on the output cell
                    NetInfo=[connections,connectionAndPopsInfo]
                    ##generate the connections, and update NetInfo as to whether expand 1 and 2 populations are required etc.
                    NetInfo=self.generateInExp1Exp2OutConnections(NetInfo,eachInputCell, eachOutputCell, wt,inCellCounter,determineMaxConnReached)
                    connectionAndPopsInfo=NetInfo[1]

                inCellCounter+=1
                HighIDvalue+=HighIDCellNeuronValue

            connectionAndPopsInfo[2]+=1# expand1Cell+=1
            connectionAndPopsInfo[3]+=1#expand2Cell+=1
            OutCellCounter+=1

        expand1PopRequired=connectionAndPopsInfo[0]
        expand2PopRequired=connectionAndPopsInfo[1]
        expand1Cell=connectionAndPopsInfo[2]
        expand2Cell=connectionAndPopsInfo[3]

        numlayers=expand1PopRequired+expand2PopRequired  #maximum of 2

        NetInfo=[expand1Cell,expand2Cell,expand1PopRequired,expand2PopRequired,numlayers]

        return [connections,NetInfo]


    def createSinADtPopulation(self,HighIDThetaPopAndInfo,sinAdtNeuronValue):
        #This function creates a population of neurons (referred to as SinThetaADtPop) to represent the result of, i.e. is the last layer (output population) of the neural
        #function that calculates, the following part of the numerical solution of the RWIP equation of motion: sin(theta)*A*dt.  The full solution is shown at
        #the top of the file.  The size of SinThetaADtPop is based on both the given value each neuron in SinThetaADtPop needs to represent
        #(i.e. sinAdtNeuronValue) and the following information regarding the population (i.e. HighIDThetaPop in HighIDThetaPopAndInfo) that
        #represents the theta value in the neural function: the value each neuron represents, and the number of neurons representing positive and negative values.
        #The function returns SinThetaADtPop along with the following information about it, i.e. the population size, the number of neurons in SinThetaADtPop
        #representing negative values, sinAdtNeuronValue, and the minimium and maximum value SinThetaADtPop can represent.
        
        ##extract information about the population HighIDThetaPop
        [thetaNeuronValue,ThetaPopSize,ThetaNegNum,ThetaPosNum]=self.extractPopInfo(HighIDThetaPopAndInfo)

        #####calculate the SinThetaADtPop size.
        minSinADtVal=self.ADt*sin(-1.0*ThetaNegNum*thetaNeuronValue)   ##minimum value SinThetaADtPop can represent
        SinThetaADtPopNumNeg=int(round((abs(minSinADtVal)/sinAdtNeuronValue))) ##number of neurons representing negative values
        
        maxSinADtVal=self.ADt*sin(ThetaPosNum*thetaNeuronValue)   ##maximum value SinThetaADtPop can represent
        SinThetaADtPopNumPos=int(round(maxSinADtVal/sinAdtNeuronValue))  ##number of neurons representing positive values.

        SinThetaADtPopSize=SinThetaADtPopNumNeg+SinThetaADtPopNumPos  #size of the population SinThetaADtPop

        ############create SinThetaADtPop population And Info Set
        model=1    #if_cond_exp neurons
        SinThetaADtPopAndInfo=self.createPopAndInfoSet(SinThetaADtPopSize,SinThetaADtPopNumNeg,sinAdtNeuronValue,minSinADtVal,maxSinADtVal,model,[])

        return SinThetaADtPopAndInfo

    def CreateSinADtNeuralFunction(self):
        ##This creates a neural function to calculate/output the result of the following calculation: sin(theta)*A*dt.

        ##The neural function consists of a minimum of two layers, the first (input) layer is HighIDThetaPop, where one neuron represents the theta
        ##value given by self.thetaPop, and a second (output) layer SinThetaADtPop represents the result of the above calculation, i.e. when a specific number
        ##of neurons in self.thetaPop fire, a specific neuron in HighIDThetaPop will fire causing a specific number of neurons in SinThetaADtPop to fire.
        ##More specifically, the result is represented by the first and consecutive neurons in, depending on the sign of the result, the set of neurons that
        ##represent positive or negative values in SinThetaADtPop.  HighID Pops are described at the beginning of the file.

        ##up to 2 expand populations may be added in between HighIDThetaPop and SinThetaADtPop.  Expand 1 and 2 populations are required if 1 or more
        ##neurons in the SinThetaADtPop require > MaxConnectionsPerCell connections.  expand 1 and 2 populations are described at the start of the file

        ##The function returns SinThetaADtPop and information related to it, i.e. the population size, the number of neurons representing negative values,
        ##the value each neuron represents, and the minimium and maximum value SinThetaADtPop can represent. The function also returns the number of layers
        ##created for the neural function

        numberOfLayers=2  #this stores the number of neural layers created for the neural function.  The minimum number is 2, i.e. HighIDThetaPop
                          #and SinThetaADtPop

        ################create the first layer of the neural function (i.e. the HighIDThetaPop) and connect the theta input (i.e. self.thetaPop) to
        ################HighIDThetaPop.  The delay between the 2 is set so the spikes from self.thetaPop (which is also the output of the previous connected
        ################RWIP model) reach HighIDThetaPop at the same time as the other RWIP model inputs (i.e. the spikes from self.forcePop and
        ################self.thetaDotPop) reach the first layer of their corresponding neural functions implemented at the same time.

        delay=1.0
        HighIDThetaPopAndInfo=self.CreateAndConnectAHighIDPop(self.thetaPop,delay+self.RWIPOutputDelay) #connects self.thetaPop to HighIDThetaPop.
                                                                                                                               
        ################create SinThetaADtPop
        sinThetaAdtNeuronValue=self.thetaDotPop[2] ###the value each neuron in SinThetaADtPop needs to represent.  The value needs to be same as
                                                    ###the value each neuron in self.thetaDotPop represents.
        
        ##create SinThetaADtPop using the population specification in HighIDThetaPopAndInfo and sinThetaAdtNeuronValue
        SinThetaADtPopAndInfo=self.createSinADtPopulation(HighIDThetaPopAndInfo,sinThetaAdtNeuronValue)

        ############connect HighIDThetaPop to SinThetaADtPop so that SinThetaADtPop represents the result of the above calculation when a theta value is
        ############given by HighIDThetaPop.  The following parameters are required for the function self.genConnsNegPosHighIDToOutPops that generates the
        ############connections and connects the populations: 
        connsToProduce=0 ####is the option to generate the connections between HighIDThetaPop and SinThetaADtPop so that the above calculation is implemented.
        connect=1  ####connect the layers (populations)

        #####extract information about HighIDThetaPop
        [thetaNeuronValue,ThetaPopSize,ThetaNegNum,ThetaPosNum]=self.extractPopInfo(HighIDThetaPopAndInfo)      

        #####extract information about SinThetaADtPop
        [SinThetaADtNeuronValue,SinThetaADtPopSize,SinThetaADtPopNumNeg,SinThetaADtPopNumPos]=self.extractPopInfo(SinThetaADtPopAndInfo) 

        ##information related to the neurons in HighIDThetaPop (i.e. the values it can represent) that (if they fire) will result in the SinThetaADtPop
        ##representing a negative value (result):
        firstCellThetaNeg=0               #the first neuron in HighIDThetaPop that will result in SinThetaADtPop representing a negative value
        firstHighIDThetaValNegRes=thetaNeuronValue*-1.0 #the value of the first neuron in HighIDThetaPop that will result in SinThetaADtPop
                                                              #representing a negative value
        HighIdThetaNeuronsResNegOut=[ThetaNegNum,firstCellThetaNeg,firstHighIDThetaValNegRes]

        ##information related to the neurons in HighIDThetaPop (i.e. the values it can represent) that (if they fire) will result in the SinThetaADtPop
        ##representing a positive value (result):

        firstCellThetaPos=ThetaNegNum   #the first neuron in HighIDThetaPop that will result in SinThetaADtPop representing a positive value
        firstHighIDThetaValPosRes=thetaNeuronValue #the value of the first neuron in HighIDThetaPop that will result in SinThetaADtPop
                                                              #representing a positive value
        HighIdThetaNeuronsResPosOut=[ThetaPosNum,firstCellThetaPos,firstHighIDThetaValPosRes]

        ##generate the connections between HighIDThetaPop to SinThetaADtPop and connect them. The function returns the number of extra layers created, i.e.
        ##whether expand 1 and 2 populations (layers) were created between HighIDThetaPop and SinThetaADtPop.  The function also returns the SinThetaADtPop,
        ##HighIDThetaPop and connections, which are redundant as the network has already been connected.

        [PopAndConns,extralayers]=self.genConnsNegPosHighIDToOutPops(HighIDThetaPopAndInfo,HighIDThetaPopAndInfo,HighIdThetaNeuronsResNegOut,HighIdThetaNeuronsResPosOut,SinThetaADtPopAndInfo,[],connect,connsToProduce)
        numberOfLayers+=extralayers 

        return [SinThetaADtPopAndInfo,numberOfLayers]

    def connectFinalLayerOfNeuralFunction(self, PopsAndConnections):
        ##This connects the final layer N of a neural function, i.e. the output population given in PopsAndConnections, to the previous layer in the
        ##function N-1, i.e. two HighID pops given in PopsAndConnections, one representing negative and the other representing positive values.
        ##If expand 1 and 2 layers/populations are required in between layers N-1 and N, which is stated in PopsAndConnections, the function will create the
        ##expand 1 and 2 populations for both HighIDpops and connect the following network:

        ##HighID Pop1 representing negative values(Layer N-3)-->exp1(Layer(N-2)-->exp2(Layer(N-1)-->outputPop(N) (representing negative values).
        ##HighID Pop2 representing positive values(Layer N-3)-->exp1(Layer(N-2)-->exp2(Layer(N-1)-->outputPop(N) (representing positive values).

        ##See the start of the file for more information related to the expand and HighID populations.

        OutputPop=PopsAndConnections[0]    #extract the output population
        if(len(OutputPop)>1):      #as the outputpop may contain other information such as the number of neurons etc, extract the population
            OutputPop=OutputPop[0]
        inputNegPop=PopsAndConnections[1]  #extract the HighID population representing negative values
        inputPosPop=PopsAndConnections[2]  #extract the HighID population representing positive values
        negConnections=PopsAndConnections[3] ##contains 5 sets of connections in the following order (1. HighID-->exp1, 2.exp1-->out, 3.exp1-->exp2, 4.exp2-->out , 5.HighID-->out)
        posConnections=PopsAndConnections[4] ##contains 5 sets of connections, same as above

        ##connect the inputNegPop to the OutputPop, and, if required, connect them to the expand 1 and 2 populations
        self.connectHighIDPopToOutPop(inputNegPop, OutputPop,negConnections)
        ##connect the inputPosPop to the OutputPop, and, if required, connect them to the expand 1 and 2 populations           
        self.connectHighIDPopToOutPop(inputPosPop, OutputPop,posConnections)
        
    def CreateRWIPModel(self):
        #This function creates a neural RWIP forward model to calculate the next RWIP state from the current RWIP state.  In other words, it builds and connects the
        #following F1-F6 components (neural mathematical functions), where the output of one function is the input to another, so that the RWIP model's
        #output, i.e. the next theta and theta dot values (represented by the number of neurons in the output populations: nextThetaPop and
        #nextThetaDotPop that fire) are calculated from its inputs, i.e. current force, theta and thetadot values (represented by the number of neurons in the
        #input populations self.forceBufferPop/self.forcePop, self.thetaPop and self.thetaDotPop that fire):

        #F1 = a*sin(theta)*dt.                  A SinADtNeuralFunction is built, where its input (theta) and output (F1) are represented by populations self.thetaPop
        #                                       and SinADtPop respectively.  Note, the input and output values of a function are represented by the number of neurons 
        #                                       in the function's input and output populations that fire.  This is the same for all of the functions.
        #F2 = (Force-friction)*B*dt.            A FFrBDtNeuralFunction is built, where its input (force) and output (F2) are represented by populations 
        #                                       self.forceBufferPop (initialised by self.forcePop) and FFrBDtPop respectively.
        #F3 = F1+F2                             An AdditionNeuralFunction is built, where its inputs (F1 and F2) are represented by populations SinADtPop
        #                                       FFrBDtPop, and its output (F3) is represented by SplusFPop.
        #F4 (next thetadot value) = F3+thetadot     An AdditionNeuralFunction is built, where its inputs (F3 and thetadot) are represented by
        #                                           populations SplusFPop and self.thetaDotPop respectively, and its output (F4) is represented by nextThetaDotPop.
        #                                           nextThetaDotPop also represents the thetadot value given to (i.e. is an input population of) the next
        #                                           connected RWIP model.
        #F5 = thetadot*dt                       A ThetaDotDtNeuralFunction is built, where its input (thetadot) and output (F5) are represented by populations
        #                                       self.thetaDotPop and ThDotDtPop respectively.
        #F6 (next theta value) = F5+theta       An AdditionNeuralFunction is built, where its inputs (F5 and theta) are represented by
        #                                       populations ThDotDtPop and self.thetaPop respectively, and its output (F6) is represented by nextThetaPop.
        #                                       nextThetaPop also represents the theta value given to (i.e. is an input population of) the next connected RWIP model.
        #                                       
        #The functions run as follows:
        #-------------> Time
        #F1-->F3-->F4 
        #F2-->
        #F5-->F6

        #This function calculates appropriate connection delays between neural functions/populations so the neural functions (and the next connected RWIP model)
        #receive all inputs (spikes) at the same time.  F1, F2 and F5 receive their inputs, i.e. theta, force and thetadot values, at the same time,
        #i.e. self.RWIPOutputDelay ms after the previous RWIP state (theta and thetadot values) was calculated by the previous RWIP model.

        #The function returns the output populations of the RWIP system, i.e. nextThetaPop and nextThetaDotPop, and the time it takes the RWIP system
        #to calculate the next RWIP state.
 
        ################################CREATE A NETWORK TO CALCULATE F4, i.e. the next thetadot value
        
        RWIPNumLayersNextThDot=1  #number of layers in the RWIP model used/created to calculate the next thetadot value.
                                  #The initial number is 1 for the theta (self.thetaPop population), thetadot (self.thetaDotPop population)
                                  #and force (self.forcePop) inputs.
                                    
        ###CREATE NEURAL FUNCTION TO CALCULATE F1: a*sin(theta)*dt, where the theta value is represented by population self.thetaPop and the result is
        ###represented by population SinADtPop.  The function below returns SinADtPop along with the number of neural layers used/created to calculate the
        ###above.
        
        [SinADtPop,SinNumLayers]=self.CreateSinADtNeuralFunction()

        ###CREATE NEURAL FUNCTION TO CALCULATE F2: (Force-friction)*B*dt,  where the force value is represented by population self.forcePop (where the initial
        ###force value is represented by self.initForcePop) and the result is represented by population FFrBDtPop.  The function below returns FFrBDtPop
        ###along with the number of neural layers used/created to calculate the above.

        [FFrBDtPop,FFrBDtNumLayers]=self.CreateFFrBDtNeuralFunction()

        ###CREATE AN ADDITION NEURAL FUNCTION TO CALCULATE F3: F1+F2, where the F1 and F2 values are represented by SinADtPop
        ###and FFrBDtPop respectively, and the result is represented by SplusFPop.

        #calculate connection delays between SinADtPop and FFrBDtPop and the first layer of the addition neural function, so that
        #the first layer receives all inputs (spikes) at the same time.

        if(SinNumLayers>FFrBDtNumLayers):
            layDifference=SinNumLayers-FFrBDtNumLayers     #difference between the number of layers used to calculate F1 and F2 
            maxNumInSFLayers=SinNumLayers                  #maximum number of layers 
            delayInputSinADt=0.0                           #delay required for connections between SinADtPop and the first layer of the addition neural function
            delayInputFFrBDt=layDifference*self.timePerLayer    #delay required for connections between FFrBDtPop and the first layer of the addition neural function
        elif(FFrBDtNumLayers>SinNumLayers):
            layDifference=FFrBDtNumLayers-SinNumLayers
            maxNumInSFLayers=FFrBDtNumLayers
            delayInputSinADt=layDifference*self.timePerLayer
            delayInputFFrBDt=0.0
        else:
            delayInputSinADt=0.0
            delayInputFFrBDt=0.0
            maxNumInSFLayers=FFrBDtNumLayers
             
        RWIPNumLayersNextThDot+=maxNumInSFLayers
        
        ##CREATE NEURAL FUNCTION
        connectFinalLayer=1 #connect the final layer of the neural function
        outputAddPop=[]     #create a neural function output population, i.e. SplusFPop.

        #The function below returns SplusFPopAndConn, containing SplusFPop, along with the number of neural layers used/created to calculate the addition.
        [SplusFPopAndConn,numAddLayers]=self.createAdditionNeuralFunction(SinADtPop, FFrBDtPop, delayInputSinADt, delayInputFFrBDt, outputAddPop,connectFinalLayer)

        SplusFPop=SplusFPopAndConn[0]       #extract the SPlusFPop population, which represents the addition result.
        RWIPNumLayersNextThDot+=numAddLayers 

        ###CREATE AN ADDITION NEURAL FUNCTION TO CALCULATE F4: F3+thetadot, where the F3 and thetadot values are inputs to the addition neural function and
        ###represented by SplusFPop and self.thetaDotPop respectively, and the result is represented by nextThetaDotPop.

        ###calculate connection delays between SplusFPop and self.thetaDotPop and the first layer of the addition neural function, so that
        ###the first layer receives all inputs (spikes) at the same time.
        
        delayInputSplusF=0.0  #delay required for connections between SplusFPop and the first layer of the addition neural function.
        delayInputThetaDot=(RWIPNumLayersNextThDot*self.timePerLayer)+self.RWIPOutputDelay  #delay required for connections between self.thetaDotPop and the
                                                                                            #first layer of the addition neural function.
        if(self.nest==1):
            delayInputThetaDot-=1.0

        ###if a nextThetaDotPop population (i.e. an output population for the addition neural function) is not provided (i.e. the self.thetaDotPop
        ###of the next connected RWIP model is not provided) in self.outputPops, create one in function self.createAdditionNeuralFunction using the
        ###specification in self.thetaDotPop.
        
        if(self.outputPops!=[]):
            nextThetaDotPop=self.outputPops[1] 
        else:                                            
            nextThetaDotPop=[[],self.thetaDotPop[1],self.thetaDotPop[2] ,self.thetaDotPop[3]]   
                             
        ##CREATE NEURAL FUNCTION
        connectFinalLayer=0         #do not connect the final layer of the neural function
        outputAddPop=nextThetaDotPop

        ###The function below returns nextThetaDotPopAndConn, containing nextThetaDotPop, the number of neural layers used/created to calculate the addition.        
        [nextThetaDotPopAndConn,numAddLayers]=self.createAdditionNeuralFunction(SplusFPop, self.thetaDotPop, delayInputSplusF, delayInputThetaDot, outputAddPop,connectFinalLayer)  

        nextThetaDotPop=nextThetaDotPopAndConn[0]  #extract the nextThetaDotPop population
        RWIPNumLayersNextThDot+=numAddLayers

        ################################CREATE A NETWOK TO CALCULATE F6, i.e. the next theta value

        RWIPNumLayersNextTh=1     #number of layers in the RWIP system used/created to calculate the next theta value.
                                  #The initial number is 1 for the theta (self.thetaPop population) and thetadot (self.thetaDotPop population) inputs.

        ###CREATE NEURAL FUNCTION TO CALCULATE F5: thetadot*dt, where the thetadot value is a represented by population self.thetaDotPop and the result is
        ###represented by population ThDotDtPop.  The function below returns ThDotDtPop along with the number of neural layers used/created to calculate the
        ###above.
        
        [ThDotDtPop,ThetaDotDtNumLayers]=self.CreateThetaDotDtNeuralFunction()
        RWIPNumLayersNextTh+=ThetaDotDtNumLayers

        ###CREATE AN ADDITION NEURAL FUNCTION TO CALCULATE F6: F5+theta, where the F5 and theta values are inputs to the addition neural function and
        ###represented by ThDotDtPop and self.thetaPop respectively, and the result is represented by nextThetaPop

        ###calculate connection delays between ThDotDtPop and self.thetaPop and the first layer of the addition neural function, so that
        ###the first layer receives all inputs (spikes) at the same time.        

        delayInputThDotDt=0.0 #delay required for connections between ThDotDtPop and the first layer of the addition neural function.
        delayInputTheta=(RWIPNumLayersNextTh*self.timePerLayer)+self.RWIPOutputDelay #delay required for connections between self.thetaPop and the first
                                                                                     #layer of the addition neural function.
        if(self.nest==1):
            delayInputTheta-=1.0

        ###if a nextThetaPop population (i.e. an output population for the addition neural function) is not provided (i.e. the self.thetaPop
        ###of the next connected RWIP model is not provided) in self.outputPops, create one in function self.createAdditionNeuralFunction using the
        ###specification in self.thetaPop.

        if(self.outputPops!=[]): 
            nextThetaPop=self.outputPops[0]  
        else:                                            
            nextThetaPop=[[],self.thetaPop[1],self.thetaPop[2],self.thetaPop[3]] 

        ###CREATE NEURAL FUNCTION
        connectFinalLayer=0    #do not connect the final layer of the neural function
        outputAddPop=nextThetaPop

        ###The function below returns nextThetaPopAndConn, containing nextThetaPop, the number of neural layers used/created to calculate the addition.        
        [nextThetaPopAndConn,numAddLayers]=self.createAdditionNeuralFunction(ThDotDtPop, self.thetaPop, delayInputThDotDt, delayInputTheta, outputAddPop,connectFinalLayer) 
        nextThetaPop=nextThetaPopAndConn[0]        #extract the nextThetaPop population
        RWIPNumLayersNextTh+=numAddLayers

        ####################MODIFY THE CONNECTION DELAYS BETWEEN THE LAST LAYER (N) OF THE ADDITION NEURAL FUNCTIONS USED TO CALCULATE F4 AND F6 ABOVE,
        ####################(I.E. NEXTTHETAPOP AND NEXTTHETADOTPOP), AND LAYER N-1 SO THAT BOTH NEXTTHETAPOP AND NEXTTHETADOTPOP RECEIVE ALL THE SPIKES, AND FIRE,
        ####################AT THE SAME TIME, I.E. THE NEXT CONNECTED RWIP SYSTEM RECEIVES THE THETA AND THETADOT VALUES AT THE SAME TIME.
 
        #####modify the connection delays
        [ModNextThetaDotPopAndConn,ModNextThetaPopAndConn]=self.ModifyNextThetaThetaDotConnDelays(RWIPNumLayersNextTh,RWIPNumLayersNextThDot, nextThetaDotPopAndConn,nextThetaPopAndConn)

        #connect the final (N) and previous (N-1) layers of the addition neural functions implemented to calculate f4 and f6 above, where the final layers of
        #f4 and f6 are nextThetaDotPop and nextThetaPop respectively, using the modified connections
        self.connectFinalLayerOfNeuralFunction(ModNextThetaDotPopAndConn)  
        self.connectFinalLayerOfNeuralFunction(ModNextThetaPopAndConn)  
        
        #####################calculate the time (neurally) it takes the RWIP model to calculate the next RWIP state, i.e. the next theta and thetadot value
            
        if(RWIPNumLayersNextThDot>RWIPNumLayersNextTh): 
            self.RWIPCalcTimePlusDelay=(RWIPNumLayersNextThDot*self.timePerLayer)+self.RWIPOutputDelay
        else:
            self.RWIPCalcTimePlusDelay=(RWIPNumLayersNextTh*self.timePerLayer)+self.RWIPOutputDelay

        #####################output populations of the RWIP model, which represent the next theta and thetadot values
        self.outputPops=[nextThetaPop, nextThetaDotPop]

    def CreateAndManageForceInput(self):
        #This creates a neural network where a population of neurons called self.forceBufferPop, once enabled at time t:

        #1.  waits for spikes (i.e. the force value) from self.forcePop, for duration self.ForceBufferDuration ms (i.e. between t and t+self.ForceBufferDuration)
        #2.  holds them once received (i.e. self.forcePop activates the same neurons in self.forceBufferPop, which continously fire)  thus, depending on when
        #    they are received, the maximum time they are held (i.e. the time self.forceBufferPop fires/is activated) for is self.ForceBufferDuration ms.)
        #3.  releases the value (i.e. the spikes (input) from self.forceBufferPop are received by the RWIP model/self.HighIDForcePop (which is also created by
        #    this function and is the input population/layer to the FFrBDtNeuralFunction) at t+self.ForceBufferDuration+1.0 ms).

        #This is important for the RWIP model to work correctly on SpiNNaker, i.e. it allows time for the force value from self.forcePop to
        #be received, as exactly when the spikes would be received is unknown.  This is not required for NEST but as its built into the system
        #it is used.

        ################create self.HighIDForcePop
        model=1  #use if_cond_exp neurons
        addzero=1 #add a neuron represent the value 0.
        
        forceNeuronValue=self.forcePop[2]  ##the minimum value self.forcePop can represent
        forceMinMaxVal=self.forcePop[3]    ##the maximum value self.forcePop can represent
        self.HighIDForcePop=self.createValuePop(forceNeuronValue, forceMinMaxVal, model,addzero,[]) #create the population.  

        ###create a force buffer network including self.forceBufferPop, i.e. create a network (including self.forceBufferPop) that, along with self.RWIPModActForceCell
        ###enables self.forceBufferPop to receive the force value from self.forcePop between t and t+self.ForceBufferDuration (as mentioned above).
        self.CreateForceBufferNet()   

        ###create a network to control the output time of the force buffer network, i.e. the time the spikes (input) from self.forceBufferPop are received by
        ###self.HighIDForcePop, and disable self.forceBufferPop.
        self.ControlForceBufferOutput()      
