# -*- coding: utf-8 -*-
"""
This is one of 2 files required to run the 5s RWIP simulation (i.e. calculate the motion of a RWIP) on SpiNNaker.  The other is NeuralRWIPModel.py.
This script does the following:
1. creates a RWIP simulation system consisting of 3 RWIP forward models (each created using NeuralRWIPModel.py and stored in list
   self.RWIPModels) connected consecutively to form a loop.  Each model, in sequence, calculates one time step of the RWIP simulation (i.e. calculates and
   outputs the next RWIP state (which is the next theta (anglular displacement of the inverted pendulum from the upright (equilibrium) position (theta=0))
   measured in radians, and thetadot (angular velocity) values) using the output of the previous model (i.e. the theta and thetadot values from the model that
   calculated the previous RWIP state) and the torque/force value calculated by a PID controller to stabilise the RWIP.  Each model contains the
   following input and output populations:
       i) input population thetaPop, which represents the theta input value (a value in this system is represented by the number of neurons in a population
          that fire).
       ii) input population thetaDotPop, which represents the thetadot input value. 
       iii) input neuron RWIPModActForceCell, which activates (fires) when its corresponding model is next to calculate a RWIP state, i.e. 
            will receive theta, thetadot, and force values.  The RWIPModActForceCell neurons (which are contained in list self.RWIPModActForceCells) are
            connected consecutively to form a loop, where each neuron, in sequence, fires.
       iv) input populations forcePop/forceBufferPop, which represent the force input value.  forcePop is a set of spike injectors, used to inject the
           force value into SpiNNaker.  However, as it unknown when SpiNNaker/a model will receive the injected spikes, the forcePop is connected to
           forceBufferPop, which waits for spikes (a force value) from forcePop, holds them once received (i.e. forcepop activates the same neurons in
           forceBufferPop, which continuously fire) for a period of time, and then releases the spikes (the value) into the model (this is discussed below and in
           NeuralRWIPModel.py). Furthermore, as the same forcePop is used for each model to reduce the number of spike injectors used, i.e. is connected to each
           model's forceBufferPop, and so only the forceBufferPop in model X (i.e. the model next to calculate a RWIP state) activates (i.e. receives the force
           value) the RWIPModActForceCell associated to model X fires (activates) along with the forcePop.  Thus,  RWIPModActForceCell indicates that model X is
           next to calculate a RWIP state and enables model X's forceBufferPop to receive the force value from (be activated by) self.forcePop.  This is described
           in more detail in NeuralRWIPModel.py.  Note that, due to the above, the force value injected is delayed entering a model and therefore the theta and
           thetadot values are also delayed so that all values are received by the model at the same time (this is described below).  The first model (i.e.
           self.RWIPModels[0]) contains an extra input population initForcePop, which is used to represent the initial force value, thus, is only used once, i.e.
           at the start of the simulation.  This is not a set of spike injectors so bypasses forceBufferPop.            
       v)  output population nextThetaPop, which represents the next theta output value.  This is also the next connected
           model's thetaPop input population.
       vi) output population nextThetaDotPop, which represents the next thetadot output value. This is also the next connected model's
           thetaDotPop input population.
   
   The RWIP simulation system works as follows when the simulation starts:
       A) populations thetaPop, thetaDotPop and initForcePop in the first model (i.e. self.RWIPModels[0]) are activated (fire) at 0ms, which represent the
          initial simulation and force values.  These populations are activated in self.RWIPModels[0] at the start of the simulation as it calculates the
          first (next) RWIP state.  Neurons in a model's thetaPop and thetaDotPop fire at the same time, self.RWIPCalcTimePlusDelay ms after its previous
          model's.  In other words, neurons in a model's thetaPop and thetaDotPop fire every self.RWIPCalcTimePlusDelay*3 ms.  Thus, each model takes
          self.RWIPCalcTimePlusDelay ms (neurally) to calculate and output the next RWIP state, which includes a delay (of self.RWIPOutputDelay+2.0 ms) sending
          the output to the next model, where self.RWIPOutputDelay is a user defined delay.
          More specifically, it takes each model (neurally) self.RWIPCalcTimePlusDelay-self.RWIPOutputDelay-2.0 ms to calculate the next RWIP state.
       B) self.initActNeuron is activated at 0ms, which activates the cell in self.RWIPModActForceCells (i.e. self.RWIPModActForceCells[1]) associated to the
          the second model (self.RWIPModels[1]) (i.e. the model that calculates the second RWIP state) at
          self.RWIPCalcTimePlusDelay+self.waitForSpikes+self.timeToInjAfter-1.0 ms, so that self.RWIPModActForceCells[1] fires 1ms before the first force
          value is injected (i.e. forcePop is activated). Thus, self.RWIPModActForceCells[1] enables self.RWIPModels[1].forceBufferPop 1ms after it fires to
          receive a force value from forcePop, therefore, self.RWIPModels[1].forceBufferPop can receive a force value from forcePop (i.e. neurons in forceBufferPop will
          fire) 2ms (or more) after self.RWIPModActForceCells[1] fires.  Each cell in self.RWIPModActForceCells fires self.RWIPCalcTimePlusDelay ms after the
          previous thereafter.  In other words, a cell in self.RWIPModActForceCells: 1. fires 1ms before a force value is injected (i.e. the forcePop is
          activated), 2.enables the cell's associated model's forceBufferPop 1ms after to receive a force value from forcePop, therefore, forceBufferPop can receive
          a force value from forcePop (i.e. the neurons in forceBufferPop will fire)  2ms (or more) after the cell fired and 3.fires every self.RWIPCalcTimePlusDelay*3
          ms.
       C) the first force value is injected (i.e. forcePop is initially activated) at self.RWIPCalcTimePlusDelay+self.waitForSpikes+self.timeToInjAfter ms,
          and as the second model's RWIPModActForceCell fires 1ms before (as discussed above), the neurons in forcePop activates the same neurons in the
          second model's forceBufferPop (i.e. self.RWIPModels[1].forceBufferPop) self.waitForSpikes+self.timeToInjAfter+1 ms after the first model
          (self.RWIPModels[0]) calculated the previous RWIP state (note, it may take longer for the forceBufferPop to receive input from forcePop as it is
          unknown when SpiNNaker will receive the injected spikes, which is one of the reasons a force input buffer is used).  In other words, it is injected
          (i.e.forcePop is activated) self.waitForSpikes+self.timeToInjAfter ms after the neurons in self.RWIPModels[1]'s thetaPop and thetaDotPop fired (
          activated), which is within the output delay between self.RWIPModels[0] and self.RWIPModels[1].  A force value is
          injected (i.e. forcePop is activated) every self.RWIPCalcTimePlusDelay ms thereafter.  In other words, it is injected
          self.waitForSpikes+self.timeToInjAfter ms after each model X has calculated a RWIP state, i.e. after the neurons in model X's output populations/
          its next connected model's input populations (thetaPop and thetaDotPop) have fired/activated, which is
          within the output delay between model X and its next connected model.          

          Note, the injected force value is delayed reaching a model (i.e. a model's HighIDForcePop population, which the model's forceBufferPop is
          connected to) by self.RWIPOutputDelay+2.0-self.timeToInjAfter-self.waitForSpikes ms.  In other words, a HighIDForcePop fires 
          self.RWIPOutputDelay+2.0-self.timeToInjAfter-self.waitForSpikes ms after forcePop is activated.
          The force value is held by a model's forceBufferPop for a maximum of self.ForceBufferDuration-1.0ms (where
          self.ForceBufferDuration=self.RWIPOutputDelay+2.0-self.timeToInjAfter-self.waitForSpikes).  In other words, the time the force value
          is held by a forceBufferPop is dependent on when forceBufferPop receives the spikes from forcePop.  Thus, the force value is received by the
          model at the same time the theta and theta dot input values are received.
          
   Using 3 connected models allows each model to return to its resting state before calculating another RWIP state.
2. sets up communication between the world and spinnaker (live injections and receivers)
3. runs a 5s RWIP simulation, where the force values are injected into SpiNNaker over the 5s period
   (i.e. at each RWIP simulation time step) and the theta and thetadot values are calculated at each time step from the number of neurons that
   fired in the output populations (nextThetaPop and nextThetaDotPop) of the model that calculated the RWIP state.
4. saves the motion/movement of the RWIP, i.e. the theta and thetadot values over the 5s period.  
"""
from NeuralRWIPModel import RWIPClass #20
from numpy import array, hstack, linspace, pi
from numpy import zeros, cos, sin, arange, around
from scipy.integrate import odeint
from matplotlib import pyplot as plt
from matplotlib import animation
from matplotlib.patches import Rectangle
from pyNN.utility.plotting import Figure, Panel
from sympy import symbols
from sympy.physics.mechanics import *
from numpy import array, hstack, zeros, linspace, pi, ones
import os
import math
from decimal import *
import numpy as np
import sys
import spynnaker8.external_devices as externaldevices
import time as ti
import pyNN.spiNNaker as sim

class SpiNNakerRWIPSystemClass:

    def __init__(self):
        
        ##RWIP Parameter Values
        self.dt=0.01            #RWIP simulation time step (10ms)
        self.ADt=70.0*self.dt   #A*self.dt
        self.BDt=1.08*self.dt   #B*self.dt.  A and B are based on the physical quantities of the RWIP (see SpiNNakerRWIPModel.py for more information)
        self.friction=0.6       #friction torque
        self.RWIPValues=[0.088,0.001] # RWIP (simulation) starting values, i.e. [theta, thetadot]

        ##RWIP Simulation Parameters
        self.duration=5#0.73 #5                          ##duration of the RWIP simulation in seconds.
        self.numTimeSteps=int(self.duration/self.dt)   ##number of RWIP simulation time steps. 
        
        ##RWIP animation parameters
        self.numframes=60*self.duration   ##number of animation frames that can be created from the simulation when 60 are created every second
        self.NthFrame=round(self.numTimeSteps/self.numframes) ##the theta values to record for the animation (i.e. one for every self.NthFrame time steps) 
        self.numThetaAnim=np.floor(self.numTimeSteps/self.NthFrame)   ##number of theta values to record for the entire RWIP simulation
        self.thetaAnim=[]  ##stores the theta value for every self.NthFrame of the simulation
        self.thetaAnim.append([self.RWIPValues[0]])  ##record the initial angle (theta value) of the RWIP
        self.l=0.6       #length of RWIP arm

        #initialise arrays and variables for the RWIP system results
        self.thetaValues=zeros(self.numTimeSteps) ##stores the theta value at each time step
        self.thetaDotValues=zeros(self.numTimeSteps) ##stores the thetaDot value at each time step
        self.forceValues=zeros(self.numTimeSteps) ##stores the force value at each time step
        self.timeValues=zeros(self.numTimeSteps)   ##stores the time at each time step  
        self.thetaValues[0]=self.RWIPValues[0]     #record the initial theta value
        self.thetaDotValues[0]=self.RWIPValues[1]  #record the initial thetadot value

        ##PID controller initial values
        self.previousProp=self.RWIPValues[0] #previous (last time step) proportional value, i.e. the previous error (difference) between the current angle
                                             #and the desired angle (i.e. 0 (upright))
        self.integral=0                      #intergral of errors over time


        ##NEURAL AND SPINNAKER SYSTEM PARAMETER VALUES
        self.numRWIPModels=3  #number of RWIP models to build and connect.  Currently this program has only been tested for 3.
        self.RWIPModels=[]   #set of RWIP model objects
        self.curRWIPModel=[]        ##the RWIP model about to/currently calculating the next RWIP state
        self.nextRWIPModel=[]    ##the next RWIP model to calculate a RWIP state
        self.nest=0                 ##neural simulator to use.  use SpiNNaker        
        self.numNeuronsPerCore=175  #number of neurons to simulate on each SpiNNaker core
        self.additionalWt=0.012     #additional weight (input) required to cause some neurons to fire.  This is important due to SpiNNaker's limitations
        self.minWeight=0.628912   #the minimum weight (input) to cause an if_cond_exp neuron to fire (activate)
        self.neuralSimTimeStep=1.0    #neural simulation time step (1.0ms)
        self.waitForSpikes=2.0       #time (neural) to wait for spikes from SpiNNaker after a model has finished calculating a
                                     #RWIP state (i.e. after a model's output populations have activated). Needs to be 1.0 or more.  (see the start of the file
                                     #for more information).

        self.timeToInjAfter=1.0  #time (neural) after a model has calculated a RWIP state + self.waitForSpikes to
                                 #inject the force value via forcePop into self.curRWIPModel.forceBufferPop.  However, as described at the start of the file,
                                 #self.curRWIPModel.forceBufferPop may receive the spikes from forcePop later than this.(see the start of the file for more
                                 #information).
        self.RWIPOutputDelay=15.0          ## the output delay (neural time) between RWIP models, i.e. the delay sending the output (the next theta and
                                           ##thetadot values) from one model to its next connected model (see the start of the file for more information). 
        self.ForceBufferDuration=self.RWIPOutputDelay+2.0-self.timeToInjAfter-self.waitForSpikes   #ForceBufferDuration cannot be >RWIPOutputDelay+1
                                   #This defines the period of time (neural) a forceBufferPop in a model is enabled to receive the force value from forcePop
                                   #In other words, its the time a forceBufferPop will wait for spikes from self.forcePop and the maximum time forceBufferPop can
                                   #hold the spikes (be activated) for, i.e. the time they are held for is dependent on when the spikes are received (see the start of the file for
                                   #more information.
        self.sim=self.init() #initialise the neural simulation
        self.timePerLayer=1.0 #the time (neural) after a pre synaptic neuron a post synaptic neuron will fire in SpiNNaker if it receives sufficient input
                              #(i.e. self.minWeight) and the connection delay and neural time step is 1.0ms.
        
        self.RWIPCalcTimePlusDelay=0     #the time (neurally) it takes a RWIP model to calculate the next RWIP state, i.e. the next theta and thetadot values.
                                         #This is calculated as a model is created.  It also includes the model's output delay (i.e. self.RWIPOutputDelay
                                         #+ 2.0ms) (see above for more information).   
        self.time_scale_factor=10.0  #####the factor to reduce the speed of the neural simulation by.  This needs to match the time_scale_factor in the spynnaker.cfg file
        self.thetaDotValue=0        #number of neurons that fired in the thetaDot (input) population of a RWIP model
        self.thetaValue=0           #number of neurons that fired in the theta (input) population of a RWIP system
        self.forcePop=[]    #population of spike injector neurons used to inject the force value into SpiNNaker
        self.initForcePop=[]   #population of neurons that represent the initial force value, which is connected to the first RWIP model (i.e. self.RWIPModels[0])
                               #and used once, i.e. activated at the start of the RWIP simulation.                               
        self.RWIPModActForceCells=[] #this contains a set of neurons, one for each RWIP model (see the start of the file for more information). 
        self.initActNeuron=[]       #a neuron activated at the start of the RWIP simulation to activate the second neuron in self.RWIPModActForceCells at
                                    #a specific time. (see the start of the file for more information).
                                                           
        self.inputMinMaxVals=[[-0.045568,0.088],[-0.1764,0.03],[-7.2,3.69]]  #min and max values that can be represented by the thetaPop, thetaDotPop and
                                                                             #forcePop/forceBufferPop populations,
                                                                             #i.e. [[thetaMin,thetaMax],[thetadotMin, thetadotMax],[forceMin,forceMax]]
        self.NeuronValues=[0.00045,0.0018,0.015]  #the value each neuron in populations thetaPop, thetaDotPop and forcePop represent, i.e.
                                                  #[theta, thetadot, force]
        self.thetaDotNeuronValue=self.NeuronValues[1]  ##the value each neuron in thetaDotPop in a RWIP model represents
        self.thetaNeuronValue=self.NeuronValues[0]   ##the value each neuron in thetaPop in a RWIP model represents.
        
        
    def InjectForceAndGetNextRWIPState(self,label,sender):

        ##This function is called as soon as the RWIP simulation starts on SpiNNaker.  It runs alongside SpiNNaker (in real time)
        ##and does the following while the simulation is running:
        ##   1. keeps track of the RWIP model about to/currently calculating the next RWIP state (i.e. self.curRWIPModel) and the next RWIP model connected to
        ##      it (i.e. self.nextRWIPModel).
        ##   2. calculates and injects a force value into SpiNNAker (i.e. into model self.curRWIPModel) every self.RWIPCalcTimePlusDelay ms
        ##      (i.e. self.waitForSpikes+self.timeToInjAfter ms after thetaPop and thetaDotPop in self.curRWIPModel fire), where the first
        ##      injection is to the second model (i.e. self.RWIPModels[1]) at self.RWIPCalcTimePlusDelay+self.waitForSpikes+self.timeToInjAfter ms.        
        ##   3. calculates the theta and thetadot values from the number of neurons that fired in the output populations of self.curRWIPModel (i.e. the
        ##      thetaPop and thetaDotPop (input) populations of self.nextRWIPModel)


        ## Note, the speed of the neural simulation on SpiNNaker needs to be reduced by a factor of self.time_scale_factor so it works, thus, the waiting and
        ## injection times need to be increased by a factor of self.time_scale_factor so that the force values are injected, and the next theta and thetadot values are 
        ## calculated, at the correct times.  
                                                                
        SimStartTime=ti.time()*1000.0  ##records the start time, in real time, of the simulation.  Note, *1000 to get the current system time in milliseconds

        ##for each RWIP simulation time step
        
        for timeStep in range(1,self.numTimeSteps):
            
            if(timeStep==1):

                ##wait for the first RWIP model (i.e. self.curRWIPModel=self.RWIPModels[0]) to calculate the next RWIP state and wait self.waitForSpikes
                ##for any spikes from its output (i.e. from SpiNNaker).  We divide by 1000.0 to convert milliseconds back into system time.  Note,
                ##self.RWIPModels[0] does not receive any input until self.RWIPOutputDelay+2.0 ms.

                WaitForFirstSys=self.RWIPCalcTimePlusDelay+self.waitForSpikes   
                                                                         
                ti.sleep((WaitForFirstSys/1000.0)*self.time_scale_factor)

                ##calculate the time (in real time) to inject the force value into the forceBufferPop of the second system (i.e.self.nextRWIPModel=
                ##self.RWIPModels[1], i.e. activate forcePop
                nextInjection=SimStartTime+((WaitForFirstSys+self.timeToInjAfter)*self.time_scale_factor)
            else:

                #####if the function has got to this point too fast, i.e. before the next injection is required, wait.
                timeLeftToWait=nextInjection-(ti.time()*1000.0)
                if(timeLeftToWait>0):
                    ti.sleep(timeLeftToWait/1000.0)

                ##inject the force value into SpiNNaker, i.e. into self.curRWIPModel.forceBufferPop via self.curRWIPModel.forcePop
                ForcePopInclZero=1  #population contains a neuron that represents a 0.
                self.curRWIPModel.ActivateCellsUsingValue(self.curRWIPModel.forcePop, Force, ForcePopInclZero,sender) 

                ##wait until self.timeToInjAfter ms before the next force value needs to be calculated and injected
                
                ti.sleep((((self.RWIPCalcTimePlusDelay-self.timeToInjAfter)/1000.0)*self.time_scale_factor)) #the function injects a force value every
                                                                                                        #self.RWIPCalcTimePlusDelay ms, where the initial
                                                                                                        #injection is at WaitForFirstSys (when the initial RWIP
                                                                                                        #model finishes and the force value is calculated)+
                                                                                                        #self.timeToInjAfter.  Thus, at every timeStep>1, the 
                                                                                                        #function waits self.RWIPCalcTimePlusDelay-self.timeToInjAfter
                                                                                                        #ms for self.curRWIPModel to finish.  A force value is 
                                                                                                        #then calculated and injected self.timeToInjAfter ms after.
                 
                #####calculate the time (in real time) to inject the force value via forcePop into self.nextRWIPModel.forceBufferPop, i.e. SpiNNaker.
                nextInjection+=(self.RWIPCalcTimePlusDelay)*self.time_scale_factor  


            ######obtain the next theta and thetadot values (next RWIP state) represented by the input populations thetaPop and
            ######thetaDotPop in model self.nextRWIPModel, which are also the output populations of self.curRWIPModel
            NextThetaDot=self.thetaDotValue  ##the thetadot value, which can be negative or positive.
            self.thetaDotValue=0             ##reset the number that fired for calculating the next RWIP state
            Nexttheta=self.thetaValue        ##the theta value, which can be negative or positive
            self.thetaValue=0                           ##reset the number that fired for calculating the next RWIP state

            ###store values
            self.thetaValues[timeStep]=Nexttheta          ##store the theta value at each time step
            self.thetaDotValues[timeStep]=NextThetaDot    ##store the thetadot value at each time step
            self.timeValues[timeStep]=(timeStep)*self.dt  ##store the time at each time step
            
            ####calculate and store the force value
            Force=self.PIDControllerRWIP(self.thetaValues[timeStep]) # calculate the new force value
            self.forceValues[timeStep]=Force

            ##record the theta (value) every self.NthFrame time steps for the animation.          
            if(timeStep%self.NthFrame==0): 
                self.thetaAnim.append([Nexttheta])

            ###track the RWIP systems
            ##the RWIP system that will calculate the next RWIP state at timeStep+1 
            self.curRWIPModel=self.nextRWIPModel
            #the next system connected to self.curRWIPModel, i.e. the system that will calculate the next RWIP state at timeStep+2
            self.nextRWIPModel=self.curRWIPModel.NextConnectedModel

    def receive_spikes(self,label, time, neuron_ids):
        # This functions listens for spikes from the theta and thetadot populations of self.nextRWIPModel and calculates the theta and thetadot value
        # from the neurons that fired.

        for neuron_id in neuron_ids:  #for each neuron that fired
            if(self.nextRWIPModel.thetaDotPop[0].label==label):
                negnum=self.nextRWIPModel.thetaDotPop[1][1]   ##number of neurons in the thetaDotPop representing a negative value.
                if(neuron_id<negnum):                           #if the neuron represents a negative value
                    self.thetaDotValue-=self.thetaDotNeuronValue 
                else:
                    self.thetaDotValue+=self.thetaDotNeuronValue 
            if(self.nextRWIPModel.thetaPop[0].label==label):
                negnum=self.nextRWIPModel.thetaPop[1][1]   ##number of neurons in the thetaPop representing a negative value.
                if(neuron_id<negnum):                        #if the neuron represents a negative value
                    self.thetaValue-=self.thetaNeuronValue                         
                else:
                    self.thetaValue+=self.thetaNeuronValue                         
    
    
    def send_spikes_init(self,label,sender):
        ##This function injects a spike into SpiNNaker via the first injector neuron of the given population (label)
        sender.send_spike(label,0,send_full_keys=True)

    def setUpLiveReceivingAndCallBacks(self):
        #This function sets up spike listening connections for populations thetaPop and thetaDotPop in the RWIP models.  Also sets up
        #callbacks. 
        live_spikes_connection_receive = externaldevices.SpynnakerLiveSpikesConnection(
            receive_labels=[self.RWIPModels[0].thetaDotPop[0].label,self.RWIPModels[1].thetaDotPop[0].label,
            self.RWIPModels[2].thetaDotPop[0].label,self.RWIPModels[0].thetaPop[0].label,self.RWIPModels[1].thetaPop[0].label,
            self.RWIPModels[2].thetaPop[0].label],
            local_port=19996)
        
        for eachModel in range(0,self.numRWIPModels):
            externaldevices.activate_live_output_for(
                self.RWIPModels[eachModel].thetaDotPop[0], database_notify_host="localhost",
                database_notify_port_num=19996)       
            externaldevices.activate_live_output_for(
                self.RWIPModels[eachModel].thetaPop[0], database_notify_host="localhost",
                database_notify_port_num=19996)
            live_spikes_connection_receive.add_receive_callback(
                self.RWIPModels[eachModel].thetaDotPop[0].label, self.receive_spikes)
            live_spikes_connection_receive.add_receive_callback(
                self.RWIPModels[eachModel].thetaPop[0].label, self.receive_spikes)


    def setUpInjectionsAndCallBacks(self):
        ###This function sets up a spike injector connection for forcePop and sets up a callback so that the InjectForceAndGetNextRWIPState
        ###function is called at the start of the simulation 
        
        #set up python injector connection for the force population
        liveSpikesInjector = externaldevices.SpynnakerLiveSpikesConnection(local_port=19999,
                                send_labels=[self.forcePop[0].label])#register python injector with injector connection        
        
        # Set up callbacks to occur at the start of simulation    
        liveSpikesInjector.add_start_resume_callback(self.forcePop[0].label,self.send_spikes_init) #inject a test spike
        liveSpikesInjector.add_start_resume_callback(self.forcePop[0].label,self.InjectForceAndGetNextRWIPState)    

    def PIDControllerRWIP(self,theta):
        #This function implements a proportional, integral and derivative controller.  It calculates and returns the force (torque) required from the
        #reaction wheel to stabilise the inverted pendulum.  In other words, it calculates the force value required to acheive the
        #desired angle of the RWIP (which is upright i.e. 0) from the current angle (the given theta value).

        ##equation values

        prop=theta                           #proportional term. error (difference) between the current and desired angles
        self.integral+=prop*self.dt             #intergral term. the intergral of errors overtime.
        derivative=(prop-self.previousProp)/self.dt  #derivative term. derivative of the error signal

        Kp=-80  #the proportional term is weighted by this factor                        
        Ki=-15  #the integral term is weighted by this factor  
        Kd=-10  #the derivative term is weighted by this factor  

        ##calculate the force value
        Force=(Kp*prop)+(Ki*self.integral)+(Kd*derivative)

        ##record the current proportional value.  This is used with the PID controller at the next time step
        self.previousProp=prop         #previous (the last time step) proportional (error) value

        return Force

    def ConnectRWIPModActForceCells(self):
        ##This function connects the neurons in self.RWIPModActForceCells consecutively to form a loop, where each neuron is associated to a different RWIP model
        ##and each neuron, in sequence, fires.  A neuron (initActNeuron) is created and connected to the second neuron in self.RWIPModActForceCells so that
        ##the second neuron fires initdelay ms after initActNeuron (activated at the start of the simulation).  See the start of the file for more information
        ##about self.RWIPModActForceCells.

        #####create initActNeuron
        model=1
        self.initActNeuron=self.RWIPModels[0].createPopulation(1,model) 


        ######connect initActNeuron to the second neuron in self.RWIPModActForceCells using self.minWeight so that the initial neuron activates
        ######the second neuron.
        initdelay=self.RWIPCalcTimePlusDelay+self.waitForSpikes+self.timeToInjAfter-1.0   #the second neuron in  self.RWIPModActForceCells needs to fire
                                                                                          #1ms before the forcePop. 

        ExcConnector=[(0, 0, self.minWeight, initdelay)]
        connList = self.sim.FromListConnector(ExcConnector)
        input_proj=self.sim.Projection(self.initActNeuron, self.RWIPModActForceCells[1], connList,  receptor_type='excitatory')

        #####connect the neurons consecutively to form a loop
        delay=self.RWIPCalcTimePlusDelay
        ExcConnector=[(0, 0, self.minWeight, delay)]
        connList = self.sim.FromListConnector(ExcConnector)
        for eachModel in range(0,  self.numRWIPModels):
            nextSystem=(eachModel+1)%self.numRWIPModels
            input_proj=self.sim.Projection(self.RWIPModActForceCells[eachModel], self.RWIPModActForceCells[nextSystem], connList,  receptor_type='excitatory')

    def CreateRWIPSimulationSystem(self):

        ##This function creates the RWIP simulation system, i.e.:
        ##   1. creates self.numRWIPModels (currently 3) RWIP forward models and connects them consecutively to form a loop, so that each model, in sequence,
        ##      calculates and outputs a RWIP state.
        ##   2. extracts from the first model (i.e.self.RWIPModels[0]):
        ##          A) self.initForcePop, which 1. is used to represent the initial force value, 2.is a set of if_cond_exp neurons, 3.is only used once (at the
        ##             start of the simulation).  This is only used as an input population in self.RWIPModels[0]. (see above for more information)
        ##          B) self.forcePop, which represents the force value after the first time step, is a set of spike injector neurons and used to inject
        ##             the force value into SpiNNaker.  This is used as an input population in all RWIP models.
        ##   3. extracts the RWIPModActForceCell from each model and connects them.

        inputPops=[]  ##input populations for the first RWIP model (i.e. self.RWIPModels[0]) need to be created.
        print "creating 3 RWIP models"
        for eachModel in range(0,self.numRWIPModels):  #for each RWIP model
            print "creating RWIP model: "+str(eachModel+1)
            #output populations of eachModel 
            if(eachModel==(self.numRWIPModels-1)): #if the last model
                ##the input populations of the first model (RWIPModels[0]) (i.e. thetaPop and thetaDotPop) are to be used as the output populations
                ##(i.e. nextThetaPop and nextThetaDotPop) of the last model (RWIPModels[self.numRWIPModels-1])
                OutputPops=[self.RWIPModels[0].thetaPop, self.RWIPModels[0].thetaDotPop] 
            else:
                OutputPops=[]  ##create output populations
                
            ##create a RWIP model
            self.RWIPModels+=[RWIPClass(self.sim, self.minWeight, self.dt,self.nest,self.additionalWt,self.ForceBufferDuration,self.numRWIPModels,
                                        self.inputMinMaxVals,eachModel,self.friction,self.BDt, self.ADt,self.RWIPOutputDelay,OutputPops,self.timePerLayer,
                                        self.NeuronValues,inputPops)]
            if(eachModel==0):    #if self.RWIPModels[0]
                self.forcePop=self.RWIPModels[eachModel].forcePop           ##extract the self.forcePop used to inject the force value into SpiNNaker
                self.initForcePop=self.RWIPModels[eachModel].initForcePop      ##extract self.initForcePop
    
            self.RWIPModActForceCells+=[self.RWIPModels[eachModel].RWIPModActForceCell]   ##extract RWIPModActForceCell from eachModel

            ###give the previous RWIP model its next connected RWIP model
            if(eachModel>0):
                self.RWIPModels[eachModel-1].NextConnectedModel=self.RWIPModels[eachModel]
            if(eachModel==self.numRWIPModels-1):                
                self.RWIPModels[eachModel].NextConnectedModel=self.RWIPModels[0]

            ##input populations of the next RWIP model
            inputPops=self.RWIPModels[eachModel].outputPops     ##the input populations (thetaPop and thetaDotPop) of the next model are the output populations of
                                                                ##eachModel            
            inputPops+=[self.forcePop]                          
                
        self.RWIPCalcTimePlusDelay=self.RWIPModels[0].RWIPCalcTimePlusDelay  #the time (neurally) it takes a RWIP model to calculate the next RWIP state +
                                                                             #the output delay.  See the start of the file for more information.
                                                                            
        self.ConnectRWIPModActForceCells()
         
    def InitialiseNeuralRWIPSimSystem(self):
        ##This function initialises the RWIP simulation, including activating the neurons in the input populations (i.e. thetaPop, thetaDotPop and initForcePop)
        ##of the first RWIP model (i.e. self.RWIPValues[0]) associated to the given initial values.
    
        ##calculate the initial force value
        Force=self.PIDControllerRWIP(self.RWIPValues[0])
        self.forceValues[0]=Force

        ####activate self.initForcePop so that it represents the Force value.
        inclZero=1  #a neuron representing a 0 value is included in this population
        sender=[] ##the neurons are not spike injectors                                                                                         
        self.RWIPModels[0].ActivateCellsUsingValue(self.initForcePop, Force, inclZero,sender) #cells activate at 0ms
                                                                                            
        ####activate neuron initActNeuron that triggers the second neuron in self.RWIPModActForceCells
        actVoltage=-51.0    #this is a voltage > the membrane potential threshold.  This is used to activate a neuron
        self.initActNeuron.initialize(v=actVoltage)

        ######activate neurons in the thetaPop and thetaDotPop of the first model (i.e. self.RWIPModels[0]) so they represent the initial values.
        InitialThetaValue=self.RWIPValues[0]
        InitialThetaDotValue=self.RWIPValues[1]
        inclZero=0  #a neuron representing a 0 value is not included in this population
        sender=[] ##the neurons are not spike injectors   
        self.RWIPModels[0].ActivateCellsUsingValue(self.RWIPModels[0].thetaPop, InitialThetaValue,inclZero,sender) 
        self.RWIPModels[0].ActivateCellsUsingValue(self.RWIPModels[0].thetaDotPop, InitialThetaDotValue,inclZero,sender)

        ######the first and next RWIP models to calculate RWIP states.
        self.curRWIPModel=self.RWIPModels[0]#the first system to calclate the next RWIP state
        self.nextRWIPModel=self.curRWIPModel.NextConnectedModel #the next system to calculate a RWIP state

    def init(self):
        ##this function sets up the SpiNNaker neural simulation.
        
        Mxdelay=self.RWIPOutputDelay+45.0  ##the maximum delay used in a RWIP model
        sim.setup(timestep=self.neuralSimTimeStep, min_delay=1.0, max_delay=Mxdelay, debug=0)
        
        sim.set_number_of_neurons_per_core(sim.IF_cond_exp,self.numNeuronsPerCore)
        return sim


    def RunRWIPSimulation(self):
        #This function runs the RWIP simulation, i.e. runs the RWIP models.

        #the neural time required to calculate self.numTimeSteps of RWIP simulation, +1 to include the last time step.
        durationTime=(self.RWIPCalcTimePlusDelay*self.numTimeSteps)+1       
                                                                            
        #run the simulation
        print "RWIP Models Uploading Then Simulation Will Start"
        self.sim.run(durationTime)                        
        print "RWIP Simulation Finished"
        self.sim.end()

    def savefig(self,fileName,x,y,xaxisLabel,yaxisLabel,title):
        #This function creates a figure of the data given in x (x axis) and y (y axis) and using the labels in xaxisLabel, yaxisLabel and title.
        #It saves the figure to the file given by fileName.
        plt.plot(x,y) 
        plt.xlabel(xaxisLabel)
        plt.ylabel(yaxisLabel)
        plt.title(title)
        plt.savefig(fileName)
        plt.close()
        #plt.show()
        
    def saveResults(self,directory):
        #This creates and saves a figure of the theta, thetadot, and force values produced by the RWIP simulation to files theta.png, thetadot.png and force.png to the given
        #directory.

        thetaValues=self.thetaValues[0:self.numTimeSteps]
        thetaDotValues=self.thetaDotValues[0:self.numTimeSteps]

        forceValues=self.forceValues[0:self.numTimeSteps]
        timeValues=self.timeValues[0:self.numTimeSteps]

        #print "theta"
        #print thetaValues
        #print "thetaDotValue"
        #print thetaDotValues 
        #print "force"
        #print forceValues

        if not os.path.isdir(directory):  #if directory does not exist
            os.makedirs(directory)        #make it

        print "Saving Results"
        self.savefig(directory+"/"+"theta.png",timeValues,thetaValues,'Time (s)','$\Theta$','RWIP Movement/Angle ($\Theta$)')
        self.savefig(directory+"/"+"thetadot.png",timeValues,thetaDotValues,'Time (s)','$\dot{\Theta}$','RWIP Angular Velocity ($\dot{\Theta}$)')
        self.savefig(directory+"/"+"force.png",timeValues,forceValues,'Time (s)','F','Force Values')
    
    def AnimatePendulum(self):
        #This produces an animation of the inverted pendulum movement.
        
        t=linspace(0, self.duration, self.numThetaAnim)
        thetaAnimValues=np.array(self.thetaAnim)
        self.animate_pendulum(t, thetaAnimValues, self.l)

        plt.show()

    def animate_pendulum(self,t, states, length, filename=None):
        """Animates the n-pendulum and optionally saves it to file.

        Parameters
        ----------
        t : ndarray, shape(m)
            Time array.
        states: ndarray, shape(m,p)
            State time history.
        length: float
            The length of the pendulum links.
        filename: string or None, optional
            If true a movie file will be saved of the animation. This may take some time.

        Returns
        -------
        fig : matplotlib.Figure
            The figure.
        anim : matplotlib.FuncAnimation
            The animation.

        """
        # the number of pendulum bobs
        numpoints=1

        # first set up the figure, the axis, and the plot elements we want to animate
        fig = plt.figure()

        # create the axes
        ax = plt.axes(xlim=(-1, 1), ylim=(-1.1, 1.1), aspect='equal')

        # display the current time
        time_text = ax.text(0.04, 0.9, '', transform=ax.transAxes)

        # blank line for the pendulum
        line, = ax.plot([], [], lw=2, marker='o', markersize=6)

        # initialization function: plot the background of each frame
        def init():
            time_text.set_text('')
            line.set_data([], [])
            return time_text, line,

            # animation function: update the objects
        def animate(i):
            time_text.set_text('time = {:2.2f}'.format(t[i]))
            x= zeros((numpoints))
            y = zeros((numpoints))
        
            for j in arange(0, numpoints):
                x[j] = length * sin(states[i, j])
                y[j] = length * cos(states[i, j])
            line.set_data([0,x],[0, y])
            return time_text, line,
    
            # call the animator function
        anim = animation.FuncAnimation(fig, animate, frames=len(t), init_func=init,
                                   interval=t[-1] / len(t) * 1000, blit=False, repeat=False)
    
        anim.save('RWIPneuralpole.mp4')
        plt.show()

    def createSetupAndRunSystem(self,directory):
        ### This function creates and sets up the RWIP simulation system and the communication
        ### between the world and spinnaker (live injectiors and receivers), runs the RWIP simulation and
        ### saves the results, i.e. the theta, thetadot and force values over the time period (defined by duration), to files theta.png
        ### thetadot.png and force.png.

        ###create the RWIP simulation system        
        self.CreateRWIPSimulationSystem()

        ##set up communication between the world and spinnaker (live injectors and receivers)
        
        #set up spike listening connections for populations thetaPop and thetaDotPop in the RWIP models.  Also sets up callbacks.
        self.setUpLiveReceivingAndCallBacks()   

        #set up a spike injector connection for forcePop and a callback so that the InjectForceAndGetNextRWIPState function is called at the
        #start of the simulation   
        self.setUpInjectionsAndCallBacks()  

        ###initialise the RWIP simulation, i.e. the RWIP models.
        self.InitialiseNeuralRWIPSimSystem() 

        ###run the RWIP simulation
        self.RunRWIPSimulation()        

        ###save the results.
        self.saveResults(directory)

        ###produce and save a video/animation of the RWIP movement
        #self.AnimatePendulum()                      


args = sys.argv
numberArgs = args.__len__()
if (numberArgs > 1):
    directory=args[1]
else:
    directory='results/'

SpiNN=SpiNNakerRWIPSystemClass()  
SpiNN.createSetupAndRunSystem(directory)    #create a neural RWIP simulation system and run it to calculate the motion of the RWIP for 5s.
print "Simulation Finished and Results Saved"
